// Copyright (C) 2014 Benoit Chachuat, Imperial College London.
// All Rights Reserved.
// This code is published under the Eclipse Public License.

#ifndef MC__POLYMODEL_H
#define MC__POLYMODEL_H

#include <iostream>
#include <iomanip>
#include <typeinfo>
#include <sstream>
#include <string>
#include <stdarg.h>
#include <cassert>
#include <climits>
#include <limits>
#include <stdlib.h>
#include <complex>

#include "mcfunc.hpp"
#include "mcop.hpp"

#undef  MC__POLYMODEL_DEBUG
#define MC__POLYMODEL_CHECK

namespace mc
{

typedef unsigned long long poly_size;

//! @brief C++ base class for the computation of polynomial models for factorable functions: Environment
////////////////////////////////////////////////////////////////////////
//! mc::PolyModel is a C++ base class for definition of polynomial model
//! environment.
////////////////////////////////////////////////////////////////////////
class PolyModel
////////////////////////////////////////////////////////////////////////
{
public:

  /** @addtogroup POLYMOD Polynomial Model Arithmetic for Factorable Functions
   *  @{
   */
  //! @brief Constructor of polynomial model environment for <tt>nvar</tt> variables and order <tt>nord</tt>
  PolyModel
    ( const unsigned int nvar, const unsigned int nord )
    { _size( nvar, nord ); }
  //! @brief Destructor of polynomial model environment

  virtual ~PolyModel
    ()
    { _cleanup(); }

  //! @brief Number of variables in polynomial model environment
  unsigned int nvar
    ()
    const
    { return _nvar; };

  //! @brief Order of polynomial model environment
  unsigned int nord
    ()
    const
    { return _nord; };

  //! @brief Total number of monomial terms in polynomial model
  unsigned int nmon
    ()
    const
    { return _nmon; };

  //! @brief Const pointer to array of size <tt>nmon()*nvar()</tt> with variable exponents on each monomial term. The exponent for variable <tt>ivar</tt> in monomial term <tt>imon</tt> is at position <tt>imon*nvar()+ivar</tt>.
  const unsigned int* expmon
    ()
    const
    { return _expmon; };

  //! @brief Index of monomial term whose variable exponents are the same as those in array <tt>iexp</tt> (of size <tt>nvar()</tt>)
  unsigned int loc_expmon
    ( const unsigned int *iexp )
    const
    { return _loc_expmon( iexp ); }

  //! @brief Array of size <tt>_nord</tt> with indices of first monomial term of order <tt>iord=1,...,_nord</tt> in polynomial model
  const unsigned int* posord
    ()
    const
    { return _posord; }

  //! @brief Get binomial coefficient \f$\left(\stackrel{n}{k}\right)\f$
  poly_size get_binom
    ( const unsigned int n, const unsigned int k )
    const
    { return _get_binom( n, k ); }

  //! @brief Exceptions of mc::PolyModel
  class Exceptions
  {
  public:
    //! @brief Enumeration type for TModel exception handling
    enum TYPE{
      SIZE=1,	//!< Number of variable in polynomial model must be nonzero
      MAXSIZE,  //!< Maximum size of polynomial model reached (monomials indexed as unsigned int)
      UNDEF=-33 //!< Feature not yet implemented in mc::PolyModel
    };
    //! @brief Constructor for error <a>ierr</a>
    Exceptions( TYPE ierr ) : _ierr( ierr ){}
    //! @brief Error flag
    int ierr(){ return _ierr; }
    //! @brief Error description
    std::string what(){
      switch( _ierr ){
      case SIZE:
        return "mc::PolyModel\t Inconsistent polynomial model dimension";
      case MAXSIZE:
        return "mc::PolyModel\t Maximum size in polynomial model reached";
      case UNDEF:
        return "mc::PolyModel\t Feature not yet implemented in mc::PolyModel class";
      default:
        return "mc::PolyModel\t Undocumented error";
      }
    }

  private:
    TYPE _ierr;
  };
  /** @} */

private:  
  //! @brief Set polynomial model order <tt>nord</tt> and number of variables <tt>nvar</tt>
  void _size
    ( const unsigned int nvar, const unsigned int nord );

  //! @brief Clean up internal arrays
  void _cleanup();

protected:
  //! @brief Order of polynomial model
  unsigned int _nord;

  //! @brief Number of variables in polynomial model
  unsigned int _nvar;

  //! @brief Total number of monomial terms in polynomial model
  unsigned int _nmon;

  //! @brief Array of size <tt>_nord</tt> with indices of first monomial term of order <tt>iord=1,...,_nord</tt> in polynomial model
  unsigned int *_posord;

  //! @brief Order used to size _posord
  unsigned _posord_size;

  //! @brief Array of size <tt>_nmon*_nvar</tt> with variable exponents in monomial terms. The exponent for variable <tt>ivar</tt> in monomial term <tt>imon</tt> is at location <tt>imon*nvar()+ivar</tt>.
  unsigned int *_expmon;

  //! @brief Number of monomial coefficients used to size _expmon
  unsigned _expmon_size;

  //! @brief Array of <tt>(_nvar+_nord-1)*(_nord+1)</tt> contining binomial coefficients
  poly_size *_binom;

  //! @brief Maximum binomial coefficients in array _binom
  std::pair<unsigned int, unsigned int> _binom_size;

  //! @brief Have any of the model variables been modified?
  bool _modvar;

  //! @brief Populate array <tt>_bndmon</tt>
  void _set_bndmon();

  //! @brief Populate array <tt>_posord</tt> up to order <tt>nord</tt>
  void _set_posord
    ( const unsigned int nord );

  //! @brief Extend array <tt>_posord</tt> up to maximum order <tt>maxord</tt>
  void _ext_posord
    ( const unsigned int maxord );

  //! @brief Populate array <tt>_expmon</tt> up to order <tt>nord</tt>
  void _set_expmon
    ( const unsigned int nord );

  //! @brief Extend array <tt>_expmon</tt> up to order <tt>maxord</tt> to accomodate <tt>maxmon</tt> coefficients
  void _ext_expmon
    ( const unsigned int maxord, const bool full=false );
  
  //! @brief Generate variable exponents <tt>iexp</tt> for subsequent monomial order <tt>iord</tt>
  void _next_expmon
    ( unsigned int *iexp, const unsigned int iord ) const;

  //! @brief Get index of monomial term with variable exponents <tt>iexp</tt> in <tt>1,...,_nmon</tt>
  unsigned int _loc_expmon
    ( const unsigned int *iexp ) const;
    
  //! @brief Populate array <tt>_binom</tt> with binomial coefficients up to order <tt>nord</tt>
  void _set_binom
    ( const unsigned int nord );
    
  //! @brief Extend array <tt>_binom</tt> with binomial coefficients up to order <tt>nord</tt>
  void _ext_binom
    ( const unsigned int nord );

  //! @brief Get binomial coefficient \f$\left(\stackrel{n}{k}\right)\f$
  poly_size _get_binom
    ( const unsigned int n, const unsigned int k ) const;
};

//////////////////////////////// PolyModel ////////////////////////////////////

inline void
PolyModel::_size
( const unsigned int nvar, const unsigned int nord )
{
  if( !nvar ) throw Exceptions( Exceptions::SIZE );

  _nvar = nvar;
  _nord = nord; 
  _binom = new poly_size[nord?(nvar+nord-1)*(nord+1):nvar];
  _binom_size = std::make_pair( nvar+nord-1, nord+1 );
  _set_binom( nord );
  _posord = new unsigned int[nord+2];
  _posord_size = nord;
  _set_posord( nord );
  _nmon = _posord[_nord+1];
  _expmon = new unsigned int[_nmon*nvar];
  _expmon_size = _nmon;
  _set_expmon( nord );
  _modvar = true;
}

inline void
PolyModel::_set_posord
( const unsigned int nord )
{
  _posord[0] = 0;
  _posord[1] = 1;
  for( unsigned int i=1; i<=nord; i++ ){
    poly_size _posord_next = _posord[i] + _get_binom( _nvar+i-1, i );
    if( _posord_next > UINT_MAX )
      throw typename PolyModel::Exceptions( PolyModel::Exceptions::MAXSIZE );
    _posord[i+1] = _posord_next;
  }

#ifdef MC__POLYMODEL_DEBUG
  mc::display( 1, nord+2, _posord, 1, "_posord", std::cout );
#endif
}
    
inline void
PolyModel::_ext_posord
( const unsigned int maxord )
{
  if( maxord < _posord_size ) return;
  delete[] _posord;
  _posord = new unsigned int[maxord+2];
  _posord_size = maxord;
  _set_posord( maxord );
}

inline void
PolyModel::_set_expmon
( const unsigned int nord )
{
  unsigned int *iexp = new unsigned int[_nvar] ;
  for( unsigned int k=0; k<_nvar; k++ ) _expmon[k] = 0;
  for( unsigned int i=1; i<=nord; i++ ){
    for( unsigned int j=0; j<_nvar; j++ ) iexp[j] = 0;
    for( unsigned int j=_posord[i]; j<_posord[i+1]; j++ ){
      _next_expmon( iexp, i );
      for( unsigned int k=0; k<_nvar; k++ )
        _expmon[j*_nvar+k] = iexp[k];
    }
  }
  delete[] iexp;

#ifdef MC__POLYMODEL_DEBUG
  mc::display( _nvar, _expmon_size, _expmon, _nvar, "_expmon", std::cout );
#endif
}
  
inline void
PolyModel::_next_expmon
( unsigned int *iexp, const unsigned int iord ) const
{
  unsigned int curord;
  do{
    iexp[_nvar-1] += iord;
    unsigned int j = _nvar;
    while( j > 0 && iexp[j-1] > iord ){
      iexp[j-1] -= iord + 1;
      j-- ;
      iexp[j-1]++;
    }
    curord = 0;
    for( unsigned int i=0; i<_nvar; i++ ) curord += iexp[i];
  } while( curord != iord );
}
    
inline void
PolyModel::_ext_expmon
( const unsigned int maxord, const bool full )
{
  _ext_binom( maxord );
  _ext_posord( maxord ); 
  delete[] _expmon;
  if( full ) _expmon_size = std::pow(maxord+1,_nvar);
  else       _expmon_size = _posord[maxord+1];
  _expmon = new unsigned int[ _expmon_size*_nvar];
  _set_expmon( maxord );
  if( !full ) return;

  unsigned int *iexp = new unsigned int[_nvar];
  for( unsigned int iord=maxord+1, jmon=_posord[maxord+1];
       jmon<_expmon_size; iord++ ){
    _ext_binom( iord );
    _ext_posord( iord );
    if( _posord[iord] >= _posord[iord+1] )
      throw typename PolyModel::Exceptions( PolyModel::Exceptions::MAXSIZE );
    for( unsigned int ivar=0; ivar<_nvar; ivar++ ) iexp[ivar] = 0;
    for( unsigned int kmon=_posord[iord]; kmon<_posord[iord+1]; kmon++ ){
      _next_expmon( iexp, iord );
      bool lt_maxord = true;
      for( unsigned int ivar=0; ivar<_nvar; ivar++ ){
        if( iexp[ivar] > maxord ){ lt_maxord = false; break; }
        _expmon[jmon*_nvar+ivar] = iexp[ivar];
      }
      if( lt_maxord ){
#ifdef MC__POLYMODEL_DEBUG
	std::cout << jmon << ":";
        for( unsigned int ivar=0; ivar<_nvar; ivar++ )
	  std::cout << "  " << _expmon[jmon*_nvar+ivar];
	std::cout << std::endl;
#endif
        jmon++;
        if( jmon >= _expmon_size ) break;
      }
    }
  }
  delete[] iexp;
  return;
}
    
inline unsigned int
PolyModel::_loc_expmon
( const unsigned int *iexp ) const
{
  unsigned int ord = 0;
  for( unsigned int i=0; i<_nvar; i++ ) ord += iexp[i];
#ifdef MC__POLYMODEL_CHECK
  assert( ord<_nord+2 );
#endif
  unsigned int pos = _posord[ord];
  
  unsigned int p = _nvar ; 
  for( unsigned int i=0; i<_nvar-1; i++ ){
    p--;
    for( unsigned int j=0; j<iexp[i]; j++ )
      pos += _get_binom( p-1+ord-j, ord-j );
    ord -= iexp[i];
  }

  return pos;    
}
 
inline void
PolyModel::_set_binom
( const unsigned int nord )
{
  poly_size *p;
  unsigned int k;
  for( unsigned int i=0; i<_nvar+nord-1; i++ ){
    p = &_binom[i*(nord+1)];
    *p = 1;
    p++;
    *p = i+1;
    p++;
    k = ( i+1<nord? i+1: nord );
    for( unsigned int j=2; j<=k; j++, p++ ) *p = *(p-1) * (i+2-j)/j;
    for( unsigned int j=k+1; j<=nord; j++, p++ ) *p = 0.;
  }
#ifdef MC__POLYMODEL_DEBUG
  std::cout << "nvar: " << _nvar << "  nord: " << nord << std::endl;
  mc::display( _binom_size.second, _binom_size.first, _binom,
    _binom_size.second, "_binom", std::cout );
#endif
}
    
inline void
PolyModel::_ext_binom
( const unsigned int maxord )
{
  if( maxord < _binom_size.second ) return;
  delete[] _binom;
  _binom = new poly_size[(_nvar+maxord-1)*(maxord+1)];
  _binom_size = std::make_pair( _nvar+maxord-1, maxord+1 );
  _set_binom( maxord );
}

inline poly_size
PolyModel::_get_binom
( const unsigned int n, const unsigned int k ) const
{
#ifdef MC__POLYMODEL_CHECK
  assert( n<=_binom_size.first );
  assert( k<=_binom_size.second );
  assert( k<=n );
#endif
  return( n? _binom[(n-1)*_binom_size.second+k]: 1 );
}

inline void
PolyModel::_cleanup()
{
  delete[] _expmon;
  delete[] _posord;
  delete[] _binom;
}

//! @brief C++ base class for the computation of polynomial models for factorable functions: Variable
////////////////////////////////////////////////////////////////////////
//! mc::PolyVar is a C++ base class for definition of polynomial model
//! variables.
////////////////////////////////////////////////////////////////////////
template <typename T>
class PolyVar
////////////////////////////////////////////////////////////////////////
{

protected:
  //! @brief Pointer to corresponding polynomial model environment
  PolyModel* _PM;

  //! @brief Array of size <tt>_nmon</tt> with monomial coefficients of variable
  double* _coefmon;

  //! @brief Array of size <tt>_nord+2</tt> with bounds for all terms of degrees <tt>iord=0,...,_nord</tt> as well as the remainder bound at position <tt>_nord+1</tt> of variable (possibly inaccurate - see _bndord_uptd)
  mutable T* _bndord; 

  //! @brief Whether the bounds in bndord are up-to-date
  mutable bool _bndord_uptd;

  //! @brief Pointer to remainder bound of variable (possibly NULL if not computed)
  T* _bndrem;

  //! @brief Pointer to polynomial bound of variable
  mutable T* _bndpol;

public:
  //! @brief Order of polynomial model
  unsigned int nord
    ()
    const
    { return( _PM? _PM->nord(): 0 ); };

  //! @brief Number of variables in polynomial model
  unsigned int nvar
    ()
    const
    { return( _PM? _PM->nvar(): 0 ); };

  //! @brief Total number of monomial terms in polynomial model
  unsigned int nmon
    ()
    const
    { return( _PM? _PM->nmon(): 1 ); };

protected:
  //! @brief Index of first monomial term of order <a>iord</a> in polynomial model
  unsigned int _posord
    ( const unsigned int iord )
    const
    { return _PM->posord()[iord]; };

  //! @brief Const pointer to array of size <a>nvar</a> of variable exponents in monomial term <tt>imon</tt> of polynomial model
  const unsigned int* _expmon
    ( const unsigned int imon )
    const
    { return _PM->expmon()+imon*nvar(); };

  //! @brief Index of monomial term whose variable exponents are the same as those in array <tt>iexp</tt> (of size <tt>nvar()</tt>)
  const unsigned int _loc_expmon
    ( const unsigned int* iexp )
    const
    { return _PM->loc_expmon( iexp ); };

  //! @brief Get binomial coefficient \f$\left(\stackrel{n}{k}\right)\f$
  poly_size _get_binom
    ( const unsigned int n, const unsigned int k )
    const
    { return _PM->get_binom( n, k ); }

  //! @brief Initialize private/protected members of variable
  void _size
    ( PolyModel* env );

  //! @brief Reinitialize private/protected members of variable
  void _resize
    ( PolyModel* env );

  //! @brief Clean up private/protected members of variable
  void _cleanup
    ();

  //! @brief Set variable equal to <a>var</a>
  PolyVar<T>& _set
    ( const PolyVar<T>&var );

  //! @brief Center remainder error term <tt>_bndrem</tt>
  virtual void _center();

  //! @brief Set polynomial bound in variable as <tt>bndpol</tt>
  virtual void _set_bndpol
    ( const T&bndpol );

  //! @brief Set polynomial bound in variable as <tt>bndpol</tt>
  virtual void _set_bndpol
    ( const T*bndpol );

  //! @brief Unset polynomial bound in variable
  virtual void _unset_bndpol
    ();

   //! @brief Polynomial range bounder using specified bounder <a>type</a> (pure virtual)
  virtual T _polybound
    ( const int type )
    const
    = 0;
 
   //! @brief Polynomial range bounder using specified bounder <a>type</a> (pure virtual)
  virtual T _polybound
    ()
    const
    = 0;
 
  //! @brief Update private/protected members of variable after manipulation the coefficients (pure virtual)
  virtual void _update_bndord
    () const
    = 0;

public:
  /** @addtogroup POLYMOD Polynomial Model Arithmetic for Factorable Functions
   *  @{
   */
  //! @brief Constructor of variable linked to polynomial model environment <a>env</a>
  PolyVar
    ( PolyModel* env=0 )
    { _size( env ); }

  //! @brief Copy constructor of variable
  PolyVar
    ( const PolyVar<T>& var )
    { _size( var._PM );
      // Set variable not linked to any environment
      if( !_PM ){
        _coefmon[0] = var._coefmon[0];
        _bndord[0]  = var._bndord[0];
      }
      // Set variable linked to an environment
      else{
        for( unsigned int i=0; i<nmon(); i++ ) _coefmon[i] = var._coefmon[i];
        for( unsigned int i=0; i<nord()+2; i++) _bndord[i] = var._bndord[i];
      }
      _bndord_uptd = var._bndord_uptd;
      if( var._bndpol ) _bndpol = new T( *var._bndpol );
    }

  //! @brief Destructor of variable
  virtual ~PolyVar()
    { delete[] _coefmon; delete[] _bndord; delete _bndpol; }

  //! @brief Set polynomial model environment in variable as <tt>env</tt>
  virtual PolyVar<T>& set
    ( PolyModel* env )
    { _resize( env );
      _unset_bndpol(); _bndord_uptd = false;
      return *this; }
    //{ _PM = env; return *this; }

  //! @brief Set multivariate polynomial coefficients in variable as <tt>coefmon</tt>
  virtual PolyVar<T>& set
    ( const double* coefmon )
    { for( unsigned int imon=0; imon<(_PM?nmon():1); imon++ )
        _coefmon[imon] = coefmon[imon];
      _unset_bndpol(); _bndord_uptd = false;
      return *this; }

  //! @brief Set multivariate polynomial coefficients in variable as <tt>coefmon</tt> - only first <tt>coefmon.first</tt> coefficients are set
  virtual PolyVar<T>& set
    ( std::pair<unsigned int, const double*>& coefmon )
    { for( unsigned int imon=0; imon<(_PM?nmon():1); imon++ )
        _coefmon[imon] = ( imon<coefmon.first && coefmon.second?
	                   coefmon.second[imon]: 0. );
      _unset_bndpol(); _bndord_uptd = false;
      return *this; }

  //! @brief Set remainder term in variable as <tt>bndrem</tt>
  virtual PolyVar<T>& set
    ( const T& bndrem )
    { *_bndrem = bndrem;
      return *this; }

  //! @brief Set multivariate polynomial coefficients and remainder term equal to those in variable <tt>var</tt>, possibly defined in another Taylor model environment with less variables or with a different expansion order. Coefficients involving other variables or higher order are initialized to 0 if <tt>reset=true</tt> (default), otherwise they are left unmodified. Higher-order terms in TV are bounded and added to the remainder bound.
  virtual PolyVar<T>& set
    ( const PolyVar<T>& var, const bool reset=true );

  //! @brief Copy multivariate polynomial coefficients from current variable into variable <tt>var</tt>, possibly defined in another polynomial model environment with less variables or with a lower expansion order. Copied coefficients are reset to 0 in current Taylor variable if <tt>reset=true</tt>, otherwise they are left unmodified (default).
  virtual PolyVar<T>& get
    ( PolyVar<T>&var, const bool reset=false );

  //! @brief Get pointer to associated polynomial model environment
  virtual PolyModel* env() const
    { return _PM; }

  //! @brief Compute bound on variable using bounder <a>type</a>
  virtual T bound
    ( const int type ) const
    { return _polybound(type) + *_bndrem; }

  //! @brief Retreive bound on variable using default bounder
  T bound() const
    { return bndpol() + *_bndrem; }

  //! @brief Retreive bound on multivariate polynomial using default bounder
  T bndpol() const
    { if( !_bndpol ) _bndpol = new T( _polybound() );
      return *_bndpol; }

  //! @brief Retreive bound on all terms with (total) order <tt>iord</tt> in polynomial model
  T bndord
    ( const unsigned int iord )
    const
    { if( !iord || (_PM && iord<=nord()) )
      { if( !_bndord_uptd ) _update_bndord(); return _bndord[iord]; }
      return 0.; }
    //{ return ( !iord || (_PM && iord<=nord()) )? _bndord[iord]: 0.; }

  //! @brief Return remainder term of variable
  T remainder
    ()
    const
    { return( *_bndrem ); }

  //! @brief Shortcut to mc::PolyVar::bound
  virtual T B
    ( const int type )
    const
    { return bound( type ); }

  //! @brief Shortcut to mc::PolyVar::bound
  virtual T B
    ()
    const
    { return bound(); }

  //! @brief Shortcut to mc::PolyVar::remainder
  T R
    ()
    const
    { return remainder(); }

  //! @brief Get (possibly scaled) coefficient in monomial term with variable exponents as given in <a>iexp</a>
  double coefmon
    ( const unsigned int*iexp )
    const;

  //! @brief Get pair of size of, and const pointer to, array of (possibly scaled) monomial coefficients in multivariate polynomial of Taylor variable
  std::pair<unsigned int, const double*> coefmon
    ()
    const;

  //! @brief Get pair of size of, and const pointer to, array of monomial exponents in multivariate polynomial of Taylor variable
  std::pair<unsigned int, const unsigned int*> expmon
    ()
    const;

  //! @brief Overloaded operator '=' for polynomial model variables
  virtual PolyVar<T>& operator=
    ( const PolyVar<T>& );
  /** @} */
};

///////////////////////////////// PolyVar /////////////////////////////////////

template <typename T> inline PolyVar<T>&
PolyVar<T>::set
( const PolyVar<T>& var, const bool reset )
{
  if( !_PM || ( var._PM && nvar() < var.nvar() ) ) return *this;

  // Reset monomial coefficients to 0
  for( unsigned int imon=0; reset && imon<nmon(); imon++ ) _coefmon[imon] = 0.;
  if( reset ) *_bndrem = 0.;

  // Copy monomial coefficients from var into *this
  _coefmon[0] = var._coefmon[0];
  if( !var._PM ){ *_bndrem = *var._bndrem; _bndord_uptd = false; return *this; }
  unsigned int*iexp = new unsigned int[nvar()];
  for( unsigned int jmon=1; jmon<var.nmon() && jmon<var._posord(nord()+1); jmon++ ){
    for( unsigned int ivar=0; ivar<nvar(); ivar++ )
      iexp[ivar] = ( ivar<var.nvar()? var._expmon(jmon)[ivar]: 0 );
    _coefmon[_loc_expmon(iexp)] = var._coefmon[jmon];
  }
  delete[] iexp;
  for( unsigned int iord=nord()+1; iord<=var.nord(); iord++ )
    *_bndrem += var.bndord(iord);
  *_bndrem += *var._bndrem;
  _unset_bndpol();
  _bndord_uptd = false;
  return *this;
}

template <typename T> inline PolyVar<T>&
PolyVar<T>::get
( PolyVar<T>& var, const bool reset )
{
  if( !_PM ){
    var._coefmon[0] = _coefmon[0];
    // Reset monomial coefficients to 0
    if( reset ) _coefmon[0] = 0.;
    return *this;
  }
  if( !var._PM || nvar() < var.nvar() || nord() < var.nord() )
    return *this;

  // Copy monomial coefficients from *this into var
  unsigned int*iexp = new unsigned int[nvar()];
  for( unsigned int jmon=0; jmon<(var._PM?var.nmon():1); jmon++ ){
    for( unsigned int ivar=0; ivar<nvar(); ivar++ )
      iexp[ivar] = (ivar<var.nvar()? var._expmon(jmon)[ivar]: 0);
    var._coefmon[jmon] = _coefmon[_loc_expmon(iexp)];
    // Reset monomial coefficients to 0
    if( reset ) _coefmon[_loc_expmon(iexp)] = 0.;
  }
  delete[] iexp;
  *var._bndrem = 0.;
  var._unset_bndpol();
  var._bndord_uptd = false;

  if( reset ){
    _unset_bndpol();
    _bndord_uptd = false;
  }
  return *this;
}

template <typename T> inline void
PolyVar<T>::_size
( PolyModel* env )
{
  _PM = env;
  if( !_PM ){
    _coefmon = new double[1];
    _bndord  = new T[1];
    _bndrem  = _bndord;
  }
  else{
    _coefmon = new double[nmon()];
    _bndord  = new T[nord()+2];
    _bndrem  = _bndord + nord()+1;
  }
  _bndord_uptd = false;
  _bndpol = 0;
  return;
}

template <typename T> inline void
PolyVar<T>::_cleanup()
{
  delete [] _coefmon; delete [] _bndord; delete _bndpol;
  _coefmon = 0; _bndord = _bndrem = _bndpol = 0;
}

template <typename T> inline void
PolyVar<T>::_resize
( PolyModel* env )
{
  if( _PM == env ) return;
  _cleanup();
  _size( env );
}

template <typename T> inline PolyVar<T>&
PolyVar<T>::operator=
( const PolyVar<T>&var )
{
  return _set( var );
}

template <typename T> inline PolyVar<T>&
PolyVar<T>::_set
( const PolyVar<T>&var )
{
  // Same PolyVar?
  if( this == &var ) return *this;

  // Reinitialization needed?
  _resize( var._PM );

  // Set variable not linked to any environment
  if( !_PM ){
    _coefmon[0] = var._coefmon[0];
    *_bndrem    = *var._bndrem;
  }
  // Set variable linked to an environment
  else{
    for( unsigned int i=0; i<nmon(); i++ ) _coefmon[i] = var._coefmon[i];
    for( unsigned int i=0; i<nord()+1 && var._bndord_uptd; i++) _bndord[i] = var._bndord[i];
    *_bndrem = *var._bndrem;
  }
  _bndord_uptd = var._bndord_uptd; 
  // Set polynomial bound
  _set_bndpol( var._bndpol );

  return *this;
}

template <typename T> inline void
PolyVar<T>::_unset_bndpol
()
{
  delete _bndpol;
  _bndpol = 0;
}

template <typename T> inline void
PolyVar<T>::_set_bndpol
( const T*bndpol )
{
  if( !bndpol ){
    if( _bndpol ) delete _bndpol;
    _bndpol = 0;
  }
  else if( !_bndpol )
    _bndpol = new T( *bndpol );
  else
    *_bndpol = *bndpol;
}

template <typename T> inline void
PolyVar<T>::_set_bndpol
( const T&bndpol )
{
  if( !_bndpol )
    _bndpol = new T( bndpol );
  else
    *_bndpol = bndpol;
}

template <typename T> inline void
PolyVar<T>::_center()
{
  const double remmid = Op<T>::mid(*_bndrem);
  _coefmon[0] += remmid;
  if( _PM ) _bndord[0] = _coefmon[0];
  *_bndrem -= remmid;
  if( _bndpol ) *_bndpol += remmid;
}

template <typename T> inline double
PolyVar<T>::coefmon
( const unsigned int*iexp )
const
{
  if( !_PM ) return 0;
  const unsigned int imon = _PM->loc_expmon( iexp );
  return( imon<nmon()? _coefmon[imon]: 0. );
}

template <typename T> inline std::pair<unsigned int, const double*>
PolyVar<T>::coefmon()
const
{
  return std::make_pair( (_PM?nmon():1), _coefmon );
}

template <typename T> inline std::pair<unsigned int, const unsigned int*>
PolyVar<T>::expmon()
const
{
  return std::make_pair( (_PM?nmon()*nvar():1), _PM->expmon );
}

} // namespace mc

#endif
