// Copyright (C) 2009-2014 Benoit Chachuat, Imperial College London.
// All Rights Reserved.
// This code is published under the Eclipse Public License.

/*!
\page page_ELIMG Ellipsoidal Arithmetic for Factorable Functions
\author Mario E. Villanueva, Jai Rajyaguru, Boris Houska, Beno&icirc;t Chachuat

The image of an \f$n_x\f$-dimensional ellipsoid \f$\mathcal{E}_{x}:=\{ q_{x}+Q^{1/2}_{x}v \,\mid\, v\in\mathbb{R}^{n_x}, v^{\rm T}v\leq 1\}\f$ under the vector-valued function \f$ f:\mathbb{R}^{n_x}\to\mathbb{R}^{n_f} \f$ is defined as \f$ f(\mathcal{E}_{x}) := \{f(x) \,\mid\, x\in \mathcal{E}_{x} \} \f$. In general, such a set can not be computed exactly and thus we are interested on computing an ellipsoidal \f$ \mathcal{E}_{f}:= \{ q_{f} + Q^{1/2}_{f} \,\mid\, v\in\mathbb{R}^{n_x}, v^{\rm T}v\leq 1 \} \f$ enclosure such that \f$\mathcal{E}_{f} \supseteq f(\mathcal{E}_{x}) \f$.

The classes mc::EllImg and mc::EllVar provide an implementation that is able to compute such an enclosure. We note that this is not a verified implementation in the sense that rounding errors are not accounted for. 

The implementation of mc::EllImg and mc::EllVar relies on the operator/function overloading mechanism of C++. This makes the computation of the ellipsoidal enclosure for the image of an ellipsoid under a factorable function both simple and intuitive, similar to computing function values in real arithmetic or bounds based on interval, Taylor or Chebyshev model arithmetics (see \ref page_INTERVAL, \ref page_TAYLOR, \ref page_CHEBYSHEV). mc::EllImg stores a column vector CPPL::dcovector and a sparse symmetric matrix CPPL::dssmatrix provided by the LAPACK wrapper <A href="http://cpplapack.sourceforge.net/">CPPLAPACK</A>. The column vector stores the center and the sparse symmetric matrix the shape of a <i>lifted</i> ellipsoid, which is the result of adding extra dimension for each operation participating in the factorable function. 

The classes mc::EllImg and mc::EllVar are templated in the interval type used to bound the nonlinearity of the function, By default, mc::EllImg and mc::EllVar can be used with the non-verified interval type mc::Interval of MC++. For reliability, however, it is strongly recommended to use verified interval arithmetic such as <A href="http://www.ti3.tu-harburg.de/Software/PROFILEnglisch.html">PROFIL</A> (header file <tt>mcprofil.hpp</tt>) or <A href="http://www.math.uni-wuppertal.de/~xsc/software/filib.html">FILIB++</A> (header file <tt>mcfilib.hpp</tt>). 

\section sec_ELIMG_1 How do I compute an ellipsoidal enclosure for the image set of an ellipsoid under a factorable function?

Suppose that we want to compute an ellipsoidal enclosure for the image set of function:
\f[
f(x) = \left(\begin{array}{c}
\log(x_{1})+x^{2}_{2}        \\
\sin(x_{1})-\cos(x_{2})       \\
\end{array} \right) \qquad \text{with} \qquad 
x \in \left\{
\left(\begin{array}{c}
3 \\
4 \\
\end{array} \right)+
\left(\begin{array}{cc}
5 & 4 \\
4 & 5 \\
\end{array} \right)^{1/2}v
:
v^{\rm T}v\leq 1 \right\}
\f].

For simplicity bounds are computed using the default interval type mc::Interval, the required header files 
are:
 
\code
#include "ellimage.hpp"
#include "ellipsoid_cpplapack.hpp"
#include "interval.hpp"

typedef mc::Interval I;
typedef mc::EllImg<I> EI;
typedef mc::EllVar<I> EV;
typedef mc::Ellipsoid E;
typedef CPPL::dcovector dcv;
typedef CPPL::dsymatrix dsm;
\endcode

First, the host ellipsoid for the independent variables (\f$x_{1}\f$ and \f$x_{2}\f$) is specified as:
\code 
dcv q0(2);      dsm Q0(2);
q0(0) = 3.;     Q0(0,0) = 5.;
q0(1) = 4.;     Q0(1,0) = 4.; Q0(1,1) = 5. ;

E Ex( Q0, q0 );
\endcode

Next the ellipsoidal enclosure \f$\mathcal E_{f}\f$ and the independent variables \f$x_{1}\f$ and \f$x_{2}\f$ are defined by mc::EllImg and mc::EllVar objects respectively:
\code 
EI Eflift(2)	;
EV X[2]			;
\endcode

Notice that the ellipsoidal enclosure is defined based on the number of independent variables and the independent variables is given as an array. This array has to be introduced to the ellipsoidal enclosure as:

\code
for( long i = 0 ; i < 2 ; ++i ) X[i].set(&Eflift , i , Ex )	;
\endcode

The dependent f_{1}(x) and f_{2}(x) are defined as:

\code
EV* F[2] = { log(X[0]) + mc::sqr(X[1]) , sin(X[0]) - cos(X[1]) }	;
\endcode

The lifted ellipsoid can be displayed to the standard output as:

\code
std::cout << "lifted ellipsoidal image of f: " << Eflift << std::endl;
\endcode

Finally, the ellipsoidal enclosure is obtained by projecting the dependent variables in the lifted ellipsoidal image, and is given as an mc::Ellipsoid object :

\code
E Ef = Eflift.project( F , 2 );
std::cout << " Ellipsoidal enclosure Ef = " << Ef << std::endl;
\endcode

After this, the ellipsoid can be manipulated according to the rules of ellipsoidal calculus (see \ref page_ELLIPSOID).
A comparison between the exact image and the ellipsoidal enclosure is presented in the following figure, in orange the exact image and in blue the boundary of the ellipsoidal enclosure.
<CENTER><TABLE BORDER=0>
<TR>
<TD>\image html ELIMG1.png</TD>
</TR>
</TABLE></CENTER>


\section sec_ELIMG_opt How are the options set for the computation of an Ellipsoidal enclosure?

The class mc::EllImg has a public member called mc::EllImg::options that can be used to set/modify a number of options; e.g.,

\code
	EIm.options.PREALLOCATE   = 5 ; 
	EIm.options.CHEBYSHEV_USE = false  ;
\endcode

The available options are the following:

<TABLE border="1">
<CAPTION><EM>Options in mc::EllImg::Options: name, type and description</EM></CAPTION>
     <TR><TH><b>Name</b>  <TD><b>Type</b><TD><b>Default</b>
         <TD><b>Description</b>
     <TR><TH><tt>PREALLOCATE</tt> <TD><tt>unsigned long</tt> <TD>0
         <TD> Number of rows to preallocate in the shape matrix and center vector
     <TR><TH><tt>CHEBYSHEV_USE</tt> <TD><tt>bool</tt> <TD>false
         <TD> Whether to use Chebyshev expansion to compute a linear approximation and bound the nonlinear dependencies of univariate terms
     <TR><TH><tt>CHEBYSHEV_ORDER</tt> <TD><tt>unsigned int</tt> <TD>5
         <TD>Order of the Chebyshev expansion (only of <tt>CHEBYSHEV_USE = true</tt>)
</TABLE>


\section sec_ELLIMAGE_err Errors What errors can I encounter during computation of a Chebyshev model?

Errors are managed based on the exception handling mechanism of the C++ language. Each time an error is encountered, a class object of type mc::EllImg::Exceptions is thrown, which contains the type of error. It is the user's responsibility to test whether an exception was thrown during the computation of the lifted ellipsoid, and then make the appropriate changes. Should an exception be thrown and not caught by the calling program, the execution will abort.

Possible errors encountered during the computation of a Chebyshev model are:

<TABLE border="1">
<CAPTION><EM>Errors during the Computation of a Chebyshev Model</EM></CAPTION>
     <TR><TH><b>Number</b> <TD><b>Description</b>
     <TR><TH> <tt> 1 </tt>  <TD> Division by zero
     <TR><TH> <tt> 2 </tt>  <TD> Inverse operation with zero in domain
     <TR><TH> <tt> 3 </tt>  <TD> Log operation with non-positive numbers in domain
     <TR><TH> <tt> 4 </tt>  <TD> Square-root operation with negative numbers in domain
     <TR><TH> <tt> 5 </tt>  <TD> Tangent operation with zero in domain of cosine, tan(x) = sin(x)/cos(x)
     <TR><TH> <tt> 6 </tt>  <TD> Sine/Cosine inverse operation with domain outside [-1,1]
     <TR><TH> <tt>-1 </tt>  <TD> Failed to construct Ellipsoidal variable
     <TR><TH> <tt>-3 </tt>  <TD> Operation between Chebyshev variables linked to different Chebyshev models
     <TR><TH> <tt>-33</tt>  <TD> Feature not yet implemented in mc::EllImg
</TABLE>

Moreover, exceptions may be thrown by the template parameter class itself.

*/

#ifndef MC__ELLIMAGE_H
#define MC__ELLIMAGE_H

#include <assert.h>
#include <exception>

#include "mcfunc.hpp"
#include "mcop.hpp"
#include "mclapack.hpp"
#include "ellipsoid.hpp"
#include "cmodel.hpp"

#include <fstream>
#include <iomanip>

#undef MC__ELLIMAGE_DEBUG

namespace mc
{

template< class T > class EllVar;


//! @brief C++ class for 
////////////////////////////////////////////////////////////////////////
//!
////////////////////////////////////////////////////////////////////////
template< class T >
class EllImg
////////////////////////////////////////////////////////////////////////
{

  friend class EllVar<T>;
  
  template< class U > friend  EllVar<U> operator+( const EllVar<U>& );
  template< class U > friend  EllVar<U> operator+( const EllVar<U>&, const EllVar<U>& );
  template< class U > friend  EllVar<U> operator+( const EllVar<U>&, const double );
  template< class U > friend  EllVar<U> operator+( const double, const EllVar<U>& );
  template< class U > friend  EllVar<U> operator-( const EllVar<U>& );
  template< class U > friend  EllVar<U> operator-( const EllVar<U>&, const EllVar<U>& );
  template< class U > friend  EllVar<U> operator-( const EllVar<U>&, const double );
  template< class U > friend  EllVar<U> operator-( const double, const EllVar<U>& );
  template< class U > friend  EllVar<U> operator*( const EllVar<U>&, const EllVar<U>& );
  template< class U > friend  EllVar<U> operator*( const EllVar<U>&, const double );			
  template< class U > friend  EllVar<U> operator*( const double, const EllVar<U>& );
  template< class U > friend  EllVar<U> operator/( const EllVar<U>&, const EllVar<U>& );
  template< class U > friend  EllVar<U> operator/( const EllVar<U>&, const double );			
  template< class U > friend  EllVar<U> operator/( const double, const EllVar<U>& );
  template <class U > friend  std::ostream& operator<< (std::ostream&, const EllImg<U>&);

  template< class U > friend  EllVar<U> inv ( const EllVar<U>& );
  template< class U > friend  EllVar<U> exp ( const EllVar<U>& );
  template< class U > friend  EllVar<U> log ( const EllVar<U>& );
  template< class U > friend  EllVar<U> sqrt( const EllVar<U>& );
  template< class U > friend  EllVar<U> sqr ( const EllVar<U>& );
  template< class U > friend  EllVar<U> pow ( const EllVar<U>&, const int );  
  template< class U > friend  EllVar<U> cos ( const EllVar<U>& );
  template< class U > friend  EllVar<U> sin ( const EllVar<U>& );
  template< class U > friend  EllVar<U> tan ( const EllVar<U>& );
  template< class U > friend  EllVar<U> acos( const EllVar<U>& );
  template< class U > friend  EllVar<U> asin( const EllVar<U>& );
  template< class U > friend  EllVar<U> atan( const EllVar<U>& );
  
  public:
    //! @brief Exceptions of mc::EllImg
    class Exceptions{
      public:
        //! @brief Enumeration type for CModel exception handling
        enum TYPE{
          DIV   = 1,	//!< Division by zero scalar
          INV,		//!< Inverse operation with zero in domain
          LOG,		//!< Log operation with non-positive numbers in domain
          SQRT,		//!< Square-root operation with negative numbers in domain
          TAN,		//!< Tangent operation with zero in domain of cosine, tan(x) = sin(x)/cos(x)
          ACOS,		//!< Sine/Cosine inverse operation with domain outside [-1,1]
          INIT  = -1,	//!< Failed to construct Ellipsoidal variable
          EIMG  = -3,	//!< Operation between Chebyshev variables linked to different Chebyshev models
          UNDEF = -33  	//!< Feature not yet implemented in mc::EllImg
        };
        //! @brief Constructor for error <a>ierr</a>
        Exceptions( TYPE ierr ) : _ierr( ierr ){}
        //! @brief Error flag
        int ierr(){ return _ierr; }
        //! @brief Error description
        std::string what(){
          switch( _ierr ){
            case DIV   : return "mc::EllImg\t Division by zero scalar";
            case INV   : return "mc::EllImg\t Inverse operation with zero in domain";
            case LOG   : return "mc::EllImg\t Log operation with non-positive numbers in domain";
            case SQRT  : return "mc::EllImg\t Square-root operation with negative numbers in domain";
            case TAN   : return "mc::EllImg\t Tangent operation with zero in domain of cosine";
            case ACOS  : return "mc::EllImg\t Inverse sine/cosine operation with domain outside [-1,1]";
            case EIMG  : return "mc::EllImg\t EllVars belong to different ellipsoids, operation not allowed";
            case UNDEF : return "mc::EllImg\t Feature not yet implemented in mc::EllImg class";
            default    : return "mc::EllImg\t Undocumented error";
          }		
        }

      private:
        //! @brief Type of error 
        TYPE _ierr;
    };

    //! @brief Structure containing the options for EllImg
    struct Options{
      //! @brief Constructor of mc::EllImg<>::Options
      Options():
        PREALLOCATE     ( 0 ),
        CHEBYSHEV_USE   ( false ),
        CHEBYSHEV_ORDER ( 5 )  	
	{}
      //! @brief Sets number of rows to preallocate in the shape matrix and center vector
      long PREALLOCATE;
      //! @brief Whether to use Chebyshev models to obtain the linearisation of univariate functions  
      bool CHEBYSHEV_USE;
      //! @brief Order of the Chebyshev model used for the linearisation of univariate functions
      int CHEBYSHEV_ORDER; //( IF CHEBYSHEV_USE is false, CHEBYSHEV_ORDER is ignored )  	 
    } options;
  	
  private:
    //! @brief Shape Matrix of the lifted Ellipsoid
    CPPL::dssmatrix _Q;
    //! @brief Centre of the lifted Ellipsoid
    CPPL::dcovector _q;
    //! @brief Map between pointers to EllVars (key) and row number
    long _current_row;
    //! @brief vector of ranges
    std::vector<T> _Irange;
    //! @brief pointer to internal Chebyshev model;
    CModel<T>* _CMpt;
    //! @brief pointer to internal Chebyshev variable;
    CVar<T>*   _CVpt;
    //! @brief variable that is true if a dependency map is available;
    bool _depmap_flag;
    //! @brief index of row where product starts;
    long _ip;
    //! @brief dependecy map
    CPPL::dssmatrix _depmap
    
#ifdef MC__ELLIMAGE_DEBUG
    //! @brief Output for debugging
    std::ofstream _dbugout;
#endif
	
  public:
    //! @brief Returns the Shape Matrix of the lifted Ellipsoid
    CPPL::dssmatrix Q()
      {	return _Q; }
    //! @brief Returns the Shape Matrix of the lifted Ellipsoid
    CPPL::dcovector q()
      { return _q; }
    //! @brief Returns the vector of ranges 
    std::vector<T> Irange()
      { return _Irange; }

    //! @brief Constructor that preallocates space for independent variables
    EllImg( long=0 );
    //! @brief Constructor that preallocates space for the full-dimensional ellipsoid
    EllImg( const CPPL::dssmatrix&  );
    //! @brief Destructor
    virtual ~EllImg();
    
    //! @brief Sets an Ellipsoidal Variable in a lifted ellipsoid
    EllVar<T>* set( EllVar<T>* EllVarpt, const CPPL::dsymatrix& Q0, const CPPL::dcovector& q0  )
      { return _set( EllVarpt, Q0, q0 ); }

    //! @brief projects an (array of length) EllVar(s) 
    Ellipsoid get( EllVar<T>* EllVarpt, long length )
      { return _get( EllVarpt, length ); }

  private:
    //! @brief Stores the range of an EllVar in the current EllImg
    void _setbound( const EllVar<T>& );
	
    //! @brief Stores the range of an EllVar in the current EllImg
    void _setbound( const EllVar<T>&, const T&);
	
    //! @brief Returns the range of an EllVar in the current EllImg
    T _bound( const EllVar<T>& ) const;
	
    //! @brief Computes the Minkowski sum of the ellipsoid and an interval
    void _minksum( const EllVar<T>&, double ); 
	
    //! @brief Computes the Minkowski sum of the ellipsoid and an interval
    void _minksum( long, double ); 
    
    //! @brief Computes the Trace of the Shape matrix for the lifted ellipsoidal image
    double _trQ();
	
	//! @brief Sets an Ellipsoidal Variable in a lifted ellipsoid
    EllVar<T>* _set( EllVar<T>*, const CPPL::dsymatrix&, const CPPL::dcovector& );
	
    //! @brief projects a given EllVar<>
    Ellipsoid _get( EllVar<T>*, long );
    
    //! @brief reset the EllImg environment
    void _reset();
	
    //! @brief univariate update
    void _univupdate( long, long, double, double, double );
	
    //! @brief univariate update used in product
    void _univupdate( long, long, CPPL::drovector, double, double, double );
	
    //! @brief Prototype for univariate function for CModel evaluation
    typedef CVar<T> (univariate_function) ( const CVar<T>&  ) ;
	
    //! @brief typedef for triplet of linearisation parameters (constant, linear, radius of remainder )
    typedef std::tuple<double, double, double> lin_param;
	
    //! @brief linearise using Chebyshev models
    lin_param _cmodel_linear( univariate_function, const T&, const int );

    //! @brief linearise using Chebyshev models (for power function)
    lin_param _linearize_sqr( const T& );    
};

//! @brief C++ class for 
////////////////////////////////////////////////////////////////////////
//!
////////////////////////////////////////////////////////////////////////
template< class T >
class EllVar
////////////////////////////////////////////////////////////////////////
{
  friend class EllImg<T>;
  
  template< class U > friend  EllVar<U> operator+( const EllVar<U>& );
  template< class U > friend  EllVar<U> operator+( const EllVar<U>&, const EllVar<U>& );
  template< class U > friend  EllVar<U> operator+( const EllVar<U>&, const double );
  template< class U > friend  EllVar<U> operator+( const double, const EllVar<U>& );
  template< class U > friend  EllVar<U> operator-( const EllVar<U>& );
  template< class U > friend  EllVar<U> operator-( const EllVar<U>&, const EllVar<U>& );
  template< class U > friend  EllVar<U> operator-( const EllVar<U>&, const double );
  template< class U > friend  EllVar<U> operator-( const double, const EllVar<U>& );
  template< class U > friend  EllVar<U> operator*( const EllVar<U>&, const EllVar<U>& );
  template< class U > friend  EllVar<U> operator*( const EllVar<U>&, const double );			
  template< class U > friend  EllVar<U> operator*( const double, const EllVar<U>& );
  template< class U > friend  EllVar<U> operator/( const EllVar<U>&, const EllVar<U>& );
  template< class U > friend  EllVar<U> operator/( const EllVar<U>&, const double );			
  template< class U > friend  EllVar<U> operator/( const double, const EllVar<U>& );
    
  template< class U > friend  EllVar<U> inv ( const EllVar<U>& );
  template< class U > friend  EllVar<U> exp ( const EllVar<U>& );
  template< class U > friend  EllVar<U> log ( const EllVar<U>& );
  template< class U > friend  EllVar<U> sqrt( const EllVar<U>& );
  template< class U > friend  EllVar<U> sqr ( const EllVar<U>& );
  template< class U > friend  EllVar<U> pow ( const EllVar<U>&, const int );  
  template< class U > friend  EllVar<U> cos ( const EllVar<U>& );
  template< class U > friend  EllVar<U> sin ( const EllVar<U>& );
  template< class U > friend  EllVar<U> tan ( const EllVar<U>& );
  template< class U > friend  EllVar<U> acos( const EllVar<U>& );
  template< class U > friend  EllVar<U> asin( const EllVar<U>& );
  template< class U > friend  EllVar<U> atan( const EllVar<U>& );
							  
  private:
    //! @brief pointer to underlying lifted ellipsoid
    EllImg<T>* _EI;
    //! @brief row index in ellipsoid
    long _RowInd;
    //! @brief flag for constant/interval
	bool _is_constant;
	//! @brief constant/interval value given as mid/rad
	std::pair<double,double> _mid_rad;  
		
  public:
    //! @brief Default constructor 
    EllVar();
    //! @brief Copy constructor
    EllVar( const EllVar<T>& );
    //! @brief Constructor for constants
    EllVar( const double d );
    //! @brief Constructor for intervals
    EllVar( const T& );
    
    //! @brief Destructor
    virtual ~EllVar(){};
    
  private:
    //! @brief constructor for internal variables
    EllVar( EllImg<T>*, long ); 
    //! @brief constructor for internal variables used in the product
    EllVar( EllImg<T>*, long, long );		

  public:
  // Public overloads
      						  EllVar<T>& operator= ( const EllVar<T>& );
  							  EllVar<T>& operator= ( const double );
							  EllVar<T>& operator= ( const T& );
 
};
 
///////////////////////////////// EllImg ///////////////////////////////////

template <class T> 
inline 
EllImg<T>::EllImg
( long dim ) :
  options      ( )       , 
  _Q           ( dim )   , 
  _q           ( dim )   ,
  _current_row ( 0 )     ,
  _Irange      (   )     ,
  _CMpt        ( NULL )  ,
  _CVpt        ( NULL )  ,
  _depmap_flag ( false ) ,
  _ip          ( -1 )
{
#ifdef MC__ELLIMAGE_DEBUG
  _dbugout.open( "debug.log" , std::ios_base::out )
#endif     
  _Q.zero();
  _q.zero();
}

template <class T> 
inline 
EllImg<T>::EllImg
( const CPPL::dssmatrix& depmap ) :
  options      ( )          , 
  _Q           ( depmap )   , 
  _q           ( depmap.n ) ,
  _current_row ( 0 )        ,
  _Irange      (  )         ,
  _CMpt        ( NULL )     ,
  _CVpt        ( NULL )     ,
  _depmap_flag ( true )     ,
  _ip          ( -1 )
{
#ifdef MC__ELLIMAGE_DEBUG
  _dbugout.open( "debug.log" , std::ios_base::out )
#endif     
  _q.zero();
}

template <class T> 
inline 
EllImg<T>::~EllImg
( )
{
#ifdef MC__ELLIMAGE_DEBUG
	_dbugout.close();
#endif     
	if( _CMpt != NULL )
	{
		delete _CMpt;
	}
}

template <class T>
inline
void 
EllImg<T>::_setbound
( const EllVar<T>& EllVar1 )
{
  long i   = EllVar1._RowInd;
  T ebound = _q(i) + std::sqrt(_Q(i,i))*T(-1.,1.);
  _Irange.push_back( ebound );
}

template <class T>
inline
void 
EllImg<T>::_setbound
( const EllVar<T>& EllVar1, const T& Ibound  )
{
  long i   = EllVar1._RowInd;
  T bound;
  T ebound = _q( i ) + std::sqrt( _Q( i,i )  )*T( -1.,1. );
  Op<T>::inter( bound, ebound, Ibound ); 
#ifdef MC__ELLIMAGE_DEBUG
  _dbugout << " Range Vector Before i= " << i << " length= " << _Irange.size() << std::endl;
  for( auto it = _Irange.begin(); it != _Irange.end(); ++it ) _dbugout << *it << std::endl;
#endif
  if( i >= (long)_Irange.size()  ) 
  {
  	_Irange.push_back( bound );
#ifdef MC__ELLIMAGE_DEBUG
  	_dbugout << " Range Vector After i= " << i << " length= " << _Irange.size() << std::endl;
  	for( auto it = _Irange.begin(); it != _Irange.end(); ++it ) _dbugout << *it << std::endl;
#endif
  	return;
  }
  else
  { 
  	_Irange[i] = bound;
#ifdef MC__ELLIMAGE_DEBUG
  	_dbugout << " Range Vector After i= " << i <<" length= " << _Irange.size() << std::endl;
  	for( auto it = _Irange.begin(); it != _Irange.end(); ++it ) _dbugout << *it << std::endl;
#endif
  	return;
  }
}

template <class T>
inline
T 
EllImg<T>::_bound
( const EllVar<T>& EllVar1 )
const
{
  long i  = EllVar1._RowInd;
  return _Irange[ i ]; 
}

template <class T>
inline
void 
EllImg<T>::_minksum
( const EllVar<T>& EllVar1, double rad )
{
  long i     = EllVar1._RowInd;
  _minksum( i , rad );	
}

template <class T>
inline
void 
EllImg<T>::_minksum
( long i, double rad  )
{
  double TOL = 1e-8, EPS = machprec(), strQ = 0.0; 

  for( long j=0; j<= i; j++ ) strQ += _Q(j,j)/(_Q(j,j)+TOL);
  strQ = std::sqrt(strQ);

  double sqrR  =  rad/std::sqrt(_Q(i,i)+TOL);
  double kappa = strQ + sqrR + EPS;
  _Q          *= (kappa/(strQ+EPS));  
  _Q(i,i)     += (std::pow(rad,2)*kappa)/(sqrR+EPS);
  _Q.diet();
}  

template <class T> 
inline
double 
EllImg<T>::_trQ
()
{
  if( !_Q.n ) return 0.;
  double tr(_Q(0,0));
  for( unsigned int i=1; i<_Q.n; i++ ) tr += _Q(i,i);
  return tr;
}


template <class T> 
inline 
EllVar<T>*
EllImg<T>::_set
( EllVar<T>* EllVarpt       ,
  const CPPL::dsymatrix& Q0 ,
  const CPPL::dcovector& q0   )
{
  assert( Q0.n == q0.l );
  assert( _Q.n >= Q0.n );
  
  // Set the underlying lifted ellipsoid
  for ( long j= 0; j < Q0.n; ++j  )
  {
	  EllVarpt[j]._EI     =  this;
	  EllVarpt[j]._RowInd = j    ;
	  // Initialise the varindex-th row of _Q with the varindex-th row of the shape matrix of Ell ; 
	  for (long i= 0; i <= j; ++i )
	  {
	  	_Q.put( i, j, Q0(i,j) ) ;
	  }
	  // Initialise the varindex-th row of _q with the varindex-th row of the centre vector of Ell ;
	  _q( j ) = q0(j);
	  // Initialise the range of the EllVar
	  _setbound( EllVarpt[j] );
	  // Increment the current row count
	  _current_row++ ;
  } 
  return EllVarpt;
}

template <class T> 
inline 
Ellipsoid
EllImg<T>::_get
( EllVar<T>* EllVarpt , 
  long  length      )
{
  CPPL::dsymatrix Q( length ); Q.zero();
  CPPL::dcovector q( length ); q.zero();
  for( long j = 0 ; j < length ; ++j  )
  {
  	long prev = EllVarpt[j]._RowInd; 
  	for ( long i = j ; i < length ; ++i )
  	{
   		Q( i , j ) = _Q( EllVarpt[i]._RowInd , prev  ) ; 
  		q( j )     = _q( EllVarpt[j]._RowInd ) ;
  	} 
  }
  return Ellipsoid( Q , q );
}

template <class T> 
inline 
void
EllImg<T>::_univupdate
( long i      ,
  long k      , 
  double c0   ,
  double c1   ,
  double eta    )
{
  // Update centre
  _q( i ) = c0;
  // Update matrix when the shape matrix has been initialised with a dependency map
  if( _depmap_flag )
  {
 	  // Update shape matrix
	  for(long j = 0; j < i ; ++j) if( _Q.isListed( i, j ) ) _Q(i, j) = c1 * _Q(k,j) ;
	  // Update diagonal element
	  _Q( i, i ) = std::pow(c1,2) * _Q(k,k) ; 
  }
  else // No dependecy map available
  {
	  // Update shape matrix
	  for(long j = 0; j < i ; ++j) _Q.put(i, j, c1 * _Q(k,j) );
	  // Update diagonal element
	  _Q.put( i, i, std::pow(c1,2) * _Q(k,k) );
  }
  // Minkowski sum with an interval centered at 0 with radius eta
  _minksum( i , eta );
}

template <class T> 
inline 
void
EllImg<T>::_univupdate
( long i             ,
  long k             , 
  CPPL::drovector rQ , 
  double c0          ,
  double c1          ,
  double eta           )
{
  // Update centre
  _q( i ) = c0;
  // Update matrix when the shape matrix has been initialised with a dependency map
  if( _depmap_flag )
  {
 	  // Update shape matrix
	  for(long j = 0; j < i ; ++j) if( _Q.isListed( i, j ) ) _Q(i, j) = c1 * _Q(k,j) ;
	  // Update diagonal element
	  _Q( i, i ) = std::pow(c1,2) * _Q(k,k) ; 
  }
  else // No dependecy map available
  {
	  // Update shape matrix
	  for(long j = 0; j < i ; ++j) _Q.put(i, j, c1 * _Q(k,j) );
	  // Update diagonal element
	  _Q.put( i, i, std::pow(c1,2) * _Q(k,k) );
  }
  // Minkowski sum with an interval centered at 0 with radius eta
  _minksum( i , eta );
}

//! @brief linearise using Chebyshev models
template <class T> 
inline 
typename EllImg<T>::lin_param
EllImg<T>::_linearize_sqr
( const T& domain )
{
  
  if( options.CHEBYSHEV_USE )
  {
  	return _cmodel_linear( sqr , domain );
  }
  else // Compute linear approximation of sqr( m + r*x ) with x \in [-1,1] using minimax linearisation
  {
	double c0, c1, eta;
	double m = Op<T>::mid( domain );
	double r = 0.5 * Op<T>::diam( domain );
	double c_sec=0., x_tan=0., c_tan=0.;
	c_sec = std::pow(m,2)+std::pow(r,2);
	c1   = 2.0*m*r;
	x_tan = (c1-2.0*m*r)/(2.0*std::pow(r,2));
	c_tan = std::pow((m+r*x_tan),2)-c1*x_tan;
	eta   = (c_sec - c_tan)/2.0;
	c0    = c_sec - eta;
    c1    =  c1 / r;
    return std::make_tuple( c0, c1, eta );
  }

}	

//! @brief linearise using Chebyshev models
template <class T> 
inline 
typename EllImg<T>::lin_param
EllImg<T>::_cmodel_linear
( univariate_function f, 
  const T& domain      ,
  const int n = 0        )
{	
  // initialise Chevyshev model if it hasnt been used
  if( _CMpt == NULL )
  {
  	_CMpt = new CModel<T>( 1 , options.CHEBYSHEV_ORDER ) ;
  }
  CVar<T> CVX( _CMpt, 0 , domain );
  CVar<T> CVF;
  if( n == 0 ) CVF = f( CVX );
  else         CVF = pow( CVX , n) ;
  double c0   = CVF.constant();
  double c1   = CVF.linear( 0 , true );
  CVF.C(); 
  double eta  = 0.5 * Op<T>::diam(CVF.B());
  return std::make_tuple( c0, c1, eta );
} 

/////////////////////////////////  EllVar  ///////////////////////////////////

template <class T> 
inline 
EllVar<T>::EllVar
( ) :
  _EI         ( NULL )                  ,
  _RowInd     ( -1 )                    ,
  _is_constant( false )                 ,
  _mid_rad    ( std::make_pair(0.,0.) )
{}

template <class T> 
inline  
EllVar<T>::EllVar
( EllImg<T>* EI , 
  long RowInd       ) :
  _EI         ( EI )                    ,
  _RowInd     ( RowInd )                ,
  _is_constant( false )                 ,
  _mid_rad    ( std::make_pair(0.,0.) )
{ 
  if( !_EI ) throw typename EllImg<T>::Exceptions( EllImg<T>::Exceptions::INIT );
  // The ellipsoid is lifted if it has not been initialised with a dependecy map
  if( ! EI->_depmap_flag ) 
  {
	  // Checks if preallocation is needed
	  if( EI->options.PREALLOCATE && RowInd >= EI->_Q.n   )
	  {
	  	long dim = EI->options.PREALLOCATE;
	  	EI->_Q.stretch( dim );
	  	EI->_q.stretch( dim );
	  }
	  else if( !EI->options.PREALLOCATE && RowInd >= EI->_Q.n )
	  {
	  	EI->_Q.stretch( 1 );
	  	EI->_q.stretch( 1 );
	  }
  }
  EI->_current_row++;
}

template <class T> 
inline  
EllVar<T>::EllVar
( EllImg<T>* EI , 
  long RowInd     ,
  long current_row  ) :
  _EI         ( EI )                    ,
  _RowInd     ( RowInd )                ,
  _is_constant( false )                 ,
  _mid_rad    ( std::make_pair(0.,0.) )
{ 
  if( !_EI ) throw typename EllImg<T>::Exceptions( EllImg<T>::Exceptions::INIT );
  // The ellipsoid is lifted if it has not been initialised with a dependecy map
  if( ! EI->_depmap_flag ) 
  {
	  // Checks if preallocation is needed
	  if( EI->options.PREALLOCATE && RowInd >= EI->_Q.n   )
	  {
	  	long dim = EI->options.PREALLOCATE;
	  	EI->_Q.stretch( dim );
	  	EI->_q.stretch( dim );
	  }
	  else if( !EI->options.PREALLOCATE && RowInd >= EI->_Q.n )
	  {
	  	EI->_Q.stretch( 1 );
	  	EI->_q.stretch( 1 );
	  }
  }
  EI->_current_row = current_row;
}


template <class T> 
inline 
EllVar<T>::EllVar
( const EllVar<T>& EllVar1 ) :
  _EI         ( EllVar1._EI )           ,
  _RowInd     ( EllVar1._RowInd )       ,
  _is_constant( EllVar1._is_constant )  ,
  _mid_rad    ( EllVar1._mid_rad )
{ }  

template <class T> 
inline 
EllVar<T>::EllVar
( const double d ) :
  _EI         ( NULL )                  ,
  _RowInd     ( -1 )                    ,
  _is_constant( true )                  ,
  _mid_rad    ( std::make_pair(d,0.) )
{ }  

template <class T> 
inline 
EllVar<T>::EllVar
( const T& B ) :
  _EI         ( NULL )                  ,
  _RowInd     ( -1 )                    ,
  _is_constant( true )                  ,
  _mid_rad    ( std::make_pair(Op<T>::mid(B),0.5*Op<T>::diam(B)) )
{ } 

template <class T> 
inline 
EllVar<T>&
EllVar<T>::operator=
( const EllVar<T>& EllVar1 ) 
{
  _EI           = EllVar1._EI;
  _RowInd       = EllVar1._RowInd;
  _is_constant  = EllVar1._is_constant;
  _mid_rad      = EllVar1._mid_rad;
  return *this; 
}

template <class T> 
inline 
EllVar<T>&
EllVar<T>::operator=
( const double d ) 
{
  _EI           = NULL;
  _RowInd       = -1  ;
  _is_constant  = true;
  _mid_rad      = std::make_pair( d, 0. );
  return *this; 
}

template <class T> 
inline 
EllVar<T>&
EllVar<T>::operator=
( const T& B ) 
{
  _EI           = NULL;
  _RowInd       = -1  ;
  _is_constant  = true;
  _mid_rad      = std::make_pair( Op<T>::mid(B), 0.5*Op<T>::diam(B) );
  return *this; 
}



/////////////////////////////////  Operators  ///////////////////////////////////

template <class T> 
inline 
EllVar<T> 
operator+
( const EllVar<T>& EllVar1 )
{
  EllImg<T>* EI = EllVar1._EI;  
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // Update Centre
  EI->_q(i)  =  EI->_q(k) ;
  // Update matrix when the shape matrix has been initialised with a dependency map
  if( EI-> _depmap_flag )
  {
  	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j )
	  { 
	  	if( EI->_Q.isListed(i,j) ) EI->_Q(i,j) = EI->_Q(k,j) ;
  	  }
	  // Update Diagonal element
	  EI->_Q(i,i) = EI->_Q(k,k) ;
  }
  else // No dependency map available 
  {
	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j ) 
	  {
	  	EI->_Q.put( i,j, EI->_Q(k,j) );
	  }
	  // Update Diagonal element
	  EI->_Q.put( i,i, EI->_Q(k,k) );
  }
  // Save range of Variable 
  EI->_setbound( EllVar3 );
  	
  return EllVar3;
}

template <class T> 
inline 
EllVar<T>  
operator+
( const EllVar<T>& EllVar1 , 
  const EllVar<T>& EllVar2  )
{
  if( EllVar1._EI != EllVar2._EI )
  {	
   	throw typename EllImg<T>::Exceptions( EllImg<T>::Exceptions::EIMG );
  }
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  long l          = EllVar2._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // Update Centre
  EI->_q( i )  =  EI->_q(k) + EI->_q(l);
  // Update matrix when the shape matrix has been initialised with a dependency map
  if( EI-> _depmap_flag )
  {
  	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j )
	  {
  		if( EI->_Q.isListed(i,j) ) EI->_Q(i,j) = EI->_Q(k,j) + EI->_Q(l,j) ;
  	  }
	  // Update Diagonal element
	  EI->_Q(i,i) = EI->_Q(k,k) + 2.*( EI->_Q(l,k) ) + EI->_Q(l,l) ;
  }
  else // No dependency map available 
  {
	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j ) 
	  {
		EI->_Q.put( i,j, EI->_Q(k,j) + EI->_Q(l,j)  );
	  }
	  // Update Diagonal element
	  EI->_Q.put( i,i, EI->_Q(k,k) + 2.*(EI->_Q(l,k)) + EI->_Q(l,l) );
  }
  // Save range of Variable 
  EI->_setbound( EllVar3 );
  	
  return EllVar3;
 }

template <class T> 
inline 
EllVar<T>  
operator+
( const EllVar<T>& EllVar1 , 
  const double  DVal )
{
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // Update Centre
  EI->_q( i )  =  EI->_q(k) + DVal;
  // Update matrix when the shape matrix has been initialised with a dependency map
  if( EI-> _depmap_flag )
  {
  	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j )
	  {
  		if( EI->_Q.isListed(i,j) ) EI->_Q(i,j) = EI->_Q(k,j) ;
  	  }
	  // Update Diagonal element
	  EI->_Q(i,i) = EI->_Q(k,k) ;
  }
  else // No dependency map available 
  {
	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j ) 
	  {
			EI->_Q.put(i,j, EI->_Q(k,j) );
	  }
	  // Update Diagonal element
	  EI->_Q.put( i,i, EI->_Q(k,k) );
  }
  
  // Save range of Variable 
  EI->_setbound( EllVar3 );
	
  return EllVar3;
 }

template <class T> 
inline 
EllVar<T>  
operator+
( const double  DVal, 
  const EllVar<T>& EllVar1 )
{
  return EllVar1 + DVal;
}
 
template <class T> 
inline 
EllVar<T>  
operator-
( const EllVar<T>& EllVar1 )
{
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // Update Centre
  EI->_q( i )  =  - EI->_q(k) ;
  // Update matrix when the shape matrix has been initialised with a dependency map
  if( EI-> _depmap_flag )
  {
  	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j )
	  {
  		if( EI->_Q.isListed(i,j) ) EI->_Q(i,j) =  EI->_Q(k,j) ;
  	  }
	  // Update Diagonal element
	  EI->_Q(i,i) = EI->_Q(k,k) ;
  }
  else // No dependency map available 
  {
	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j ) 
	  {
	  	EI->_Q.put( i,j, - EI->_Q(k,j) );
	  }
	  // Update Diagonal element
	  EI->_Q.put( i,i, EI->_Q(k,k) );
  }
  // Save range of Variable 
  EI->_setbound( EllVar3 );
  
  return EllVar3;
}
 
template <class T> 
inline 
EllVar<T>  
operator-
( const EllVar<T>& EllVar1 , 
  const EllVar<T>& EllVar2  )
{
  if( EllVar1._EI != EllVar2._EI )
  {	
   	throw typename EllImg<T>::Exceptions( EllImg<T>::Exceptions::EIMG );
  }
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  long l          = EllVar2._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // Update Centre
  EI->_q( i )  =  EI->_q(k) - EI->_q(l);
  // Update matrix when the shape matrix has been initialised with a dependency map
  if( EI-> _depmap_flag )
  {
  	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j )
	  {
  		if( EI->_Q.isListed(i,j) ) EI->_Q(i,j) = EI->_Q(k,j) - EI->_Q(l,j) ;
  	  }
	  // Update Diagonal element
	  EI->_Q(i,i) = EI->_Q(k,k) - 2.*(EI->_Q(l,k)) + EI->_Q(l,l) ;
  }
  else // No dependency map available 
  {
	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j ) 
	  {
		EI->_Q.put(i,j, EI->_Q(k,j) - EI->_Q(l,j)  );
	  }
	  // Update Diagonal element
	  EI->_Q.put( i,i, EI->_Q(k,k) - 2.*(EI->_Q(l,k)) + EI->_Q(l,l) );
  }
  // Save range of Variable 
  EI->_setbound( EllVar3 );

  return EllVar3;
}

template <class T> 
inline 
EllVar<T>  
operator-
( const EllVar<T>& EllVar1 , 
  const double  DVal )
{
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // Update Centre
  EI->_q( i )  =  EI->_q(k) - DVal;
  // Update matrix when the shape matrix has been initialised with a dependency map
  if( EI-> _depmap_flag )
  {
  	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j )
	  {
  		if( EI->_Q.isListed(i,j) ) EI->_Q(i,j) = EI->_Q(k,j) ;
  	  }
	  // Update Diagonal element
	  EI->_Q(i,i) = EI->_Q(k,k) ;
  }
  else // No dependency map available 
  {
	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j )  
	  {
		EI->_Q.put(i,j, EI->_Q(k,j) );
	  }
	  // Update Diagonal element
	  EI->_Q.put( i,i, EI->_Q(k,k) );
  }
  // Save range of Variable 
  EI->_setbound( EllVar3 );

  return EllVar3;
}

template <class T> 
inline 
EllVar<T> 
operator-
( const double  DVal, 
  const EllVar<T>& EllVar1 )
{
  EllImg<T>* EI = EllVar1._EI;
  long i       = EI->_current_row;
  long k       = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // Update Centre
  EI->_q( i )  =  DVal - EI->_q(k) ;
  // Update matrix when the shape matrix has been initialised with a dependency map
  if( EI-> _depmap_flag )
  {
  	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j )
	  {
  		if( EI->_Q.isListed(i,j) ) EI->_Q(i,j) = - EI->_Q(k,j) ;
  	  }
	  // Update Diagonal element
	  EI->_Q(i,i) = EI->_Q(k,k) ;
  }
  else // No dependency map available 
  {
	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j ) 
	  {
		EI->_Q.put(i,j, - EI->_Q(k,j) );
	  }
	  // Update Diagonal element
	  EI->_Q.put( i,i, EI->_Q(k,k) );
  }
  // Save range of Variable 
  EI->_setbound( EllVar3 );

  return EllVar3;
}
 
template <class T> 
inline 
EllVar<T>  
operator*
( const EllVar<T>& EllVar0 , 
  const EllVar<T>& EllVar1  )
{
  if( EllVar0._EI != EllVar1._EI )
  {	
   	throw typename EllImg<T>::Exceptions( EllImg<T>::Exceptions::EIMG );
  }
  EllImg<T>* EI = EllVar0._EI;
  long i;
  long k  = EllVar0._RowInd;
  long l  = EllVar1._RowInd;
  T domain_EllVar0  = EI->_q( k ) + std::sqrt( EI->_Q(k,k) ) *T(-1.,1);
  T domain_EllVar1  = EI->_q( l ) + std::sqrt( EI->_Q(l,l) ) *T(-1.,1);
  if( EI->_ip == -1 )
  {
	  EI->_ip = EI->_current_row    ;	// The product block starts at the current_row 
	  if( EI->_ip  + 7 > EI->_Q.n   )         // If there are not at least 7 more rows  to accomodate the product block then
	  {
	  	EI->_Q.stretch( 7 )     ;   // Stretch to make space for the product block
	  	EI->_q.stretch( 7 )     ;	// in every row, updates are performed until i (i,i) is the diagonal element
  	  }
  	  for( long ip = 0 ; ip < 7 ; ++ip ) EI->_Irange.push_back(T(0.,0.)); // Add 6 trivial elements to the range vector
      i  = EI->_current_row + 6 ; // Index for product result is the last row created 
  }
  else i = EI->_current_row;
   
  long ii = EI->_ip		  ;	// internal row counter for product block
  long kk = ii            ; // internal counter for argument in univariate operations;
  long ll				  ; // internal counter for 2nd argument 
  // long jmax = i           ;
  EllVar<T> EllVar3( EI, i , i+1 );
#ifdef MC__ELLIMAGE_DEBUG
  EI->_dbugout << std::scientific << std::setprecision(3) << std::right;
  EI->_dbugout << "Product starts: i= " << i <<" k= "<< k << " l= " << l <<std::endl;
  EI->_dbugout << "Product block starts: "  << EI->_ip << std::endl;  
  EI->_dbugout << "Range, q , Q\n" ;
  for( long i = 0 ; i < EI->_Q.n ; ++i )
  {
    EI->_dbugout << EI->_Irange[i] << "\t" << EI->_q(i) << "\t" << EI->_Q.row(i); 
  }
  EI->_dbugout << std::endl;
#endif 
  // compute v2 = ( v0 + v1 )
  for( long j = 0 ; j <= i ; ++j )
  {
  	if(  EI->_Q.isListed(ii,j) ) EI->_Q(ii,j) = EI->_Q( k,j ) + EI->_Q( l,j );
  	else EI->_Q.put( ii,j, EI->_Q( k,j ) + EI->_Q( l,j ) );
  } 
  if( EI->_Q.isListed(ii,ii) )  EI->_Q(ii,ii) =  EI->_Q( k,k ) + 2.* EI->_Q( l,k ) + EI->_Q( l,l );
  else EI->_Q.put( ii,ii, EI->_Q( k,k ) + 2.* EI->_Q( l,k ) + EI->_Q( l,l ) ); 
  EI->_q( ii ) = EI->_q( k ) + EI->_q( l ); 
  ii += 1; // increase row counter for product block
  // compute v3 = ( v0 - v1 )
  for( long j = 0 ; j <= i ; ++j )
  {
  	if(  EI->_Q.isListed(ii,j) ) EI->_Q(ii,j) = EI->_Q( k,j ) - EI->_Q( l,j );
  	else EI->_Q.put( ii,j, EI->_Q( k,j ) - EI->_Q( l,j ) );
  } 
  if( EI->_Q.isListed(ii,ii) )  EI->_Q(ii,ii) =  EI->_Q( k,k ) - 2.* EI->_Q( l,k ) + EI->_Q( l,l );
  else EI->_Q.put( ii,ii, EI->_Q( k,k ) - 2.* EI->_Q( l,k ) + EI->_Q( l,l ) ); 
  EI->_q( ii ) = EI->_q( k ) - EI->_q( l ); 
  T domainv3 =  EI->_q( ii ) + std::sqrt( EI->_Q( ii , ii ) ) * T(-1.,1.); // Compute domain for v5 before Msum (in v4) 
  ii += 1; // increase row counter for product block
  // compute v4 = ( v2 )^2
  T domain = EI->_q( kk ) + std::sqrt( EI->_Q(kk,kk) )*T(-1.,1.);
  double c_0, c_1, eta;
  std::tie( c_0, c_1, eta ) = EI->_linearize_sqr( domain );
  for( long j = 0 ; j <= i ; ++j )
  {
  	if(  EI->_Q.isListed(ii,j) ) EI->_Q(ii,j) =  c_1*EI->_Q( kk ,j );
  	else EI->_Q.put( ii,j, c_1*EI->_Q( kk ,j ) );
  } 
  if( EI->_Q.isListed(ii,ii) )  EI->_Q(ii,ii) =  std::pow(c_1,2)*EI->_Q(kk,kk);
  else EI->_Q.put( ii,ii, std::pow(c_1,2)*EI->_Q(kk,kk) );    
  EI->_q( ii ) = c_0;    
  EI->_minksum( ii, eta );
  ii += 1; // increase row counter for product block
  kk += 1; // increase counter for argument
  // compute v5 = ( v3 )^2  
  std::tie( c_0, c_1, eta ) = EI->_linearize_sqr( domainv3 );                
  for( long j = 0 ; j <= i ; ++j )
  {
  	if(  EI->_Q.isListed(ii,j) ) EI->_Q(ii,j) =  c_1*EI->_Q( kk ,j );
  	else EI->_Q.put( ii,j, c_1*EI->_Q( kk ,j ) );
  } 
  if( EI->_Q.isListed(ii,ii) )  EI->_Q(ii,ii) =  std::pow(c_1,2)*EI->_Q(kk,kk);
  else EI->_Q.put( ii,ii, std::pow(c_1,2)*EI->_Q(kk,kk) );   
  EI->_q( ii ) = c_0;
  EI->_minksum( ii, eta );
  ii += 1; // increase row counter for product block
  kk += 1; // increase counter for argument
  // compute v6 = (1/4)*( v4 )
  for( long j = 0 ; j <= i ; ++j )
  {
  	if(  EI->_Q.isListed(ii,j) ) EI->_Q(ii,j) =  (1./4.)*EI->_Q( kk ,j );
  	else EI->_Q.put( ii,j, (1./4.)*EI->_Q( kk ,j ) );
  } 
  if( EI->_Q.isListed(ii,ii) )  EI->_Q(ii,ii) =  (1./16.)*EI->_Q(kk,kk);
  else EI->_Q.put( ii,ii,(1./16.)*EI->_Q(kk,kk) );   
  EI->_q( ii ) = (1./4.)*EI->_q( kk );
  ii += 1; // increase row counter for product block
  kk += 1; // increase counter for argument
  // compute v7 = (1/4)*( v5 )
  for( long j = 0 ; j <= i ; ++j )
  {
  	if(  EI->_Q.isListed(ii,j) ) EI->_Q(ii,j) =  (1./4.)*EI->_Q( kk ,j );
  	else EI->_Q.put( ii,j, (1./4.)*EI->_Q( kk ,j ) );
  } 
  if( EI->_Q.isListed(ii,ii) )  EI->_Q(ii,ii) =  (1./16.)*EI->_Q(kk,kk);
  else EI->_Q.put( ii,ii,(1./16.)*EI->_Q(kk,kk) );  
  EI->_q( ii ) = (1./4.)*EI->_q( kk );
  kk += 1;
  ll  = kk+1;
  // compute v8 = ( v6 - v7 )
  for( long j = 0 ; j < i ; ++j )
  {
  	if(  EI->_Q.isListed(i,j) ) EI->_Q(i,j) = EI->_Q( kk,j ) - EI->_Q( ll,j );
  	else EI->_Q.put( i,j, EI->_Q( kk,j ) - EI->_Q( ll,j ) );
  } 
  if( EI->_Q.isListed(i,i) )  EI->_Q(i,i) =  EI->_Q( kk,kk ) - 2.* EI->_Q( ll,kk ) + EI->_Q( ll,ll );
  else EI->_Q.put( i,i, EI->_Q( kk,kk ) - 2.* EI->_Q( ll,kk ) + EI->_Q( ll,ll ) ); 
  EI->_q( i ) = EI->_q( kk ) - EI->_q( ll ); 
  //cleanup the product block
  for (long ip= EI->_ip ; ip <= ii ; ++ip )
  {
  	for (long jp= 0 ; jp <= i ; ++jp ) 
  	{
	  	EI -> _Q.del( ip , jp ) ; 
  	}
//  	EI -> _Q.del( ip , ip ) ; 
  }
  //save bound the product 
  EI->_setbound( EllVar3 , domain_EllVar0 * domain_EllVar1 );
#ifdef MC__ELLIMAGE_DEBUG
  EI->_dbugout << std::scientific << std::setprecision(3) << std::right;
  EI->_dbugout << "Product ends: i= " << i <<" k= "<< k << " l= " << l <<std::endl;
  EI->_dbugout << "Product block starts: "  << EI->_ip << std::endl;  
  EI->_dbugout << "Range, q , Q\n" ;
  for( long i = 0 ; i < EI->_Q.n ; ++i )
  {
    EI->_dbugout << EI->_Irange[i] << "\t" << EI->_q(i) << "\t" << EI->_Q.row(i); 
  }
  EI->_dbugout << std::endl;
#endif 
  
  return EllVar3;
}

template <class T> 
inline 
EllVar<T> 
operator*
( const EllVar<T>& EllVar1 , 
  const double  DVal )
{
  EllImg<T>* EI = EllVar1._EI;
  long i       = EI->_current_row;
  long k       = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // Update Centre
  EI->_q( i )  =  EI->_q(k) * DVal;
  // Update matrix when the shape matrix has been initialised with a dependency map
  if( EI-> _depmap_flag )
  {
  	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j )
	  {
  		if( EI->_Q.isListed(i,j) ) EI->_Q(i,j) = DVal * EI->_Q(k,j) ;
  	  }
	  // Update Diagonal element
	  EI->_Q(i,i) = std::pow(DVal,2)*EI->_Q(k,k) ;
  }
  else // No dependency map available 
  {
	  // Update current row in the Shape matrix
	  for ( long j = 0; j < i ; ++j ) 
	  {
		EI->_Q.put(i,j, DVal * EI->_Q(k,j) );
	  }
	  // Update Diagonal element
	  EI->_Q.put( i,i, std::pow(DVal,2)*EI->_Q(k,k) );
  }
  // Save range of Variable 
  EI->_setbound( EllVar3 );
  return EllVar3;
}

template <class T> 
inline 
EllVar<T>  
operator*
( const double  DVal, 
  const EllVar<T>& EllVar1 )
{
  return EllVar1 * DVal;
}
 
template <class T> 
inline 
EllVar<T>  
operator/
( const EllVar<T>& EllVar1, 
  const EllVar<T>& EllVar2 )
{
  return EllVar1 * inv( EllVar2 );
}

template <class T> 
inline 
EllVar<T> 
operator/
( const EllVar<T>& EllVar1, 
  const double  DVal  )
{
  if( isequal( DVal, 0. ) )
  {	
   	throw typename EllImg<T>::Exceptions( EllImg<T>::Exceptions::DIV );
  }
  return ( 1./DVal ) * EllVar1;
}

template <class T> 
inline 
EllVar<T>  
operator/
( const double  DVal, 
  const EllVar<T>& EllVar1 )
{
  return DVal * inv( EllVar1 );
}

template <class T> 
inline 
EllVar<T> 
inv
( const EllVar<T>& EllVar1 )
{
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // bound the argument EllVar1 in the current ellipsoid
  T domain = EI->_bound( EllVar1 );
  if( Op<T>::l(domain) <= 0. && Op<T>::u(domain) >= 0. )
  {	
   	throw typename EllImg<T>::Exceptions( EllImg<T>::Exceptions::INV );
  } 
  // Compute linear approximation of inv( m + r*x ) with x \in [-1,1] using Chebyshev models
  double c_0, c_1, eta;
  if( EI->options.CHEBYSHEV_USE )
  {
	std::tie( c_0, c_1, eta ) = EI->_cmodel_linear( inv , domain );
  }
  else // Compute linear approximation of inv( m + r*x ) with x \in [-1,1] using minimax linearisation
  {
  	double m = Op<T>::mid( domain );
  	double r = 0.5 * Op<T>::diam( domain );
	double c_sec=0., x_tan=0., c_tan=0.;
	c_sec   = m/(std::pow(m,2) - std::pow(r,2));
	c_1     = r/(std::pow(r,2) - std::pow(m,2));
	if ( m > 0 ) x_tan =  (std::sqrt(-r/c_1)-m)/r;
	else         x_tan =  1./((std::sqrt(-r/c_1)-m)/r);
	c_tan   = 1./(m+r*x_tan)-c_1*x_tan;
	eta     = (c_sec - c_tan)/2.;
	c_0     = c_sec - eta;
	c_1     = c_1/r;
  } 
  // Update matrix
  EI->_univupdate( i , k , c_0 , c_1 , eta ); 
  // Save range of Variable 
  EI->_setbound( EllVar3 , 1./(domain) );
#ifdef MC__ELLIMAGE_DEBUG
  EI->_dbugout << std::scientific << std::setprecision(3) << std::right;
  EI->_dbugout << "inv ends: i= " << i <<" k= "<< k <<std::endl;
  EI->_dbugout << "Range, q , Q\n" ;
  for( long i = 0 ; i < EI->_Q.n ; ++i )
  {
    EI->_dbugout << EI->_Irange[i] << "\t" << EI->_q(i) << "\t" << EI->_Q.row(i); 
  }
  EI->_dbugout << std::endl;
#endif 
  return EllVar3;  
}
 
template <class T> 
inline 
EllVar<T> 
exp
( const EllVar<T>& EllVar1 )
{
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // bound the argument EllVar1 in the current ellipsoid
  T domain = EI->_bound( EllVar1 );
  // Compute linear approximation of exp( m + r*x ) with x \in [-1,1] using Chebyshev models
  double c_0, c_1, eta;
  if( EI->options.CHEBYSHEV_USE )
  {
	std::tie( c_0, c_1, eta ) = EI->_cmodel_linear( exp , domain );
  }
  else // Compute linear approximation of exp( m + r*x ) with x \in [-1,1] using minimax linearisation
  {
	double m = Op<T>::mid( domain );
	double r = 0.5 * Op<T>::diam( domain );
	double c_sec=0., x_tan=0., c_tan=0.;
	c_sec  = 0.5*(std::exp(m-r)+std::exp(m+r));
	c_1    = 0.5*(std::exp(m+r)-std::exp(m-r));
	x_tan  = (std::log(c_1/r) - m)/r;
	c_tan  = std::exp(m+r*x_tan)-c_1*x_tan;
	eta = (c_sec - c_tan)/2.0;
	c_0 = c_sec - eta;
	c_1 = c_1/r;
  }
  // Update matrix 
  EI->_univupdate( i , k , c_0 , c_1 , eta ); 
  // Save range of Variable 
  EI->_setbound( EllVar3 , exp(domain) );
  
  return EllVar3; 
}

template <class T> 
inline 
EllVar<T> 
log
( const EllVar<T>& EllVar1 )
{
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
#ifdef MC__ELLIMAGE_DEBUG
  EI->_dbugout << std::scientific << std::setprecision(3) << std::right;
  EI->_dbugout << "log starts: i= " << i <<" k= "<< k <<std::endl;
  EI->_dbugout << "Range, q , Q\n" ;
  for( long i = 0 ; i < EI->_Q.n ; ++i )
  {
    EI->_dbugout << EI->_Irange[i] << "\t" << EI->_q(i) << "\t" << EI->_Q.row(i); 
  }
  EI->_dbugout << std::endl;
#endif 
  // bound the argument EllVar1 in the current ellipsoid
  T domain = EI->_bound( EllVar1 );
  if( Op<T>::l(domain) <= 0.  )
  {	
   	throw typename EllImg<T>::Exceptions( EllImg<T>::Exceptions::LOG );
  } 
  // Compute linear approximation of log( m + r*x ) with x \in [-1,1] using Chebyshev models
  double c_0, c_1, eta;
  if( EI->options.CHEBYSHEV_USE )
  {
  	std::tie( c_0, c_1, eta ) = EI->_cmodel_linear( log , domain );
  }
  else // Compute linear approximation of log( m + r*x ) with x \in [-1,1] using minimax linearisation
  {
	double m = Op<T>::mid( domain );
	double r = 0.5 * Op<T>::diam( domain );
	double c_sec=0., x_tan=0., c_tan=0.;
	c_sec = (1./2.)*(std::log(m-r)+std::log(m+r));
	c_1   = (1./2.)*(std::log(m+r)-std::log(m-r));
	x_tan = (1./c_1)-(m/r);
	c_tan = std::log(m+r*x_tan)-c_1*x_tan; 
	eta   = std::fabs((c_sec - c_tan)/2.);
	c_0   = c_sec + eta;
	c_1   = c_1/r;
  }
  // Update matrix 
  EI->_univupdate( i , k , c_0 , c_1 , eta ); 
  // Save range of Variable 
  EI->_setbound( EllVar3 , log(domain) );
#ifdef MC__ELLIMAGE_DEBUG
  EI->_dbugout << std::scientific << std::setprecision(3) << std::right;
  EI->_dbugout << "log ends: i= " << i <<" k= "<< k <<std::endl;
  EI->_dbugout << "Range, q , Q\n" ;
  for( long i = 0 ; i < EI->_Q.n ; ++i )
  {
    EI->_dbugout << EI->_Irange[i] << "\t" << EI->_q(i) << "\t" << EI->_Q.row(i); 
  }
  EI->_dbugout << std::endl;
#endif   
  return EllVar3;  
}

template <class T> 
inline 
EllVar<T> 
sqrt
( const EllVar<T>& EllVar1 )
{
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // bound the argument EllVar1 in the current ellipsoid
  T domain = EI->_bound( EllVar1 );
  if( Op<T>::l(domain) < 0.  )
  {	
   	throw typename EllImg<T>::Exceptions( EllImg<T>::Exceptions::SQRT );
  }
  // Compute linear approximation of sqrt( m + r*x ) with x \in [-1,1] using Chebyshev models
  double c_0, c_1, eta;
  if( EI->options.CHEBYSHEV_USE )
  {
  	std::tie( c_0, c_1, eta ) = EI->_cmodel_linear( sqrt , domain );
  }
  else // Compute linear approximation of sqrt( m + r*x ) with x \in [-1,1] using minimax linearisation
  {
	double m = Op<T>::mid( domain );
	double r = 0.5 * Op<T>::diam( domain );
	double c_sec=0., x_tan=0., c_tan=0.;
	c_sec = (0.5)*(std::sqrt(m-r)+std::sqrt(m+r)); 
	c_1   = (0.5)*(std::sqrt(m+r)-std::sqrt(m-r)); 
	x_tan = (std::pow(r,2) - 4.0*std::pow((c_1),2)*m )/(4.0*std::pow((c_1),2)*r); 
	c_tan = std::sqrt(m+r*x_tan)-c_1*x_tan; 
	eta   = std::fabs((c_sec - c_tan)/2.0); 
	c_0   = c_sec + eta;
	c_1   = c_1 / r ;   
  }								
  // Update matrix 
  EI->_univupdate( i , k , c_0 , c_1 , eta ); 
  // Save range of Variable 
  EI->_setbound( EllVar3 , sqrt(domain) );
  
  return EllVar3; 
}

template <class T> 
inline 
EllVar<T> 
sqr
( const EllVar<T>& EllVar1 )
{
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
#ifdef MC__ELLIMAGE_DEBUG
  EI->_dbugout << std::scientific << std::setprecision(3) << std::right;
  EI->_dbugout << "sqr starts: i= " << i <<" k= "<< k <<std::endl;
  EI->_dbugout << "Range, q , Q\n" ;
  for( long i = 0 ; i < EI->_Q.n ; ++i )
  {
    EI->_dbugout << EI->_Irange[i] << "\t" << EI->_q(i) << "\t" << EI->_Q.row(i); 
  }
  EI->_dbugout << std::endl;
#endif 
  // bound the argument EllVar1 in the current ellipsoid
  T domain = EI->_bound( EllVar1 );
  // Compute linear approximation of sqr( m + r*x ) with x \in [-1,1] using Chebyshev models
  double c_0, c_1, eta;
  if( EI->options.CHEBYSHEV_USE )
  {
  	std::tie( c_0, c_1, eta ) = EI->_cmodel_linear( sqr , domain );
  }
  else // Compute linear approximation of sqr( m + r*x ) with x \in [-1,1] using minimax linearisation
  {
	double m = Op<T>::mid( domain );
	double r = 0.5 * Op<T>::diam( domain );
	double c_sec=0., x_tan=0., c_tan=0.;
	c_sec = std::pow(m,2)+std::pow(r,2);
	c_1   = 2.0*m*r;
	x_tan = (c_1-2.0*m*r)/(2.0*std::pow(r,2));
	c_tan = std::pow((m+r*x_tan),2)-c_1*x_tan;
	eta   = (c_sec - c_tan)/2.0;
	c_0   = c_sec - eta;
    c_1   = c_1 / r;
  }
  // Update matrix 
  EI->_univupdate( i , k , c_0 , c_1 , eta ); 
  // Save range of Variable 
  EI->_setbound( EllVar3 , sqr(domain) );
#ifdef MC__ELLIMAGE_DEBUG
  EI->_dbugout << std::scientific << std::setprecision(3) << std::right;
  EI->_dbugout << "sqr ends: i= " << i <<" k= "<< k <<std::endl;
  EI->_dbugout << "Range, q , Q\n" ;
  for( long i = 0 ; i < EI->_Q.n ; ++i )
  {
    EI->_dbugout << EI->_Irange[i] << "\t" << EI->_q(i) << "\t" << EI->_Q.row(i); 
  }
  EI->_dbugout << std::endl;
#endif 
  return EllVar3; 
}

template <class T> 
inline 
EllVar<T> 
pow
( const EllVar<T>& EllVar1 , const int n )
{
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // bound the argument EllVar1 in the current ellipsoid
  T domain = EI->_bound( EllVar1 );
  // Compute linear approximation of pow( m + r*x , n ) with x \in [-1,1] using Chebyshev models
  double c_0, c_1, eta;
  if( EI->options.CHEBYSHEV_USE )
  {
  	std::tie( c_0, c_1, eta ) = EI->_cmodel_linear( pow , domain , n );
  }
  else // Compute linear approximation of log( m + r*x , n ) with x \in [-1,1] using minimax linearisation
  {
	double m = Op<T>::mid( domain );
	double r = 0.5 * Op<T>::diam( domain );
	double c_sec=0., x_tan=0., c_tan=0.;
    double x_tanL, x_tanU, c_tanL, c_tanU;
	if( n % 2 == 0 ) // even power
	{
		c_sec   = (1./2.)*(std::pow((m+r),n)+std::pow((m-r),n));
		c_1     = (1./2.)*(std::pow((m+r),n)-std::pow((m-r),n));
		x_tan   = (std::pow((c_1/(n*r)),1./(n-1))-m)/r;
		c_tan   = std::pow((m+r*x_tan),n)-c_1*x_tan;
		eta     = (c_sec - c_tan)/2.;
		c_0     = c_sec - eta;   	
	}
	else
	{
		if(Op<T>::l( domain ) >= 0 || Op<T>::u( domain ) <= 0)
		{
			c_1     = (1./2.)*(std::pow((m+r),n)-std::pow((m-r),n));
			c_sec   = (1./2.)*(std::pow((m+r),n)+std::pow((m-r),n));
			x_tan   = (sign(m)*std::pow((c_1/(n*r)),1./(n-1))-m)/r;
			c_tan   = std::pow((m+r*x_tan),n)-c_1*x_tan;
			eta     = (c_sec - c_tan)/2.;
			c_0     = c_sec - eta;
		}
		else
		{
			c_1    = (1./2.)*(std::pow((m+r),n)-std::pow((m-r),n));
			x_tanU = (std::pow((c_1/(n*r)),1./(n-1))-m)/r;
			x_tanL = (-1.0*std::pow((c_1/(n*r)),1./(n-1))-m)/r;
			c_tanU = std::pow((m+r*x_tanU),n)-c_1*x_tanU;
			c_tanL = std::pow((m+r*x_tanL),n)-c_1*x_tanL;
			eta    = std::fabs((c_tanL - c_tanU)/2.);
			c_0    = std::fabs(c_tanL - eta);
		}
	}
	c_1 = c_1 / r;
  }
  // Update matrix 
  EI->_univupdate( i , k , c_0 , c_1 , eta );   
  // Save range of Variable 
  EI->_setbound( EllVar3 , pow(domain, n) );
  
  return EllVar3; 
}

template <class T> 
inline 
EllVar<T> 
cos
( const EllVar<T>& EllVar1 )
{
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // bound the argument EllVar1 in the current ellipsoid
  T domain = EI->_bound( EllVar1 );
  // Compute linear approximation of cos( m + r*x ) with x \in [-1,1] using Chebyshev models
  double c_0, c_1, eta;
  std::tie( c_0, c_1, eta ) = EI->_cmodel_linear( cos , domain );
  // Update matrix 
  EI->_univupdate( i , k , c_0 , c_1 , eta ); 
  // Save range of Variable 
  EI->_setbound( EllVar3 , cos(domain) );
  
  return EllVar3; 
}

template <class T> 
inline 
EllVar<T> 
sin
( const EllVar<T>& EllVar1 )
{
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // bound the argument EllVar1 in the current ellipsoid
  T domain = EI->_bound( EllVar1 );
  // Compute linear approximation of sin( m + r*x ) with x \in [-1,1] using Chebyshev models
  double c_0, c_1, eta;
  std::tie( c_0, c_1, eta ) = EI->_cmodel_linear( sin , domain );
  // Update matrix 
  EI->_univupdate( i , k , c_0 , c_1 , eta ); 
  // Save range of Variable 
  EI->_setbound( EllVar3 , sin(domain) );
  
  return EllVar3; 
}

template <class T> 
inline 
EllVar<T> 
tan
( const EllVar<T>& EllVar1 )
{
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // bound the argument EllVar1 in the current ellipsoid
  T domain = EI->_bound( EllVar1 );
  if( Op<T>::l(cos(domain)) <= 0. && Op<T>::u(cos(domain)) >= 0. )
  {	
   	throw typename EllImg<T>::Exceptions( EllImg<T>::Exceptions::TAN );
  } 
  // Compute linear approximation of tan( m + r*x ) with x \in [-1,1] using Chebyshev models
  double c_0, c_1, eta;
  std::tie( c_0, c_1, eta ) = EI->_cmodel_linear( tan , domain );
  // Update matrix 
  EI->_univupdate( i , k , c_0 , c_1 , eta ); 
  // Save range of Variable 
  EI->_setbound( EllVar3 , tan(domain) );
  
  return EllVar3; 
}

template <class T> 
inline 
EllVar<T> 
acos
( const EllVar<T>& EllVar1 )
{
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // bound the argument EllVar1 in the current ellipsoid
  T domain = EI->_bound( EllVar1 );
  if( Op<T>::l(domain) < -1. && Op<T>::u(domain) > 1. )
  {	
   	throw typename EllImg<T>::Exceptions( EllImg<T>::Exceptions::ACOS );
  } 
  // Compute linear approximation of acos( m + r*x ) with x \in [-1,1] using Chebyshev models
  double c_0, c_1, eta;
  std::tie( c_0, c_1, eta ) = EI->_cmodel_linear( acos , domain );
  // Update matrix 
  EI->_univupdate( i , k , c_0 , c_1 , eta ); 
  // Save range of Variable 
  EI->_setbound( EllVar3 , acos(domain) );
  
  return EllVar3; 
}

template <class T> 
inline 
EllVar<T> 
asin
( const EllVar<T>& EllVar1 )
{
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // bound the argument EllVar1 in the current ellipsoid
  T domain = EI->_bound( EllVar1 );
  if( Op<T>::l(domain) < -1. && Op<T>::u(domain) > 1. )
  {	
   	throw typename EllImg<T>::Exceptions( EllImg<T>::Exceptions::ACOS );
  } 
  // Compute linear approximation of asin( m + r*x ) with x \in [-1,1] using Chebyshev models
  double c_0, c_1, eta;
  std::tie( c_0, c_1, eta ) = EI->_cmodel_linear( asin , domain );
  // Update matrix 
  EI->_univupdate( i , k , c_0 , c_1 , eta ); 
  // Save range of Variable 
  EI->_setbound( EllVar3 , asin(domain) );
  
  return EllVar3; 
}

template <class T> 
inline 
EllVar<T> 
atan
( const EllVar<T>& EllVar1 )
{
  EllImg<T>* EI = EllVar1._EI;
  long i          = EI->_current_row;
  long k          = EllVar1._RowInd;
  // Construct Evar3 to return (increases _current_row)
  EllVar<T> EllVar3( EI, i );
  // bound the argument EllVar1 in the current ellipsoid
  T domain = EI->_bound( EllVar1 );
  // Compute linear approximation of acos( m + r*x ) with x \in [-1,1] using Chebyshev models
  double c_0, c_1, eta;
  std::tie( c_0, c_1, eta ) = EI->_cmodel_linear( atan , domain );
  // Update matrix 
  EI->_univupdate( i , k , c_0 , c_1 , eta ); 
  // Save range of Variable 
  EI->_setbound( EllVar3 , atan(domain) );
  
  return EllVar3; 
}
template <class T> 
inline
std::ostream& 
operator<< 
( std::ostream&os      , 
  const EllImg<T>& EI  )
{
  os << "Lifted Ellipsoidal Image\n";
  os << "center: \n";
  os << EI._q << std::endl; 
  os << "Shape: \n";
  os << EI._Q << std::endl;
  return os;
}

} // namespace mc

#endif
