#include <fstream>
#include "odebnd_gsl.hpp"
#include "interval.hpp"
#include "tmodel.hpp"

typedef mc::Interval I;
typedef mc::Ellipsoid E;
typedef mc::TModel<I> PM;
typedef mc::TVar<I> PV;


int main()
{
  // DEFINE DIMENSIONS OF THE PROBLEM	
  const unsigned int NP = 1;  // Parameter dimension
  const unsigned int NX = 2;  // State dimension
  const unsigned int NSAMP = 10; // number of samples

  // time span
  double t0 = 0., tf = 5.;    // Time span
  const unsigned int NS = 100;  // Time stages
  double tk[NS+1]; tk[0] = t0;
  for( unsigned k=0; k<NS; k++ ) tk[k+1] = tk[k] + (tf-t0)/(double)NS;

  // DEFINE GRAPH, VARIABLES, PARAMETERS
  mc::FFGraph IVP;   						// DAG describing the IVP 	
  mc::FFVar P[NP];  						// Parameter array
  mc::FFVar X[NX];  						// State array  
  mc::FFVar RHS[NX];  						// Right-hand side function
  mc::FFVar IC[NX];   						// Initial value function
  for( unsigned int i=0; i<NP; i++ ) P[i].set( &IVP );  	// Set parameters in the DAG
  for( unsigned int i=0; i<NX; i++ ) X[i].set( &IVP );  	// Set states in the DAG
  
  // DEFINE RHS   
  RHS[0] = P[0] * X[0] * ( 1. - X[1] );
  RHS[1] = P[0] * X[1] * ( X[0] - 1. );
  
  // DEFINE IC
  IC[0] = 1.2;
  IC[1] = 1.1;

  // DEFINE SET ARITHMETICS
  // Interval		
  I Ip[NP];
  Ip[0] = I(2.95,3.05);
  I *Ixk[NS+1];
  for( unsigned k=0; k<=NS; k++ ) Ixk[k] = new I[NX];
  
  // Polynomial Model
  const unsigned int NTM = 1;
  PM PMEnv( NP, NTM );
  PV PMp[NP] ;
  for( unsigned int i = 0; i<NP; i++) PMp[i] = PV( &PMEnv, i, Ip[i] ) ;
  PV *PMxk[NS+1];
  for( unsigned k=0; k<=NS; k++ ){
    PMxk[k] = new PV[NX];
  }

  // DEFINE ODE FOR EXACT TUBE 
  // (mc::ODESLV_GSL for sampling)
  mc::ODESLV_GSL<I> EXACT_ODE;

  EXACT_ODE.set_dag( &IVP );
  EXACT_ODE.set_state( NX, X );
  EXACT_ODE.set_parameter( NP, P );
  EXACT_ODE.set_dynamic( RHS );
  EXACT_ODE.set_initial( IC );

  EXACT_ODE.options.DISPLAY = 1;
  EXACT_ODE.options.RESRECORD = true;

  std::cout << "\nNON_VALIDATED INTEGRATION - APPROXIMATE ENCLOSURE OF REACHABLE SET:\n\n";
  EXACT_ODE.bounds( NS, tk, Ip, Ixk, 0, NSAMP );
  std::ofstream file0( "ODE_EXACT.out", std::ios_base::out );
  EXACT_ODE.record( file0 );


  // DEFINE SET PROPAGATION METHOD 
  // mc::ODEBND_GSL<> for continuous time, 
  // mc::ODEBND_VAL<> for   discrete time )
  mc::ODEBND_GSL<I> BOUND_ODE ; 
		
  BOUND_ODE.set_dag( &IVP );			
  BOUND_ODE.set_state( NX, X );
  BOUND_ODE.set_parameter( NP, P );
  BOUND_ODE.set_dynamic( RHS );
  BOUND_ODE.set_initial( IC );

 
  BOUND_ODE.options.RESRECORD = true;
  BOUND_ODE.options.WRAPMIT = mc::ODEBND_GSL<I>::Options::ELLIPS;
  std::cout << "\nAPPROXIMATE ENCLOSURE OF REACHABLE SET:\n\n";
  BOUND_ODE.bounds( NS, tk, Ip, Ixk );
//  BOUND_ODE.bounds( NS, tk, PMp, PMxk  );
  std::ofstream file1( "ODE_BOUNDS.out", std::ios_base::out );
  BOUND_ODE.record( file1 );

  return 0;

  for( unsigned k=0; k<=NS; k++ ){
    delete[] Ixk[k];
    delete[] PMxk[k]; 
  }

}
