// Copyright (C) 2012-2014 Benoit Chachuat, Imperial College London.
// All Rights Reserved.
// This code is published under the Eclipse Public License.

#ifndef MC__BASE_ODE_HPP
#define MC__BASE_ODE_HPP

#include "ffunc.hpp"

namespace mc
{
//! @brief C++ base class for defining of parametric ODEs
////////////////////////////////////////////////////////////////////////
//! mc::BASE_ODE is a C++ base class for defining the variables,
//! parameters and functions participating in parametric ordinary
//! differential equations (ODEs).
////////////////////////////////////////////////////////////////////////
class BASE_ODE
{
protected:
  //! @brief pointer to DAG of ODE system
  FFGraph* _pDAG;

  //! @brief pointer to time/independent variable
  const FFVar* _pT;

  //! @brief number of states
  unsigned int _nx;

  //! @brief pointer to state variables (differential states)
  const FFVar* _pX;

  //! @brief number of adjoints
  unsigned int _ny;

  //! @brief pointer to adjoint variables (differential costates)
  FFVar* _pY;

  //! @brief number of parameters
  unsigned int _np;

  //! @brief pointer to parameters
  const FFVar* _pP;

  //! @brief vector of const pointers to initial value in each stage of ODE system
  std::vector<const FFVar*> _vIC;

  //! @brief vector of const pointers to RHS functions in each stage of ODE system
  std::vector<const FFVar*> _vRHS;

  //! @brief number of invariants
  unsigned int _ni;

  //! @brief vector of const pointers to invariant functions in each stage of ODE system
  std::vector<const FFVar*> _vINV;

  //! @brief number of state functionals
  unsigned int _nf;

  //! @brief vector of const pointers to functionals (assumes additive contributions) in each stage of ODE system
  std::vector<const FFVar*> _vFCT;

  //! @brief current time
  double _t;

  //! @brief current stage
  unsigned int _istg;

public:
  /** @ingroup ODESLV_GSL
   *  @ingroup ODEBND_GSL
   *  @ingroup ODEBND_VAL
   *  @{
   */

  //! @brief Class constructor
  BASE_ODE()
    : _pDAG(0), _pT(0), _nx(0), _pX(0), _ny(0), _pY(0), _np(0), _pP(0),
      _ni(0), _nf(0)
    {}

  //! @brief Class destructor
  virtual ~BASE_ODE()
    { delete[] _pY; }

  //! @brief Integrator status
  enum STATUS{
     NORMAL=0,	//!< Normal execution
     FAILURE,	//!< Integration breakdown (bounds explosion)
     FATAL	//!< Interruption due to errors in third-party libraries
  };

  //! @brief Get pointer to DAG
  FFGraph* dag() const
    { return _pDAG; }

  //! @brief Set pointer to DAG
  void set_dag
    ( FFGraph*pDAG )
    { _pDAG = pDAG; }

  //! @brief Get size of state
  unsigned int nx() const
    { return _nx; }

  //! @brief Get pointer to state
  const FFVar* state() const
    { return _pX; }

  //! @brief Set pointer to state
  void set_state
    ( const unsigned nx, const FFVar*pX )
    { _nx = nx; _pX = pX; }

  //! @brief Get pointer to adjoint
  const FFVar* adjoint() const
    { return _pY; }

  //! @brief Set pointer to adjoint
  void set_adjoint()
    { if( _ny != _nx ){ delete[] _pY; _ny = _nx; _pY = new FFVar[_ny]; }
      for( unsigned iy=0; iy<_ny; iy++ )
        if( _pY[iy].dag() != _pDAG ) _pY[iy].set( _pDAG ); }

  //! @brief Return size of parameter
  unsigned int np() const
    { return _np; }

  //! @brief Get pointer to parameter
  const FFVar* parameter() const
    { return _pP; }

  //! @brief Set pointer to parameter
  void set_parameter
    ( const unsigned np, const FFVar*pP )
    { _np = np; _pP = pP; }

  //! @brief Get pointer to time
  const FFVar* time() const
    { return _pT; }

  //! @brief Set pointer to time
  void set_time
    ( const FFVar*pT )
    { _pT = pT; }

  //! @brief Define RHS function in single-stage IVP-ODE
  void set_dynamic
    ( const FFVar*const RHS, const unsigned ni=0, const FFVar*const INV=0 )
    { _vRHS.clear(); _vRHS.push_back( RHS );
      _ni = ni; _vINV.clear(); _vINV.push_back( INV ); }

  //! @brief Define RHS function in multi-stage IVP-ODE with <a>ns</a> stages -- Assumes state is already defined using set_state()
  void set_dynamic
    ( const unsigned ns, const FFVar*const RHS, const unsigned ni=0, const FFVar*const INV=0 )
    { _vRHS.clear(); for( unsigned i=0; i<ns; i++ ) _vRHS.push_back( RHS+i*_nx );
      _ni = ni; _vINV.clear(); for( unsigned i=0; i<ns; i++ ) _vINV.push_back( INV+i*_ni ); }

  //! @brief Define IC function in single-stage IVP-ODE
  void set_initial
    ( const FFVar*const IC )
    { _vIC.clear(); _vIC.push_back( IC ); }

  //! @brief Define IC function in multi-stage IVP-ODE with <a>ns</a> stages -- Assumes state is already defined
  void set_initial
    ( const unsigned ns, const FFVar*const IC )
    { _vIC.clear(); for( unsigned i=0; i<ns; i++ ) _vIC.push_back( IC+i*_nx ); }

  //! @brief Define state function in single-stage IVP-ODE
  void set_function
    ( const unsigned nf, const FFVar*const FCT )
    { _nf = nf; _vFCT.clear(); _vFCT.push_back( FCT ); }

  //! @brief Define state function in multi-stage IVP-ODE with <a>ns</a> stages
  void set_function
    ( const unsigned nf, const unsigned ns, const FFVar*const FCT )
    { _nf = nf; _vFCT.clear(); for( unsigned i=0; i<ns; i++ ) _vFCT.push_back( FCT+i*_nf ); }

  //! @brief Define state function in multi-stage IVP-ODE with <a>ns</a> stages
  void set
    ( const BASE_ODE&ode )
    { _pDAG = ode._pDAG;
      _nx = ode._nx; _ny = ode._ny; _np = ode._np; _ni = ode._ni; _nf = ode._nf;
      _pT = ode._pT; _pX = ode._pX; _pY = ode._pY; _pP = ode._pP;
      _vIC = ode._vIC; _vRHS = ode._vRHS; _vINV = ode._vINV; _vFCT = ode._vFCT; };

  //! @brief Return last successful integration time
  double final_time() const
    { return _t; }
  /** @} */

protected:
  //! @brief Private methods to block default compiler methods
  BASE_ODE(const BASE_ODE&);
  BASE_ODE& operator=(const BASE_ODE&);
};

} // end namescape mc

#endif

