// Copyright (C) 2012-2014 Benoit Chachuat, Imperial College London.
// All Rights Reserved.
// This code is published under the Eclipse Public License.

#ifndef MC__ODEBND_GSL_HPP
#define MC__ODEBND_GSL_HPP

#undef  MC__ODEBND_GSL_PM_DEBUG
#undef  MC__ODEBND_GSL_SAMPLE_DEBUG
#undef  MC__ODEBND_GSL_LINTR_DEBUG
#define MC__ODEBND_GSL_LINTR_PMJAC
#undef  MC__ODEBND_GSL_ELLTR_SCALED
#undef  MC__ODEBND_GSL_DINEQI_DEBUG

#include <stdexcept>
#include <cassert>
#include <cmath>
#include <fstream>
#include <vector>
#include <sys/time.h>

#include "ellipsoid.hpp"
#include "tmodel.hpp"
#include "cmodel.hpp"

#include "odeslv_gsl.hpp"

namespace mc
{
//! @brief C++ class computing enclosures of the reachable set of parametric ODEs using continuous-time set-valued integration.
////////////////////////////////////////////////////////////////////////
//! mc::ODEBND_GSL is a C++ class that computes enclosures of the
//! reachable set of parametric ordinary differential equations
//! (ODEs) using continuous-time set-valued integration. It implements
//! the methods of differential inequalities, whereby polynomial models
//! with interval or ellipsoidal remainders are used to enable high-
//! order convergence. The use of ellipsoidal remainders enables
//! stability of the enclosures for asymptotically stable ODE systems
//! when the parameter host is sufficiently small. The numerical
//! integrator is gsl_odeiv2 in GSL.
////////////////////////////////////////////////////////////////////////
template <typename T, typename PMT=mc::TModel<T>, typename PVT=mc::TVar<T> >
class ODEBND_GSL: public BASE_GSL
{
private:
  typedef Ellipsoid E;

  //! @brief GSL data type for ODE system (auxiliary ODEs: differential inequalities)
  gsl_odeiv2_system _sys_dineqI;

  //! @brief GSL driver for ODE system (auxiliary ODEs: differential inequalities)
  gsl_odeiv2_driver *_driver_dineqI;

  //! @brief GSL data type for ODE system (auixiliary ODEs: differential inequalities w/ polynomial models)
  gsl_odeiv2_system _sys_dineqPM;

  //! @brief GSL driver for ODE system (auixiliary ODEs: differential inequalities w/ polynomial models)
  gsl_odeiv2_driver *_driver_dineqPM;

  //! @brief ODE solver based on GSL
  ODESLV_GSL<T>* _ODESLV_GSL;

  //! @brief list of operations in RHS evaluation
  std::list<const FFOp*> _opRHS;

  //! @brief array of list of operations in individual RHS evaluations
  std::list<const FFOp*> *_opRHSi;

  //! @brief list of operations in RHS Jacobian
  std::list<const FFOp*> _opJAC;

  //! @brief list of operations in IC evalution
  std::list<const FFOp*> _opIC;

  //! @brief preallocated array for evaluation of RHS function in T arithmetic
  T* _IRHS;

  //! @brief preallocated array for evaluation of RHS Jacobian in T arithmetic
  T* _IJAC;

  //! @brief preallocated array for evaluation of RHS function in PM arithmetic
  PVT* _PMRHS;

  //! @brief preallocated array for evaluation of RHS Jacobian in PM arithmetic
  PVT* _PMJAC;

  //! @brief const pointer to RHS function in current stage of ODE system
  const FFVar* _pRHS;

  //! @brief const pointer to RHS Jacobian in current stage of ODE system
  const FFVar* _pJAC;

  //! @brief const pointer to IC function in current stage of ODE system
  const FFVar* _pIC;

  //! @brief number of variables for DAG evaluation
  unsigned int _nVAR;

  //! @brief pointer to variables for DAG evaluation
  FFVar* _pVAR;

  //! @brief pointer of variable bounds for DAG evaluation
  T *_IVAR;

  //! @brief pointer to state interval bounds **DO NOT FREE**
  T *_Ix;

  //! @brief pointer to parameter interval bounds **DO NOT FREE**
  T *_Ip;

  //! @brief polynomial model environment
  PMT *_PMenv;

  //! @brief pointer of variable bounds for DAG evaluation
  PVT *_PMVAR;

  //! @brief state polynomial model **DO NOT FREE**
  PVT *_PMx;

  //! @brief parameter polynomial model **DO NOT FREE**
  PVT *_PMp;

  //! @brief state derivative polynomial model
  PVT *_PMxdot;

  //! @brief state PM remainder lower bound time derivatives
  double *_RxLdot;

  //! @brief state PM remainder upper bound time derivatives
  double *_RxUdot;

  //! @brief parameter reference
  double *_pref;

  //! @brief state reference
  double *_xref;

  //! @brief linear transformation A matrix
  double *_A;

  //! @brief linear transformation B matrix
  double *_B;

  //! @brief linear transformed state interval bounds
  T *_Id;

  //! @brief RHS Jacobian interval bounds
  T *_Idfdx;

  //! @brief linear transformed state ellipsoidal bounds
  E _Ed;

  //! @brief shape matrix (lower triangular) in ellipsoidal bounds
  double *_Q;

  //! @brief state reference time derivatives
  double *_xrefdot;

  //! @brief linear transformation B matrix time derivatives
  double *_Bdot;

  //! @brief rotated state interval bounds time derivatives
  T *_Iddot;

  //! @brief state interval bounds time derivatives
  T *_Ixdot;

  //! @brief shape matrix time directives in ellipsoidal bounds
  double *_Qdot;

  //! @brief polynomial model environment for mean-value theorem in (X,P)
  PMT *_MVXPenv;

  //! @brief rotated state polynomial model
  PVT *_MVXPd;

  //! @brief RHS polynomial model
  PVT *_MVXPf;

  //! @brief RHS Jacobian polynomial model
  PVT *_MVXPdfdx;

  //! @brief polynomial models for variables in mean-value theorem
  PVT *_MVXPVAR;

  //! @brief pointer to state polynomial models **DO NOT FREE**
  PVT *_MVXPx;

  //! @brief pointer to parameter polynomial models **DO NOT FREE**
  PVT *_MVXPp;

  //! @brief polynomial model environment for linear transformation of intial value
  PMT *_MVPenv;

  //! @brief array of parameter polynomial model for initial value
  PVT *_MVPp;

  //! @brief array of initial state polynomial model for initial value
  PVT *_MVPx;

public:
  /** @defgroup ODEBND_GSL Continuous-time set-valued integration of parametric ODEs
   *  @{
   */
  //! @brief Default constructor
  ODEBND_GSL();

  //! @brief Virtual destructor
  virtual ~ODEBND_GSL();

  //! @brief Integration results at a given time instant
  struct Results
  {
    //! @brief Constructors
    Results
      ( const double tk, const unsigned int nxk, const T*Ixk ):
      t( tk ), nx( nxk )
      { X = new T[nx];
        for( unsigned int ix=0; ix<nx; ix++ ) X[ix] = Ixk[ix]; }
    Results
      ( const double tk, const unsigned int nxk, const PVT*PMxk ):
      t( tk ), nx( nxk )
      { X = new T[nx];
        for( unsigned int ix=0; ix<nx; ix++ ) X[ix] = PMxk[ix].B(); }
    Results
      ( const Results&res ):
      t( res.t ), nx( res.nx )
      { X = new T[nx];
        for( unsigned int ix=0; ix<nx; ix++ ) X[ix] = res.X[ix]; }
    //! @brief Destructor
    ~Results()
      { delete[] X; }
    //! @brief Time instant
    double t;
    //! @brief Solution dimension
    unsigned int nx;
    //! @brief Solution bounds
    T* X;
  };

  //! @brief Integrator options
  struct Options: public BASE_GSL::Options
  {
    //! @brief Constructor
    Options():
      BASE_GSL::Options(), WRAPMIT(ELLIPS), ORDMIT(2), PMOPT(typename PMT::Options()),
      ODESLVOPT(typename ODESLV_GSL<T>::Options()), QTOL(machprec()), USEINV(true),
      DISPLAY(1), RESRECORD(false)
      {}
    //! @brief Assignment operator
    Options& operator=
      ( Options&options ){
        BASE_GSL::Options::operator=(options);
        WRAPMIT   = options.WRAPMIT;
        ORDMIT    = options.ORDMIT;
        PMOPT     = options.PMOPT;
        ODESLVOPT = options.ODESLVOPT;
        QTOL      = options.QTOL;
        USEINV    = options.USEINV;
        DISPLAY   = options.DISPLAY;
        RESRECORD = options.RESRECORD;
        return *this;
      }
    //! @brief Enumeration of wrapping mitigation strategies
    enum WRAPPING_STRATEGY{
      NONE=0,		//!< No wrapping mitigation
      DINEQ,		//!< Differential inequality contractor
      ELLIPS		//!< Ellipsoidal contractor with linear preconditioning [Default]
    };
    //! @brief Wrapping mitigation strategy
    WRAPPING_STRATEGY WRAPMIT;
    //! @brief Order of wrapping mitigation strategy (Default: 2)
    unsigned int ORDMIT;
    //! @brief Options of polynomial model arithmetic for wrapping mitigation
    typename PMT::Options PMOPT;
    //! @brief Options of non-validated ODE solver for reachable set approximation
    typename ODESLV_GSL<T>::Options ODESLVOPT;
    //! @brief Tolerance when dividing by trace of shape matrix in ellipsoidal bounds (Default: machprec())
    double QTOL;
    //! @brief Whether or not to use the specified invariants for bounds contraction (default: true)
    bool USEINV;
    //! @brief Display level (default: 1)
    int DISPLAY;
    //! @brief Whether or not to record results (default: false)
    bool RESRECORD;
  } options;

  //! @brief Structure for setting up storing the solver exceptions
  class Exceptions
  {
  public:
    //! @brief Enumeration type for ODEBND_GSL exception handling
    enum TYPE{
      REINIT=1,	//!< Error due to state reinitialization at intermediate stage
      UNDEF=-33	//!< Error due to calling a function/feature not yet implemented
    };
    //! @brief Constructor for error <a>ierr</a>
    Exceptions( TYPE ierr ) : _ierr( ierr ){}
    //! @brief Inline function returning the error flag
    int ierr(){ return _ierr; }
    //! @brief Inline function returning the error description
    std::string what(){
      switch( _ierr ){
      case REINIT:
        return "ODEBND_GSL::Exceptions  State reinitialization at intermediate stage not allowed";
      case UNDEF: default:
        return "ODEBND_GSL::Exceptions  Calling a feature not yet implemented";
       }
    }
  private:
    TYPE _ierr;
  };

  //! @brief Statistics for state bounds integration
  Stats stats_traj;

  //! @brief Computes interval enclosure of reachable set of parametric ODEs
  STATUS bounds
    ( const unsigned int ns, const double*tk, const T*Ip, T**Ixk,
      std::ostream&os=std::cout );

  //! @brief Computes polynomial model enclosure of reachable set of parametric ODEs
  STATUS bounds
    ( const unsigned int ns, const double*tk, const PVT*PMp, PVT**PMxk,
      E*Exk=0, std::ostream&os=std::cout );

  //! @brief Computes Hausdorff distance between interval enclosure and actual reachable set of parametric ODEs, using parameter sampling
  STATUS hausdorff
    ( const unsigned int ns, const double*tk, const T*Ip, double**Hxk,
      const unsigned int nsamp, std::ostream&os=std::cout );

  //! @brief Computes Hausdorff distance between polynomial model remainder enclosure and actual remainder function range, using parameter sampling
  STATUS hausdorff
    ( const unsigned int ns, const double*tk, const PVT*PMp, double**Hxk,
      const unsigned int nsamp, std::ostream&os=std::cout );

  //! @brief Record results in file <a>bndrec</a>, with accuracy of <a>iprec</a> digits
  void record
    ( std::ofstream&bndrec, const unsigned int iprec=5 ) const;

  //! @brief Return value of final time reached
  double final_time() const
    { return _t; };
  /** @} */

  //! @brief static pointer to class
  static ODEBND_GSL<T,PMT,PVT> *pODEBND_GSL;

private:

  //! @brief Vector storing interval bound results (upon request only)
  std::vector< Results > _results;

  //! @brief Function to finalize statistics for GSL
  void _GSL_term();

  //! @brief Function to initialize GSL driver
  void _GSL_init
    ( gsl_odeiv2_system &sys, gsl_odeiv2_driver *&driver );

  //! @brief Function to bound the remainder function relative to a polynomial model at sampling points -- return value is status
  STATUS _remainders
    ( const unsigned int ns, const double*tk, const T*Ip, const PVT*const*PMxk,
      T**Rxk, const unsigned int nsamp, std::ostream&os=std::cout );

  //! @brief Recrusive function computing bounds on errors between solutions of IVP in ODEs and polynomial approximant using sampling
  STATUS _remainders
    ( const unsigned int ns, const double*tk, const T*Ip, const PVT*const*PMxk,
      T**Rxk, const unsigned int nsamp, unsigned int* vsamp,
      const unsigned int ip, double*p, double**xk, std::ostream&os );

  //! @brief Function to initialize GSL for state interval bounds
  void _GSL_init
    ( const T*Ip );

  //! @brief Function to initialize state interval bounds
  void _init
    ( T*Ix0 );

  //! @brief Function to set RHS pointer
  bool _set_RHS_dineqI
    ( const unsigned int iRHS );

  //! @brief Function converting interval bounds to GSL array
  void _I2vec
    ( const double*dL, const double*dU, double*vec );

  //! @brief Function converting interval bounds to GSL array
  void _I2vec
    ( const T*Id, double*vec );

  //! @brief Function converting ellipsoidal bounds to GSL array
  void _E2vec
    ( const double*xref, const double*Q, const double*B, double*vec );

  //! @brief Function converting GSL array to interval bounds
  void _vec2I
    ( const double*vec, T*Ix );

  //! @brief Function converting GSL array to ellipsoidal bounds
  void _vec2E
    ( const double*vec, T*Ix );

  //! @brief Function converting rotated variables into original coordinates
  template<typename U> void _ep2x
    ( const U*d, const U*p, U*x ) const;

  //! @brief Function converting ellipsoids into interval bounds
  void _ep2x
    ( const double*Q, E&Ed, T*Id, const T*Ip, T*Ix ) const;

  //! @brief Static wrapper to function to calculate the DINEQs RHS values
  static int MC_GSLRHSI__
    ( double t, const double* x, double* xdot, void* user_data );

  //! @brief Function to calculate the DINEQs RHS values
  int _ODEBND_GSL_RHSI
    ( double t, const double* x, double* xdot, void* user_data );

  //! @brief Static wrapper to function to calculate the DINEQs RHS derivatives
  static int MC_GSLJACI__
    ( double t, const double* x, double* jac, double* xdot, void* user_data );

  //! @brief Function to calculate the DINEQs RHS derivatives
  int _ODEBND_GSL_JACI
    ( double t, const double* x, double* jac, double* xdot, void* user_data );

  //! @brief Function to initialize GSL for state polynomial models
  bool _GSL_init
    ( const PVT*PMp );

  //! @brief Function to initialize state polynomial model
  void _init
    ( PVT*PMx0 );

  //! @brief Function to set RHS pointer
  bool _set_RHS_dineqPM
    ( const unsigned int iRHS );

  //! @brief Function converting polynomial model with interval remainder to GSL array
  void _PMI2vec
    ( const PVT*PM, const double*RxL, const double*RxU, double*vec );

  //! @brief Function converting polynomial model with interval remainder to GSL array
  void _PMI2vec
    ( const PVT*PM, const T*D, double*vec );

  //! @brief Function converting polynomial model with interval remainder to GSL array
  void _PMI2vec
    ( const PVT*PM, double*vec );

  //! @brief Function converting polynomial model with ellipsoidal remainder to GSL array
  void _PME2vec
    ( const PVT*PM, const double*Q, double*vec );

  //! @brief Function converting GSL array to polynomial model with interval remainder
  void _vec2PMI
    ( const double*vec, PVT*PM );

  //! @brief Function converting GSL array to polynomial model with ellipsoidal remainder
  void _vec2PME
    ( const double*vec, PVT*PM );

  //! @brief Function converting rotated interval remainder bound back into original coordinates
  template<typename U> void _d2x
    ( const U*d, U*x, const bool reinit=true ) const;

  //! @brief Function converting rotated interval remainder bound back into original coordinates
  template<typename U> void _e2x
    ( const U*d, U*x, const bool reinit=true ) const;

  //! @brief Function converting ellipsoidal remainder bound into interval remainder bound
  void _e2x
    ( const double*Q, E&Ed, T*Id ) const;

  //! @brief Static wrapper to function to calculate the DINEQ-PMs RHS values
  static int MC_GSLRHSPM__
    ( double t, const double* x, double* xdot, void* user_data );

  //! @brief Function calculating the DINEQ-PMs RHS values
  int _ODEBND_GSL_RHSPM
    ( double t, const double* x, double* xdot, void* user_data );

  //! @brief Function propagating enclosures in PM arithmetic with ellipsoidal remainder -- approximation using mean-value theorem and interval analysis
  void _PROP_PM_ELL0
    ( const double t );

  //! @brief Function propagating enclosures in PM arithmetic with ellipsoidal remainder -- approximation using mean-value theorem and PM arithmetic
  void _PROP_PM_ELL1
    ( const double t );

  //! @brief Function propagating enclosures in PM arithmetic with ellipsoidal remainder -- joint polynomial model in states and parameters
  void _PROP_PM_ELL2
    ( const double t );

  //! @brief Static wrapper to function to calculate the DINEQ-PMs RHS derivatives
  static int MC_GSLJACPM__
    ( double t, const double* x, double* jac, double* xdot, void* user_data );

  //! @brief Function to calculate the DINEQ-PMs RHS jacobian
  int _ODEBND_GSL_JACPM
    ( double t, const double* x, double* jac, double* xdot, void* user_data );

  //! @brief Function to display intermediate results
  template<typename U> void _print_interm
    ( const double t, const U*x, const std::string&var, std::ostream&os=std::cout ) const;

  //! @brief Function to display intermediate results
  template<typename U, typename V> void _print_interm
    ( const double t, const U*x, const V&r, const std::string&var, std::ostream&os=std::cout ) const;

  //! @brief Position in symmetric matrix stored in lower triangular form
  unsigned int _ndxLT
    ( const unsigned int i, const unsigned int j ) const
    { return( i<j? _ndxLT(j,i): i+j*_nx-j*(j+1)/2 ); }

  //! @brief Private methods to block default compiler methods
  ODEBND_GSL(const ODEBND_GSL&);
  ODEBND_GSL& operator=(const ODEBND_GSL&);
};

template <typename T, typename PMT, typename PVT>
 ODEBND_GSL<T,PMT,PVT>* ODEBND_GSL<T,PMT,PVT>::pODEBND_GSL = 0;

template <typename T, typename PMT, typename PVT> inline
ODEBND_GSL<T,PMT,PVT>::ODEBND_GSL
()
: BASE_GSL(), _pRHS(0), _pJAC(0), _pIC(0), _nVAR(0), _pVAR(0)
{
  // Initalize state/parameter arrays
  _opRHSi = 0;
  _IRHS = _IJAC = _IVAR = _Ix = _Ip = 0;
  _PMRHS = _PMJAC = _PMVAR = _PMp = _PMx = _PMxdot = 0;
  _PMenv = 0;

  // Initialize parameterization arrays
  _pref = _xref = _xrefdot = 0;
  _A = _B = _Bdot = _Q = _Qdot = 0;
  _Id = _Idfdx = _Iddot = _Ixdot = 0;
  _RxLdot = _RxUdot = 0;

  // Initialize polynomial model environments
  _MVXPenv = _MVPenv = 0;
  _MVXPd = _MVXPf = _MVXPdfdx = _MVXPVAR = _MVXPx = _MVXPp = _MVPp = _MVPx = 0;

  // Initialize ellipsoidal calculus
  E::options.PSDCHK = false;
     
  // Initalize GSL
  _driver_dineqI = 0;
  _driver_dineqPM = 0;
  _ODESLV_GSL = new ODESLV_GSL<T>();
}

template <typename T, typename PMT, typename PVT> inline
ODEBND_GSL<T,PMT,PVT>::~ODEBND_GSL
()
{
  /* DO NOT FREE _pRHS */
  delete[] _opRHSi;
  delete[] _IRHS;
  delete[] _PMRHS;
  delete[] _IJAC;
  delete[] _PMJAC;
  delete[] _pJAC;
  delete[] _pVAR;

  // Free state arrays -- Do *NOT* delete _p _Ip _PMp _PMenv
  delete[] _IVAR;   // **DO NOT DELETE _Ix, _Ip**
  delete[] _PMVAR;  // **DO NOT DELETE _PMx, _PMp**
  delete[] _PMxdot;
  delete[] _RxLdot;
  delete[] _RxUdot;
      
  // Free linear transformation arrays
  delete[] _pref;
  delete[] _xref;
  delete[] _A;
  delete[] _B;
  delete[] _Id;
  delete[] _Idfdx;
  delete[] _Q;
  delete[] _xrefdot;
  delete[] _Bdot;
  delete[] _Iddot;
  delete[] _Ixdot;
  delete[] _Qdot;
  delete[] _MVXPd;
  delete[] _MVXPf;
  delete[] _MVXPdfdx;
  delete[] _MVXPVAR;  // **DO NOT DELETE _MVXPx, _MVXPp**
  delete   _MVXPenv;
  delete[] _MVPp;
  delete[] _MVPx;
  delete   _MVPenv;
      
  // Free GSL arrays
  if( _driver_dineqI )  gsl_odeiv2_driver_free( _driver_dineqI );
  if( _driver_dineqPM ) gsl_odeiv2_driver_free( _driver_dineqPM );
  delete _ODESLV_GSL;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_GSL_init
( gsl_odeiv2_system &sys, gsl_odeiv2_driver *&driver )
{
  // Set GSL driver
  if( driver ) gsl_odeiv2_driver_free( driver );
  switch( options.INTMETH ){
  case Options::RK8PD:
    driver = gsl_odeiv2_driver_alloc_y_new( &sys, gsl_odeiv2_step_rk8pd,
      options.H0, options.ATOL, options.RTOL );
    break;
  case Options::MSADAMS:
    driver = gsl_odeiv2_driver_alloc_y_new( &sys, gsl_odeiv2_step_msadams,
      options.H0, options.ATOL, options.RTOL );
    break;
  case Options::MSBDF:
    driver = gsl_odeiv2_driver_alloc_y_new( &sys, gsl_odeiv2_step_msbdf,
      options.H0, options.ATOL, options.RTOL );
    break;
  case Options::RKF45: default:
    driver = gsl_odeiv2_driver_alloc_y_new( &sys, gsl_odeiv2_step_rkf45,
      options.H0, options.ATOL, options.RTOL );
    break;
  }
  gsl_odeiv2_driver_set_hmin( driver, options.HMIN );  
  gsl_odeiv2_driver_set_nmax( driver, options.NMAX );  

  // Set state storage
  delete [] _vec_state;
  _vec_state = new double[ sys.dimension ];
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_GSL_term()
{
  // Get final CPU time
  _final_stats( stats_traj );
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_vec2I
( const double*vec, T*Ix )
{
  for( unsigned int ix=0, ivec=0; ix<_nx; ix++ ){
    Ix[ix] = T( vec[ivec], vec[ivec+1] );
    ivec+=2;
  }
  return;
}

template <typename T, typename PMT, typename PVT> template<typename U> inline void
ODEBND_GSL<T,PMT,PVT>::_ep2x
( const U*d, const U*p, U*x ) const
{
  for( unsigned int ix=0; ix<_nx; ix++ ){   
    x[ix] = _xref[ix] + d[ix];
    for( unsigned int jp=0; jp<_np; jp++ )
      x[ix] += ( p[jp] - _pref[jp] ) * _B[jp*_nx+ix];
  }
  return;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_ep2x
( const double*Q, E&Ed, T*Id, const T*Ip, T*Ix ) const
{
  Ed.set( _nx, Q );
#ifdef MC__ODEBND_GSL_LINTR_DEBUG
  std::cout << "Ed =" << Ed << std::endl;
#endif
  for( unsigned int ix=0; ix<_nx; ix++ )
    Id[ix] = T( Ed.l(ix), Ed.u(ix) );
  _ep2x( Id, Ip, Ix );

#ifdef MC__ODEBND_GSL_LINTR_DEBUG
  for( unsigned int ix=0; ix<_nx; ix++ )
    std::cout << "Id[" << ix << "] = " << _Id[ix] << std::endl;
  for( unsigned int ix=0; ix<_nx; ix++ )
    std::cout << "Ix[" << ix << "] = " << Ix[ix] << std::endl;
  { int dum; std::cin >> dum; }
#endif
  return;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_vec2E
( const double*vec, T*Ix )
{
  unsigned int ivec = 0;
  for( unsigned int ix=0; ix<_nx; ix++ )
    _xref[ix] = vec[ivec++];
  for( unsigned int iQ=0; iQ<_nx*(_nx+1)/2; iQ++ )
    _Q[iQ] = vec[ivec++];
  for( unsigned int iB=0; iB<_nx*_np; iB++ )
    _B[iB] = vec[ivec++];
  return _ep2x( _Q, _Ed, _Id, _Ip, Ix );
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_I2vec
( const double*xL, const double*xU, double*vec )
{
  for( unsigned int ix=0, ivec=0; ix<_nx; ix++ ){
    vec[ivec++] = xL[ix];
    vec[ivec++] = xU[ix];
  }
  return;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_I2vec
( const T*Ix, double*vec )
{
  for( unsigned int ix=0, ivec=0; ix<_nx; ix++ ){
    vec[ivec++] = Op<T>::l( Ix[ix] );
    vec[ivec++] = Op<T>::u( Ix[ix] );
  }
  return;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_E2vec
( const double*xref, const double*Q, const double*B, double*vec )
{
  unsigned int ivec = 0;
  for( unsigned int ix=0; ix<_nx; ix++ )
    vec[ivec++] = xref[ix];
  for( unsigned int iQ=0; iQ<_nx*(_nx+1)/2; iQ++ )
    vec[ivec++] = Q[iQ];
  for( unsigned int iB=0; iB<_nx*_np; iB++ )
    vec[ivec++] = B[iB];
  return;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_init
( T*Ix0 )
{
  switch( options.WRAPMIT){

  case Options::NONE:
  case Options::DINEQ:
    _pDAG->eval( _opIC, _nx, _pIC, Ix0, _np, _pVAR+_nx, _Ip );
    return _I2vec( Ix0, _vec_state );

  case Options::ELLIPS:
  default:
    for( unsigned int ip=0; ip<_np; ip++ ){
      _pref[ip] = Op<T>::mid( _Ip[ip] );
      _MVPp[ip].set( _MVPenv, ip, _Ip[ip] );
    }
    _pDAG->eval( _opIC, _nx, _pIC, _MVPx, _np, _pVAR+_nx, _MVPp );

    double trR2 = 0.;
    for( unsigned int ix=0, iQ=0; ix<_nx; ix++ ){
      _xref[ix] = _MVPx[ix].constant();
      for( unsigned int jp=0; jp<_np; jp++ ) _B[jp*_nx+ix] = _MVPx[ix].linear( jp );
      Ix0[ix] = _MVPx[ix].bound();
      trR2 += sqr( Op<T>::diam( _MVPx[ix].remainder() ) / 2. );
      _Q[iQ++] = Op<T>::diam( _MVPx[ix].remainder() ) / 2.;
      for( unsigned int jx=ix+1; jx<_nx; jx++ ) _Q[iQ++] = 0.;
    }
    for( unsigned int ix=0, iQ=0; ix<_nx; ix++, iQ+=_nx-ix ) _Q[iQ] *= std::sqrt( trR2 );
#ifdef MC__ODEBND_GSL_LINTR_DEBUG
    E Ex0( _nx, _Q, _xref );
    std::cout << "Ex0 =" << Ex0 << std::endl;
#endif
    return _E2vec( _xref, _Q, _B, _vec_state );
  }
}

template <typename T, typename PMT, typename PVT> inline int
ODEBND_GSL<T,PMT,PVT>::_ODEBND_GSL_RHSI
( double t, const double* x, double* xdot, void* user_data )
{
  if( !_pRHS ) return GSL_EBADFUNC;
  stats_traj.numRHS++;

  switch( options.WRAPMIT){
  case Options::NONE:
    _vec2I( x, _Ix );   // set current state bounds
    _IVAR[_nx+_np] = t; // set current time
    _pDAG->eval( _opRHS, _IRHS, _nx, _pRHS, _Ixdot, _nVAR, _pVAR, _IVAR );
    _I2vec( _Ixdot, xdot );
    return GSL_SUCCESS;  
   
  case Options::DINEQ:
    if( !_opRHSi ) return GSL_EBADFUNC;
    _vec2I( x, _Ix );   // set current state bounds
    _IVAR[_nx+_np] = t; // set current time
    for( unsigned int ix=0, ivec=0; ix<_nx; ix++ ){
      T Ixi = _Ix[ix];
      _Ix[ix] = Op<T>::l( Ixi );
      _pDAG->eval( _opRHSi[ix], _IRHS, 1, _pRHS+ix, _Ixdot+ix, _nVAR, _pVAR, _IVAR );
      xdot[ivec++] = Op<T>::l( _Ixdot[ix] );
      _Ix[ix] = Op<T>::u( Ixi );
      _pDAG->eval( _opRHSi[ix], _IRHS, 1, _pRHS+ix, _Ixdot+ix, _nVAR, _pVAR, _IVAR );
      xdot[ivec++] = Op<T>::u( _Ixdot[ix] );
      _Ix[ix] = Ixi;
    }
    return GSL_SUCCESS;  
   
  case Options::ELLIPS:{
  default:
    _vec2E( x, _Ix );   // set current state enclosure
#ifdef MC__ODEBND_GSL_DINEQI_DEBUG
    std::cout << "@t=" << t << std::endl;
    //E::options.PSDCHK = true;
    for( unsigned int ix=0; ix<_nx; ix++ )
      std::cout << "xref[" << ix << "] = " << _xref[ix] << std::endl;
    for( unsigned int ix=0; ix<_nx; ix++ ){
      std::cout << "B[" << ix << ",#] = ";
      for( unsigned int jp=0; jp<_np; jp++ )
        std::cout << _B[jp*_np+ix] << "  ";
      std::cout << std::endl;
    }
    std::cout << "Ed = " << _Ed << std::endl;
    { int dum; std::cin >> dum; }
#endif
    double trQ = 0.;
    for( unsigned int ix=0; ix<_nx; ix++ )
      trQ += ( _Q[_ndxLT(ix,ix)]>0? _Q[_ndxLT(ix,ix)]: machprec() );

    // Compute polynomial expansion of ODE's RHS
    for( unsigned int jx=0; jx<_nx; jx++ )
      _MVXPd[jx].set( _MVXPenv, jx, _Id[jx] );
    for( unsigned int jp=0; jp<_np; jp++ )
      _MVXPp[jp].set( _MVXPenv, _nx+jp, _Ip[jp] );
    _ep2x( _MVXPd, _MVXPp, _MVXPx );
#ifdef MC__ODEBND_GSL_DINEQI_DEBUG
    for( unsigned int ix=0; ix<_nx; ix++ )
      std::cout << "MVXPx[ " << ix << "] = " << _MVXPx[ix] << std::endl;
    { int dum; std::cin >> dum; }
#endif
    _MVXPVAR[_nx+_np] = t; // set current time   
    _pDAG->eval( _opRHS, _PMRHS, _nx, _pRHS, _MVXPf, _nVAR, _pVAR, _MVXPVAR );
/*
    // Extract time derivatives of constant, linear and remainder parts
    double sumkappa = 0.;
    for( unsigned int ix=0; ix<_nx; ix++ ){
#ifdef MC__ODEBND_GSL_DINEQI_DEBUG
      std::cout << "MVXPf[" << ix << "] = " << _MVXPf[ix] << std::endl;
#endif
      _xrefdot[ix] = _MVXPf[ix].constant() + Op<T>::mid( _MVXPf[ix].remainder() );
      _Iddot[ix] = _MVXPf[ix].remainder() - Op<T>::mid( _MVXPf[ix].remainder() );
      for( unsigned int iord=2; iord<=_MVXPenv->nord(); iord++ ){
        _xrefdot[ix] += Op<T>::mid( _MVXPf[ix].bndord( iord ) );
        _Iddot[ix] += _MVXPf[ix].bndord( iord )
                    - Op<T>::mid( _MVXPf[ix].bndord( iord ) );
      }
#ifdef MC__ODEBND_GSL_DINEQI_DEBUG
      std::cout << "xrefdot[" << ix << "] = " << _xrefdot[ix] << std::endl;
      std::cout << "Iddot[" << ix << "] = " << _Iddot[ix]
                << " : " << Op<T>::mid(_Iddot[ix]) << std::endl;
      std::cout << "kappa[" << ix << "] = "
                << Op<T>::diam( _Iddot[ix] ) / 2.
                 / ( std::sqrt( trQ ) + options.QTOL ) << std::endl;
#endif
      sumkappa += ( Op<T>::diam( _Iddot[ix] ) / 2. )
                / ( std::sqrt( trQ ) + options.QTOL );
      for( unsigned int jx=0; jx<_nx; jx++ )
        _A[ix+jx*_nx] = _MVXPf[ix].linear(jx);
#ifdef MC__ODEBND_GSL_DINEQI_DEBUG
      std::cout << "A[" << ix << ",#] = ";
      for( unsigned int jx=0; jx<_nx; jx++ )
        std::cout << _A[ix+jx*_nx] << "  ";
      std::cout << std::endl;
#endif
      for( unsigned int jp=0; jp<_np; jp++ )
        _Bdot[ix+jp*_nx] = _MVXPf[ix].linear(_nx+jp);
#ifdef MC__ODEBND_GSL_DINEQI_DEBUG
      std::cout << "Bdot[" << ix << ",#] = ";
      for( unsigned int jp=0; jp<_np; jp++ )
        std::cout << _Bdot[ix+jp*_nx] << "  ";
      std::cout << std::endl;
#endif
    }
*/
    // Extract time derivatives of constant, linear and remainder parts
    //sumkappa = 0.;
    double sumkappa = 0.;
    for( unsigned int ix=0; ix<_nx; ix++ ){
#ifdef MC__ODEBND_GSL_DINEQI_DEBUG
      std::cout << "MVXPf[" << ix << "] = " << _MVXPf[ix] << std::endl;
#endif
      for( unsigned int jx=0; jx<_nx; jx++ )
        _A[ix+jx*_nx] = _MVXPf[ix].linear(jx,true);
#ifdef MC__ODEBND_GSL_DINEQI_DEBUG
      std::cout << "A[" << ix << ",#] = ";
      for( unsigned int jx=0; jx<_nx; jx++ )
        std::cout << _A[ix+jx*_nx] << "  ";
      std::cout << std::endl;
#endif
      for( unsigned int jp=0; jp<_np; jp++ )
        _Bdot[ix+jp*_nx] = _MVXPf[ix].linear(_nx+jp,true);
#ifdef MC__ODEBND_GSL_DINEQI_DEBUG
      std::cout << "Bdot[" << ix << ",#] = ";
      for( unsigned int jp=0; jp<_np; jp++ )
        std::cout << _Bdot[ix+jp*_nx] << "  ";
      std::cout << std::endl;
#endif
#ifdef MC__ODEBND_GSL_DINEQI_DEBUG
      std::cout << "MVXPf[" << ix << "] = " << _MVXPf[ix] << std::endl;
#endif
      T Rfi = _MVXPf[ix].B();
      _xrefdot[ix] = Op<T>::mid(Rfi);
      _Iddot[ix] = Rfi - Op<T>::mid(Rfi);
#ifdef MC__ODEBND_GSL_DINEQI_DEBUG
      std::cout << "xrefdot[" << ix << "] = " << _xrefdot[ix] << std::endl;
      std::cout << "Iddot[" << ix << "] = " << _Iddot[ix]
                << " : " << Op<T>::mid(_Iddot[ix]) << std::endl;
      std::cout << "kappa[" << ix << "] = "
                << Op<T>::diam( _Iddot[ix] ) / 2.
                 / ( std::sqrt( trQ ) + options.QTOL ) << std::endl;
#endif
      sumkappa += ( Op<T>::diam( _Iddot[ix] ) / 2. )
                / ( std::sqrt( trQ ) + options.QTOL );
    }

    for( unsigned int jx=0; jx<_nx; jx++ ){
      for( unsigned int ix=jx; ix<_nx; ix++ ){
        _Qdot[_ndxLT(ix,jx)] = sumkappa * _Q[_ndxLT(ix,jx)];
        for( unsigned int kx=0; kx<_nx; kx++ )
          _Qdot[_ndxLT(ix,jx)] += _Q[_ndxLT(ix,kx)] * _A[jx+kx*_nx]
	                        + _A[ix+kx*_nx] * _Q[_ndxLT(kx,jx)];
      }
      _Qdot[_ndxLT(jx,jx)] += ( Op<T>::diam( _Iddot[jx] ) / 2. )
                            * ( std::sqrt( trQ ) + options.QTOL );
    }
#ifdef MC__ODEBND_GSL_DINEQI_DEBUG
    for( unsigned int ix=0; ix<_nx; ix++ ){
      std::cout << "Qdot[" << ix << ",#] = ";
      for( unsigned int jx=0; jx<=ix; jx++ )
        std::cout << _Qdot[_ndxLT(ix,jx)] << "  ";
      std::cout << std::endl;
    }
    E Exdot( _nx, _Qdot, _xrefdot );
    std::cout << "Exdot =" << Exdot << std::endl;
    { int dum; std::cin >> dum; }
#endif

    // Copy time derivative in GSL array
    _E2vec( _xrefdot, _Qdot, _Bdot, xdot );
    
    return GSL_SUCCESS;
   }
  }
}

template <typename T, typename PMT, typename PVT> inline int
ODEBND_GSL<T,PMT,PVT>::MC_GSLRHSI__
( double t, const double* x, double* xdot, void* user_data )
{
  ODEBND_GSL<T,PMT,PVT> *pODEBND_GSL = ODEBND_GSL<T,PMT,PVT>::pODEBND_GSL;
  int flag = pODEBND_GSL->_ODEBND_GSL_RHSI( t, x, xdot, user_data );
  ODEBND_GSL<T,PMT,PVT>::pODEBND_GSL = pODEBND_GSL;
  return flag;
}

template <typename T, typename PMT, typename PVT> inline int
ODEBND_GSL<T,PMT,PVT>::_ODEBND_GSL_JACI
( double t, const double* x, double* jac, double* xdot, void* user_data )
{
  // Jacobian not (yet) implemented -- Use RK methods
  return GSL_EBADFUNC;
}

template <typename T, typename PMT, typename PVT> inline int
ODEBND_GSL<T,PMT,PVT>::MC_GSLJACI__
( double t, const double* x, double* jac, double* xdot, void* user_data )
{
  ODEBND_GSL<T,PMT,PVT> *pODEBND_GSL = ODEBND_GSL<T,PMT,PVT>::pODEBND_GSL;
  int flag = pODEBND_GSL->_ODEBND_GSL_JACI( t, x, jac, xdot, user_data );
  ODEBND_GSL<T,PMT,PVT>::pODEBND_GSL = pODEBND_GSL;
  return flag;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_GSL_init
( const T* Ip )
{
  // Set GSL driver
  _sys_dineqI.function = MC_GSLRHSI__;
  _sys_dineqI.jacobian = MC_GSLJACI__;
  _sys_dineqI.params = 0;
  _sys_dineqI.dimension = 0;
  switch( options.WRAPMIT){
  case Options::NONE:
  case Options::DINEQ:
    _sys_dineqI.dimension += 2*_nx;
    break;
  case Options::ELLIPS:
  default:
    _sys_dineqI.dimension += _nx*(1+_np)+_nx*(_nx+1)/2;
    break;
  }
  _GSL_init( _sys_dineqI, _driver_dineqI );

  // Size and set DAG evaluation arrays
  _nVAR = _nx+_np+1;
  delete[] _pVAR; _pVAR = new FFVar[_nVAR];
  delete[] _IVAR; _IVAR = new T[_nVAR];
  for( unsigned ix=0; ix<_nx; ix++ ) _pVAR[ix] = _pX[ix];
  for( unsigned ip=0; ip<_np; ip++ ) _pVAR[_nx+ip] = _pP[ip];
  _pVAR[_nx+_np] = (_pT? *_pT: 0. );
  for( unsigned ip=0; ip<_np; ip++ ) _IVAR[_nx+ip] = Ip[ip];
  _Ix = _IVAR;
  _Ip = _IVAR + _nx;

  // Reset _MVXPVenv and related variables
  if( _MVXPenv && ( _MVXPenv->nord() != options.ORDMIT 
                 || _MVXPenv->nvar() != _nx+_np ) ){
    delete[] _MVXPf;   _MVXPf = 0;
    delete[] _MVXPd;   _MVXPd = 0;
    delete[] _MVXPVAR; _MVXPVAR = 0;
    delete   _MVXPenv; _MVXPenv = 0;
  }

  // Reset _MVPVenv and related variables
  if( _MVPenv && ( _MVPenv->nord() != 1 
                || _MVPenv->nvar() != _np ) ){
    delete[] _MVXPx;   _MVPx = 0;
    delete[] _MVXPp;   _MVPp = 0;
    delete   _MVPenv; _MVPenv = 0;
  }

  // Set parameterization variables
  switch( options.WRAPMIT){
  case Options::NONE:
  case Options::DINEQ:
    if( !_Ixdot )   _Ixdot   = new T[_nx];
    break;

  case Options::ELLIPS:
  default:
    if( !_pref )    _pref    = new double[_np];
    if( !_xref )    _xref    = new double[_nx];
    if( !_A )	    _A       = new double[_nx*_nx];
    if( !_B )	    _B       = new double[_nx*_np];
    if( !_Q )	    _Q       = new double[_nx*(_nx+1)/2];
    if( !_Id )      _Id      = new T[_nx];
    if( !_xrefdot ) _xrefdot = new double[_nx];
    if( !_Bdot )    _Bdot    = new double[_nx*_np];
    if( !_Qdot )    _Qdot    = new double[_nx*(_nx+1)/2];
    if( !_Iddot )   _Iddot   = new T[_nx];
    if( !_MVXPenv ) _MVXPenv = new PMT( _nx+_np, options.ORDMIT );
    _MVXPenv->options = options.PMOPT;
    if( !_MVXPd )   _MVXPd   = new PVT[_nx];
    if( !_MVXPf )   _MVXPf   = new PVT[_nx];
    if( !_MVXPVAR ) _MVXPVAR = new PVT[_nVAR];
    _MVXPp = _MVXPVAR + _nx;
    _MVXPx = _MVXPVAR;
    if( !_MVPenv )  _MVPenv  = new PMT( _np, 1 );
    _MVPenv->options = options.PMOPT;
    if( !_MVPp )    _MVPp    = new PVT[_np];
    if( !_MVPx )    _MVPx    = new PVT[_nx];
    break;
  }

  // Reset result record and statistics
  _results.clear();
  _init_stats( stats_traj );
}

template <typename T, typename PMT, typename PVT> inline bool
ODEBND_GSL<T,PMT,PVT>::_set_RHS_dineqI
( const unsigned int iRHS )
{
  if( _vRHS.size() <= iRHS ) return false;

  _pRHS = _vRHS.at( iRHS );
  _opRHS = _pDAG->subgraph( _nx, _pRHS );

  delete[] _IRHS;   _IRHS = 0;
  delete[] _PMRHS;  _PMRHS = 0;
  delete[] _opRHSi; _opRHSi = 0;
  switch( options.WRAPMIT){
  case Options::NONE:
    _IRHS = new T[ _opRHS.size() ]; break;
  case Options::DINEQ:
    _opRHSi = new std::list<const FFOp*>[_nx];
    for( unsigned ix=0; ix<_nx; ix++ )
      _opRHSi[ix] = _pDAG->subgraph( 1, _pRHS+ix );
    _IRHS = new T[ _opRHS.size() ]; break;   
  case Options::ELLIPS:
  default:
    _PMRHS = new PVT[ _opRHS.size() ]; break;
  }

  delete[] _pJAC;
  _pJAC = 0;
  _opJAC.clear();

  return true;
}

//! @fn template <typename T, typename PMT, typename PVT> inline typename ODEBND_GSL<T,PMT,PVT>::STATUS ODEBND_GSL<T,PMT,PVT>::bounds(
//! const unsigned int ns, const double*tk, const T*Ip, T**Ixk,
//! std::ostream&os )
//!
//! This function computes an interval enclosure of the reachable set of 
//! the parametric ODEs defined in IVP using equally spaced samples:
//!   - <a>ns</a> [input] number of time stages
//!   - <a>tk</a> [input] stage times, including the initial time
//!   - <a>Ip</a> [input] interval parameter set
//!   - <a>Ixk</a> [output] interval state enclosures at stage times
//!   - <a>os</a> [input] output stream
//! .
//! The return value is the status.
template <typename T, typename PMT, typename PVT>
inline typename ODEBND_GSL<T,PMT,PVT>::STATUS
ODEBND_GSL<T,PMT,PVT>::bounds
( const unsigned int ns, const double*tk, const T*Ip, T**Ixk, std::ostream&os )
{
  // Initialize trajectory integration with GSL
  _GSL_init( Ip );

  try{
    // Bounds on initial conditions
    if( !_vIC.size() )
      { _GSL_term(); return FATAL; }
    _pIC = _vIC.at(0);
    _opIC = _pDAG->subgraph( _nx, _pIC );
    _init( Ixk[0] );
    if( options.DISPLAY >= 1 ) _print_interm( tk[0], Ixk[0], "x", os );
    // Record initial results
    if( options.RESRECORD )
      _results.push_back( Results( tk[0], _nx, Ixk[0] ) );

    // Integrate ODEs through each stage using GSL
    _t = tk[0];
    _h = options.H0;
    pODEBND_GSL = this;

    for( _istg=0; _istg<ns; _istg++ ){
      if( _istg && _vIC.size()>1 ) throw Exceptions( Exceptions::REINIT );

      // update list of operations in RHS and JAC
      const unsigned int iRHS = ( _vRHS.size()<2? 0: _istg );
      if( (!_istg || iRHS) && !_set_RHS_dineqI( iRHS ) )
        { _GSL_term(); return FATAL; }

      while( _t < tk[_istg+1] ){
        if( gsl_odeiv2_evolve_apply( _driver_dineqI->e, _driver_dineqI->c, _driver_dineqI->s,
            &_sys_dineqI, &_t, tk[_istg+1], &_h, _vec_state ) != GSL_SUCCESS
         || _h < options.HMIN
         || (options.NMAX && stats_traj.numSteps > options.NMAX) )
          throw Exceptions( Exceptions::UNDEF );
        stats_traj.numSteps++;
        if( options.HMAX > 0 && _h > options.HMAX ) _h = options.HMAX;
      }

      switch( options.WRAPMIT){
      case Options::NONE:
      case Options::DINEQ:
        _vec2I( _vec_state, Ixk[_istg+1] ); break;
      case Options::ELLIPS:
      default:
        _vec2E( _vec_state, Ixk[_istg+1] ); break;
      }
      if( options.DISPLAY >= 1 ) _print_interm( tk[_istg+1], Ixk[_istg+1], "x", os );
      // Record intermediate results
      if( options.RESRECORD )
        _results.push_back( Results( tk[_istg+1], _nx, Ixk[_istg+1] ) );
    }
  }
  catch(...){
    _GSL_term();
    if( options.DISPLAY >= 1 ) _print_stats( stats_traj, os );
    return FAILURE;
  }

  _GSL_term();
  if( options.DISPLAY >= 1 ) _print_stats( stats_traj, os );
  return NORMAL;
}

//! @fn template <typename T, typename PMT, typename PVT> inline typename ODEBND_GSL<T,PMT,PVT>::STATUS ODEBND_GSL<T,PMT,PVT>::hausdorff(
//! const unsigned int ns, const double*tk, const T*Ip, double**Hxk,
//! const unsigned int nsamp, std::ostream&os )
//!
//! This function computes the Hausdorff distance between the interval enclosure
//! and the exact reachable set projected onto each variable:
//!   - <a>ns</a> [input] number of time stages
//!   - <a>tk</a> [input] stage times, including the initial time
//!   - <a>Ip</a> [input] interval parameter set
//!   - <a>Hxk</a> [output] Hausdorff distance between the interval enclosure
//!     and the exact reachable set projected onto each variable, at stage times
//!   - <a>nsamp</a> [input] number of samples for each parameter
//!   - <a>os</a> [input] output stream
//! .
//! The return value is the status.
template <typename T, typename PMT, typename PVT> inline typename ODEBND_GSL<T,PMT,PVT>::STATUS
ODEBND_GSL<T,PMT,PVT>::hausdorff
( const unsigned int ns, const double*tk, const T*Ip, double**Hxk,
  const unsigned int nsamp, std::ostream&os )
{
  int DISPLAY_ODEBND = options.DISPLAY;
  int DISPLAY_ODESLV = options.ODESLVOPT.DISPLAY;
  options.DISPLAY = 0;
  options.ODESLVOPT.DISPLAY = 0;
  _ODESLV_GSL->set( *this ); // in base class BASE_ODE
  _ODESLV_GSL->options = options.ODESLVOPT;
  _ODESLV_GSL->options.RESRECORD = false;

  // Compute exact bounds 
  T** Ixk0 = new T*[ns+1];
  for( unsigned int is=0; is<ns+1; is++ ) Ixk0[is] = new T[_nx];
  if( _ODESLV_GSL->bounds( ns, tk, Ip, Ixk0, 0, nsamp, os ) != NORMAL ){
    for( unsigned int is=0; is<ns+1; is++ ) delete[] Ixk0[is];
    delete[] Ixk0;
    return FAILURE;
  }

  // Compute approximate bounds
  T** Ixk = new T*[ns+1];
  for( unsigned int is=0; is<ns+1; is++ ) Ixk[is] = new T[_nx];
  try{ bounds( ns, tk, Ip, Ixk, os ); }
  catch(...){;}
  unsigned int nsf = _istg;
  for( unsigned int is=nsf; is<ns; is++ )
    for( unsigned int ix=0; ix<_nx; ix++ )
      Ixk[_istg+1][ix] = T( -1e20, 1e20 );

  // Compute Hausdorff metric at each time step
  struct loc{ static double hausdorff
    ( const T&Ix, const T&Ix0 )
    { return std::max( std::fabs(Op<T>::l(Ix)-Op<T>::l(Ix0)),
                       std::fabs(Op<T>::u(Ix)-Op<T>::u(Ix0)) ); }
  };

  options.DISPLAY = DISPLAY_ODEBND;
  options.ODESLVOPT.DISPLAY = DISPLAY_ODESLV;
  for( unsigned int is=0; is<ns+1; is++ ){
    for( unsigned int ix=0; ix<_nx; ix++ )
      Hxk[is][ix] = loc::hausdorff( Ixk[is][ix], Ixk0[is][ix] );
    if( options.DISPLAY >= 1 ) _print_interm( tk[is], Hxk[is], "dH", os );
  }

  for( unsigned int is=0; is<ns+1; is++ ) delete[] Ixk0[is];
  delete[] Ixk0;
  for( unsigned int is=0; is<ns+1; is++ ) delete[] Ixk[is];
  delete[] Ixk;

  return NORMAL;
}

template <typename T, typename PMT, typename PVT> template<typename U> inline void
ODEBND_GSL<T,PMT,PVT>::_d2x
( const U*d, U*x, const bool reinit ) const
{
  for( unsigned int ix=0; ix<_nx; ix++ ){
   if( reinit ) x[ix] = 0.;
    for( unsigned int jx=0; jx<_nx; jx++ )
      x[ix] += d[jx] * _A[jx*_nx+ix];
  }
  return;
}

template <typename T, typename PMT, typename PVT> template<typename U> inline void
ODEBND_GSL<T,PMT,PVT>::_e2x
( const U*d, U*x, const bool reinit ) const
{
  for( unsigned int ix=0; ix<_nx; ix++ ){   
    if( reinit ) x[ix] = 0.;
    x[ix] += d[ix];
  }
  return;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_e2x
( const double*Q, E&Ed, T*Id ) const
{
  Ed.set( _nx, _Q );
#ifdef MC__ODEBND_GSL_LINTR_DEBUG
  std::cout << "Ed =" << Ed << std::endl;
#endif

  for( unsigned int ix=0; ix<_nx; ix++ )
    Id[ix] = T( Ed.l(ix), Ed.u(ix) );

#ifdef MC__ODEBND_GSL_LINTR_DEBUG
  for( unsigned int ix=0; ix<_nx; ix++ )
    std::cout << "Id[" << ix << "] = " << Id[ix] << std::endl;
  { int dum; std::cin >> dum; }
#endif
  return;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_vec2PMI
( const double*vec, PVT*PMx )
{
  unsigned int ivec=0;
  for( unsigned int ix=0; ix<_nx; ix++  ){
    PMx[ix].set( _PMenv );
    PMx[ix].set( vec+ivec );
    ivec+=_PMenv->nmon();
    PMx[ix].set( T( vec[ivec], vec[ivec+1] ) );
    ivec+=2;
  }
  return;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_vec2PME
( const double*vec, PVT*PMx )
{
  unsigned int ivec=0;
  for( unsigned int ix=0; ix<_nx; ix++  ){
    PMx[ix].set( _PMenv );
    PMx[ix].set( vec+ivec );
    ivec+=_PMenv->nmon();
  }
  for( unsigned int iQ=0; iQ<_nx*(_nx+1)/2; iQ++ )
    _Q[iQ] = vec[ivec++];
  _e2x( _Q, _Ed, _Id );
  for( unsigned int ix=0; ix<_nx; ix++  )
    PMx[ix].set( _Id[ix] );
  return;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_PMI2vec
( const PVT*PM, double*vec )
{
  unsigned int ivec=0;
  for( unsigned int ix=0; ix<_nx; ix++ ){
    std::pair<unsigned int, const double*> PMcoef = PM[ix].coefmon();
    unsigned int imon = 0;
    for( ; imon<PMcoef.first; imon++ )
      vec[ivec++] = PMcoef.second[imon];
    for( ; imon<_PMenv->nmon(); imon++ )
      vec[ivec++] = 0.;
    vec[ivec++] = Op<T>::l( PM[ix].remainder() );
    vec[ivec++] = Op<T>::u( PM[ix].remainder() );
  }
  return;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_PMI2vec
( const PVT*PM, const double*RxL, const double*RxU, double*vec )
{
  unsigned int ivec=0;
  for( unsigned int ix=0; ix<_nx; ix++ ){
    std::pair<unsigned int, const double*> PMcoef = PM[ix].coefmon();
    unsigned int imon = 0;
    for( ; imon<PMcoef.first; imon++ )
      vec[ivec++] = PMcoef.second[imon];
    for( ; imon<_PMenv->nmon(); imon++ )
      vec[ivec++] = 0.;
    vec[ivec++] = RxL[ix];
    vec[ivec++] = RxU[ix];
  }
  return;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_PMI2vec
( const PVT*PM, const T*Rx, double*vec )
{
  unsigned int ivec=0;
  for( unsigned int ix=0; ix<_nx; ix++ ){
    std::pair<unsigned int, const double*> PMcoef = PM[ix].coefmon();
    unsigned int imon = 0;
    for( ; imon<PMcoef.first; imon++ )
      vec[ivec++] = PMcoef.second[imon];
    for( ; imon<_PMenv->nmon(); imon++ )
      vec[ivec++] = 0.;
    vec[ivec++] = Op<T>::l( Rx[ix] );
    vec[ivec++] = Op<T>::u( Rx[ix] );
  }
  return;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_PME2vec
( const PVT*PM, const double*QRx, double*vec )
{
  unsigned int ivec=0;
  for( unsigned int ix=0; ix<_nx; ix++ ){
    std::pair<unsigned int, const double*> PMcoef = PM[ix].coefmon();
    unsigned int imon = 0;
    for( ; imon<PMcoef.first; imon++ )
      vec[ivec++] = PMcoef.second[imon];
    for( ; imon<_PMenv->nmon(); imon++ )
      vec[ivec++] = 0.;
  }
  for( unsigned int iQ=0; iQ<_nx*(_nx+1)/2; iQ++ )
    vec[ivec++] = QRx[iQ];
  return;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_init
( PVT*PMx0 )
{
  switch( options.WRAPMIT){

  case Options::NONE:
  case Options::DINEQ:
    _pDAG->eval( _opIC, _nx, _pIC, PMx0, _np, _pVAR+_nx, _PMp );
    return _PMI2vec( PMx0, _vec_state );

  case Options::ELLIPS:
  default:{
    _pDAG->eval( _opIC, _nx, _pIC, PMx0, _np, _pVAR+_nx, _PMp );
    for( unsigned int ip=0; ip<_np; ip++ )
      _pref[ip] = Op<T>::mid( _PMp[ip].B() );
    double trR2 = 0.;
    for( unsigned int ix=0, iQ=0; ix<_nx; ix++ ){
      trR2 += sqr( Op<T>::diam( PMx0[ix].remainder() ) / 2. );
      _Q[iQ++] = Op<T>::diam( PMx0[ix].remainder() ) / 2.;
      for( unsigned int jx=ix+1; jx<_nx; jx++ ) _Q[iQ++] = 0.;
      _Id[ix] = PMx0[ix].remainder();
    }
    for( unsigned int ix=0, iQ=0; ix<_nx; ix++, iQ+=_nx-ix )
      _Q[iQ] *= std::sqrt( trR2 );
    _Ed.set( _nx, _Q );
#ifdef MC__ODEBND_GSL_LINTR_DEBUG
    std::cout << "ERx0 =" << _Ed << std::endl;
#endif
    return _PME2vec( PMx0, _Q, _vec_state ); }
  }
}

template <typename T, typename PMT, typename PVT> inline int
ODEBND_GSL<T,PMT,PVT>::MC_GSLRHSPM__
( double t, const double* x, double* xdot, void* user_data )
{
  ODEBND_GSL<T,PMT,PVT> *pODEBND_GSL = ODEBND_GSL<T,PMT,PVT>::pODEBND_GSL;
  int flag = pODEBND_GSL->_ODEBND_GSL_RHSPM( t, x, xdot, user_data );
  ODEBND_GSL<T,PMT,PVT>::pODEBND_GSL = pODEBND_GSL;
  return flag;
}

template <typename T, typename PMT, typename PVT> inline int
ODEBND_GSL<T,PMT,PVT>::_ODEBND_GSL_RHSPM
( double t, const double* x, double* xdot, void* user_data )
{
  if( !_pRHS ) return GSL_EBADFUNC;
  stats_traj.numRHS++;

  switch( options.WRAPMIT){
  case Options::NONE:
    _vec2PMI( x, _PMx );   // set current state polynomial model
#ifdef MC__ODEBND_GSL_PM_DEBUG
    std::cerr << "PMx Intermediate" << std::endl;
    _print_interm( t, _PMx, "x", std::cerr );
#endif
    _PMVAR[_nx+_np] = t; // set current time
    _pDAG->eval( _opRHS, _PMRHS, _nx, _pRHS, _PMxdot, _nVAR, _pVAR, _PMVAR );
    _PMI2vec( _PMxdot, xdot );
#ifdef MC__ODEBND_GSL_PM_DEBUG
    std::cerr << "PMxdot Intermediate" << std::endl;
    _print_interm( t, _PMxdot, "x", std::cerr );
    {int dum; std::cin >> dum;}
#endif
    return GSL_SUCCESS;  

  case Options::DINEQ:
    _vec2PMI( x, _PMx );   // set current state polynomial model
    for( unsigned int ix=0; ix<_nx; ix++ ) _PMx[ix].center();
#ifdef MC__ODEBND_GSL_PM_DEBUG
    std::cerr << "PMx Intermediate" << std::endl;
    _print_interm( t, _PMx, "x", std::cerr );
#endif
    _PMVAR[_nx+_np] = t; // set current time
    for( unsigned int ix=0; ix<_nx; ix++ ){
      T Rxi = _PMx[ix].remainder();
      _PMx[ix].set( Op<T>::l( Rxi ) );
      _pDAG->eval( _opRHSi[ix], _PMRHS, 1, _pRHS+ix, _PMxdot+ix, _nVAR, _pVAR, _PMVAR );
      _RxLdot[ix] = Op<T>::l( _PMxdot[ix].remainder() );
      _PMx[ix].set( Op<T>::u( Rxi ) );
      _pDAG->eval( _opRHSi[ix], _PMRHS, 1, _pRHS+ix, _PMxdot+ix, _nVAR, _pVAR, _PMVAR );
      _RxUdot[ix] = Op<T>::u( _PMxdot[ix].remainder() );
      _PMx[ix].set( Rxi );
    }
    _PMI2vec( _PMxdot, _RxLdot, _RxUdot, xdot );
#ifdef MC__ODEBND_GSL_PM_DEBUG
    std::cerr << "PMxdot Intermediate" << std::endl;
    _print_interm( t, _PMxdot, "x", std::cerr );
    {int dum; std::cin >> dum;}
#endif
    return GSL_SUCCESS;  
   
  case Options::ELLIPS:{
  default:
    _vec2PME( x, _PMx );   // set current state polynomial model
    _PMVAR[_nx+_np] = t;   // set current time   
#ifdef MC__ODEBND_GSL_PM_DEBUG
    std::cout << "@t=" << t << std::endl;
    for( unsigned int ix=0; ix<_nx; ix++ )
      std::cout << "PMx[ " << ix << "] = " << _PMx[ix] << std::endl;
    std::cout << "Ed = " << _Ed << std::endl;
    for( unsigned int ix=0; ix<_nx; ix++ )
      std::cout << "Id[" << ix << "] = " << _Id[ix]
                << " : " << Op<T>::mid(_Id[ix]) << std::endl;
#endif

    // In this variant a bound on the Jacobian matrix is computed and the
    // linear part is taken as the mid-point of this matrix
    if( !options.ORDMIT )
      _PROP_PM_ELL0( t );

    // In this variant a polynomial model of the Jacobian matrix is computed and the
    // linear part is taken as the mid-point of this matrix
    else if( _PMenv->nord() > _MVXPenv->nord() )
      _PROP_PM_ELL1( t );

    // In this variant a polynomial model in the joint state-parameter and
    // of the same order as the parameter polynomial model is computed
    else
      _PROP_PM_ELL2( t );

    // Set dynamics of shape matrix
    double trQ = 0., sumkappa = 0.;
    for( unsigned int ix=0; ix<_nx; ix++ )
      trQ += ( _Q[_ndxLT(ix,ix)]>0? _Q[_ndxLT(ix,ix)]: machprec() );
    const double srqt_trQ = (trQ>0? std::sqrt( trQ ): 0.) + options.QTOL;
    for( unsigned int ix=0; ix<_nx; ix++ ){
#ifdef MC__ODEBND_GSL_PM_DEBUG
        std::cout << "kappa[" << ix << "] = "
                  << Op<T>::diam( _Iddot[ix] ) / ( 2. * srqt_trQ ) << std::endl;
#endif
        sumkappa += Op<T>::diam( _Iddot[ix] ) / ( 2. * srqt_trQ );
    }

    for( unsigned int jx=0; jx<_nx; jx++ ){
      for( unsigned int ix=jx; ix<_nx; ix++ ){
        _Qdot[_ndxLT(ix,jx)] = sumkappa * _Q[_ndxLT(ix,jx)];
        for( unsigned int kx=0; kx<_nx; kx++ )
          _Qdot[_ndxLT(ix,jx)] += _Q[_ndxLT(ix,kx)] * _A[jx+kx*_nx]
                         + _A[ix+kx*_nx] * _Q[_ndxLT(kx,jx)];
      }
      _Qdot[_ndxLT(jx,jx)] += Op<T>::diam( _Iddot[jx] ) / 2. * srqt_trQ;
    }

#ifdef MC__ODEBND_GSL_PM_DEBUG
    for( unsigned int ix=0; ix<_nx; ix++ )
      std::cout << "PMxdot[" << ix << "] =" << _PMxdot[ix] << std::endl;
    for( unsigned int ix=0; ix<_nx; ix++ )
      std::cout << "Iddot[" << ix << "] =" << _Iddot[ix] << std::endl;
    E ERxdot( _nx, _Qdot );
    std::cout << "ERxdot =" << ERxdot << std::endl;
    { int dum; std::cin >> dum; }
#endif
   
    // Copy time derivatives in GSL array
    _PME2vec( _PMxdot, _Qdot, xdot );
 
    return GSL_SUCCESS;
   }
  }
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_PROP_PM_ELL0
( const double t )
{
  // Compute polynomial approximant and ellipsoidal remainder of RHS
  for( unsigned int ix=0; ix<_nx; ix++ ){
    _IVAR[ix] = _PMx[ix].bound(); // set current state bounds
    ( _PMx[ix].center() ).set( T(0.) ); // cancel remainder term
  }
  _IVAR[_nx+_np] = t; // set current time

  _pDAG->eval( _opRHS, _PMRHS, _nx, _pRHS, _PMxdot, _nVAR, _pVAR, _PMVAR );
  _pDAG->eval( _opJAC, _IJAC, _nx*_nx, _pJAC, _Idfdx, _nVAR, _pVAR, _IVAR );

  for( unsigned int ix=0; ix<_nx; ix++ ){
    for( unsigned int jx=0; jx<_nx; jx++ )
      _A[ix+jx*_nx] = Op<T>::mid( _Idfdx[ix*_nx+jx] );
    _Iddot[ix] = _PMxdot[ix].remainder();
    for( unsigned int jx=0; jx<_nx; jx++ )
      _Iddot[ix] += ( _Idfdx[ix*_nx+jx] - _A[ix+jx*_nx] ) * _Id[jx];
#ifdef MC__ODEBND_GSL_PM_DEBUG
    std::cout << "Iddot[" << ix << "] = " << _Iddot[ix]
              << " : " << Op<T>::mid(_Iddot[ix]) << std::endl;
#endif
  }
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_PROP_PM_ELL1
( const double t )
{
  // Compute polynomial approximant and ellipsoidal remainder of RHS
  for( unsigned int jx=0; jx<_nx; jx++ ){
    _MVXPd[jx].set( _MVXPenv, _np+jx, _Id[jx] );
    _MVXPx[jx].set( _MVXPenv ).set( _PMx[jx].center().set( T(0.) ), true );
  }
  _e2x( _MVXPd, _MVXPx, false );
#ifdef MC__ODEBND_GSL_PM_DEBUG
  for( unsigned int ix=0; ix<_nx; ix++ )
    std::cout << "MVXPx[ " << ix << "] = " << _MVXPx[ix] << std::endl;
  //{ int dum; std::cin >> dum; }
#endif
  _MVXPVAR[_nx+_np] = t; // set current time   

  _pDAG->eval( _opRHS, _PMRHS, _nx, _pRHS, _PMxdot, _nVAR, _pVAR, _PMVAR );
  _pDAG->eval( _opJAC, _PMJAC, _nx*_nx, _pJAC, _MVXPdfdx, _nVAR, _pVAR, _MVXPVAR );

  for( unsigned int ix=0; ix<_nx; ix++ ){
#ifdef MC__ODEBND_GSL_PM_DEBUG
    for( unsigned int jx=0; jx<_nx; jx++ )
      std::cout << "PMdfdx[" << ix << "," << jx << "] = " << _MVXPdfdx[ix*_nx+jx] << std::endl;
#endif
    // Extract constant coefficients and set to 0
    for( unsigned int jx=0; jx<_nx; jx++ )
      _A[ix+jx*_nx] = ( _MVXPdfdx[ix*_nx+jx].center() ).constant( true );
#ifdef MC__ODEBND_GSL_PM_DEBUG
    std::cout << "A[" << ix << ",#] = ";
    for( unsigned int jx=0; jx<_nx; jx++ )
      std::cout << _A[ix+jx*_nx] << "  ";
    std::cout << std::endl;
    for( unsigned int jx=0; jx<_nx; jx++ )
      std::cout << "PMdfdx[" << ix << "," << jx << "] = " << _MVXPdfdx[ix*_nx+jx] << std::endl;
#endif
    // Bound remaining terms
    _Iddot[ix] = _PMxdot[ix].remainder();
    for( unsigned int jx=0; jx<_nx; jx++ )
      _Iddot[ix] += _MVXPdfdx[ix*_nx+jx].bound() * _Id[jx];
#ifdef MC__ODEBND_GSL_PM_DEBUG
    std::cout << "Iddot[" << ix << "] = " << _Iddot[ix]
              << " : " << Op<T>::mid(_Iddot[ix]) << std::endl;
#endif
  }
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_PROP_PM_ELL2
( const double t )
{
  // Compute polynomial approximant and ellipsoidal remainder of RHS
  for( unsigned int jx=0; jx<_nx; jx++ ){
    _MVXPd[jx].set( _MVXPenv, _np+jx, _Id[jx] );
    _MVXPx[jx].set( _MVXPenv ).set( _PMx[jx].center().set( T(0.) ), true );
  }
  _e2x( _MVXPd, _MVXPx, false );
#ifdef MC__ODEBND_GSL_PM_DEBUG
  for( unsigned int ix=0; ix<_nx; ix++ )
    std::cout << "MVXPx[ " << ix << "] = " << _MVXPx[ix] << std::endl;
  //{ int dum; std::cin >> dum; }
#endif
  _MVXPVAR[_nx+_np] = t; // set current time   

  _pDAG->eval( _opRHS, _PMRHS, _nx, _pRHS, _MVXPf, _nVAR, _pVAR, _MVXPVAR );

  for( unsigned int ix=0; ix<_nx; ix++ ){
#ifdef MC__ODEBND_GSL_PM_DEBUG
    std::cout << "MVXPf[" << ix << "] = " << _MVXPf[ix] << std::endl;
#endif
    // Extract polynomial model in P and set to 0
    _MVXPf[ix].get( _PMxdot[ix].set(_PMenv), true );
#ifdef MC__ODEBND_GSL_PM_DEBUG
    std::cout << "PMxdot[" << ix << "] =" << _PMxdot[ix] << std::endl;
    std::cout << "MVXPf[" << ix << "] = " << _MVXPf[ix] << std::endl;
    { int dum; std::cin >> dum; }
#endif
    // Extract linear part in X and set to 0
    for( unsigned int jx=0; jx<_nx; jx++ )
      _A[ix+jx*_nx] = _MVXPf[ix].linear( _np+jx, true );
#ifdef MC__ODEBND_GSL_PM_DEBUG
    std::cout << "A[" << ix << ",#] = ";
    for( unsigned int jx=0; jx<_nx; jx++ )
      std::cout << _A[ix+jx*_nx] << "  ";
    std::cout << std::endl;
    std::cout << "MVXPf[" << ix << "] = " << _MVXPf[ix] << std::endl;
#endif
    // Bound remaining terms
    _Iddot[ix] = _MVXPf[ix].bound();
#ifdef MC__ODEBND_GSL_PM_DEBUG
    std::cout << "Iddot[" << ix << "] = " << _Iddot[ix]
              << " : " << Op<T>::mid(_Iddot[ix]) << std::endl;
#endif
  }
}
/*
template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::_INV_PM_ELL0()
{
  // Compute Taylor model polynomial part and ellipsoidal remainder of RHS
  E Exdot.set( _Qdot );
  for( unsigned int ix=0; ix<_nx; ix++ ){
    _PMVAR[ix] = _PMxdot[ix].set( T(0.) ); // polynomial model of state derivatives
    _Iddot[ix] = T( Exdot.l(ix), Exdot.u(ix) );
    _IVAR[ix] = _PMxdot[ix].bound() + _Iddot[ix]; // bound of state derivatives
  }

  _pDAG->eval( _opINV, _ni, _pINV, _PMinv, _nVAR, _pVAR, _PMVAR );
  _pDAG->eval( _opDINV, _ni*_nx, _pDINV, _Idinvdx, _nVAR, _pVAR, _IVAR );

  for( unsigned int ii=0; ii<_ni; ii++ ){
    _Rinv[ii] = _PMinv[ii].remainder();
    for( unsigned int jx=0; jx<_nx; jx++ ){
      _Ainv[ii+jx*_ni] = Op<T>::mid( _Idinvdx[ii*_nx+jx] );
      _Rinv[ii] += ( _Idinvdx[ii*_nx+jx] - _Ainv[ii+jx*_ni] ) * _Iddot[jx];
    }
#ifdef MC__ODEBND_GSL_PM_DEBUG
    std::cout << "Rinv[" << ii << "] =" << _Rinv[ii]
              << "  diam(Rinv[" << ii << "]) =" << Op<T>::diam(_Rinv[ii])
              << "  midp(Rinv[" << ii << "]) =" << Op<T>::mid(_Rinv[ii])
              << std::endl;
    std::cout << "Ainv[" << ii << ",#] = ";
    for( unsigned int jx=0; jx<_nx; jx++ )
      std::cout << _Ainv[ii+jx*_ni] << "  ";
    std::cout << std::endl;
#endif
  }
}
*/
template <typename T, typename PMT, typename PVT> inline int
ODEBND_GSL<T,PMT,PVT>::MC_GSLJACPM__
( double t, const double* x, double* jac, double* xdot, void* user_data )
{
  ODEBND_GSL<T,PMT,PVT> *pODEBND_GSL = ODEBND_GSL<T,PMT,PVT>::pODEBND_GSL;
  int flag = pODEBND_GSL->_ODEBND_GSL_JACPM( t, x, jac, xdot, user_data );
  ODEBND_GSL<T,PMT,PVT>::pODEBND_GSL = pODEBND_GSL;
  return flag;
}

template <typename T, typename PMT, typename PVT> inline int
ODEBND_GSL<T,PMT,PVT>::_ODEBND_GSL_JACPM
( double t, const double* x, double* jac, double* xdot, void* user_data )
{
  stats_traj.numJAC++;
  return GSL_EBADFUNC;
}

template <typename T, typename PMT, typename PVT> inline bool
ODEBND_GSL<T,PMT,PVT>::_GSL_init
( const PVT* PMp )
{
  // Check polynomial model compatibility and size
  unsigned int kp=_np;
  for( unsigned int ip=0; ip<_np && kp==_np; ip++ )
    if( PMp[ip].env() ) kp = ip;
  if( kp==_np || PMp[kp].env()->nvar()!=_np ) return false;
  _PMenv = PMp[kp].env();  

  // Define ODE system in GSL format
  _sys_dineqPM.function = MC_GSLRHSPM__;
  _sys_dineqPM.jacobian = MC_GSLJACPM__;
  _sys_dineqPM.params = 0;
  _sys_dineqPM.dimension = _PMenv->nmon()*_nx;
  switch( options.WRAPMIT){
  case Options::NONE:
  case Options::DINEQ:
    _sys_dineqPM.dimension += 2*_nx;
    break;
  case Options::ELLIPS:
  default:
    _sys_dineqPM.dimension += _nx*(_nx+1)/2;
    break;
  }
  _GSL_init( _sys_dineqPM, _driver_dineqPM );

  // Size and set DAG evaluation arrays
  _nVAR = _nx+_np+1;
  delete[] _pVAR; _pVAR = new FFVar[_nVAR];
  delete[] _IVAR; _IVAR = new T[_nVAR];
  delete[] _PMVAR; _PMVAR = new PVT[_nVAR];
  for( unsigned ix=0; ix<_nx; ix++ ) _pVAR[ix] = _pX[ix];
  for( unsigned ip=0; ip<_np; ip++ ) _pVAR[_nx+ip] = _pP[ip];
  _pVAR[_nx+_np] = (_pT? *_pT: 0. );
  for( unsigned ip=0; ip<_np; ip++ ) _IVAR[_nx+ip] = PMp[ip].bound();
  _Ix = _Ip = 0; // not used
  for( unsigned ip=0; ip<_np; ip++ ) _PMVAR[_nx+ip] = PMp[ip];
  _PMx = _PMVAR;
  _PMp = _PMVAR + _nx;

  // Reset _MVXPVenv and related variables
  unsigned int MVXPsize = ( options.ORDMIT<_PMenv->nord()? options.ORDMIT: _PMenv->nord() ); 
  if( _MVXPenv && ( _MVXPenv->nord() != MVXPsize || _MVXPenv->nvar() != _nx+_np ) ){
    delete[] _MVXPf;    _MVXPf = 0;
    delete[] _MVXPdfdx; _MVXPdfdx = 0;
    delete[] _MVXPd;    _MVXPd = 0;
    delete[] _MVXPVAR;  _MVXPVAR = _MVXPx = _MVXPp = 0; 
    delete   _MVXPenv;  _MVXPenv = 0;
  }

  // Size state/parameter arrays
  if( !_PMx )	 _PMx	 = new PVT[_nx];
  if( !_PMxdot ) _PMxdot = new PVT[_nx];

  switch( options.WRAPMIT){
  case Options::NONE:
    break;

  case Options::DINEQ:
    if( !_RxLdot ) _RxLdot = new double[_nx];
    if( !_RxUdot ) _RxUdot = new double[_nx];
    break;

  case Options::ELLIPS:
    if( !_pref )     _pref     = new double[_np];
    if( !_xref )     _xref     = new double[_nx];
    if( !_A )	     _A        = new double[_nx*_nx];
    if( !_Q )	     _Q        = new double[_nx*(_nx+1)/2];
    if( !_Id )       _Id       = new T[_nx];
    if( !_Qdot )     _Qdot     = new double[_nx*(_nx+1)/2];
    if( !_Iddot )    _Iddot    = new T[_nx];
    if( !_Idfdx )    _Idfdx    = new T[_nx*_nx];
    if( !_MVXPenv )  _MVXPenv  = new PMT( _nx+_np, MVXPsize );
    _MVXPenv->options = _PMenv->options;
    if( !_MVXPd )    _MVXPd    = new PVT[_nx];
    if( !_MVXPf )    _MVXPf    = new PVT[_nx];
    if( !_MVXPdfdx ) _MVXPdfdx = new PVT[_nx*_nx];
    if( !_MVXPVAR )  _MVXPVAR  = new PVT[_nVAR];
    _MVXPx = _MVXPVAR;
    _MVXPp = _MVXPVAR + _nx;
    for( unsigned int ip=0; ip<_np; ip++ )
      _MVXPp[ip].set( _MVXPenv, ip, _PMp[ip].B() );
    break;
  }

  // Reset result record and statistics
  _results.clear();  
  _init_stats( stats_traj );

  return true;
}

template <typename T, typename PMT, typename PVT> inline bool
ODEBND_GSL<T,PMT,PVT>::_set_RHS_dineqPM
( const unsigned int iRHS )
{
  if( _vRHS.size() <= iRHS ) return false;

  _pRHS = _vRHS.at( iRHS );
  _opRHS = _pDAG->subgraph( _nx, _pRHS );

  delete[] _IRHS;   _IRHS = 0;
  delete[] _PMRHS;  _PMRHS = 0;
  delete[] _opRHSi; _opRHSi = 0;
  delete[] _IJAC;   _IJAC = 0;
  delete[] _PMJAC;  _PMJAC = 0;
  delete[] _pJAC;   _pJAC = 0; _opJAC.clear();
  switch( options.WRAPMIT){
  case Options::NONE:
    _PMRHS = new PVT[ _opRHS.size() ]; break;
  case Options::DINEQ:
    _opRHSi = new std::list<const FFOp*>[_nx];
    for( unsigned ix=0; ix<_nx; ix++ )
      _opRHSi[ix] = _pDAG->subgraph( 1, _pRHS+ix );
    _PMRHS = new PVT[ _opRHS.size() ]; break;
  case Options::ELLIPS: default:
    _PMRHS = new PVT[ _opRHS.size() ];
    _pJAC = _pDAG->FAD( _nx, _pRHS, _nx, _pVAR );
    _opJAC = _pDAG->subgraph( _nx*_nx, _pJAC );
    if( !options.ORDMIT )
      _IJAC = new T[ _opJAC.size() ];
    else if( _PMenv->nord() > _MVXPenv->nord() )
      _PMJAC = new PVT[ _opJAC.size() ];
    break;
  }

  return true;
}

//! @fn template <typename T, typename PMT, typename PVT> inline typename ODEBND_GSL<T,PMT,PVT>::STATUS ODEBND_GSL<T,PMT,PVT>::bounds(
//! const unsigned int ns, const double*tk, const PVT*PMp, PVT**PMxk, E*Exk=0, 
//! std::ostream&os=std::cout )
//!
//! This function computes an enclosure of the reachable set of the parametric ODEs
//! using propagation of polynomial models with convex remainders (intervals, ellipsoids):
//!   - <a>ns</a> [input] number of time stages
//!   - <a>tk</a> [input] stage times, including the initial time
//!   - <a>PMp</a> [input] polynomial model of parameter set
//!   - <a>PMxk</a> [output] polynomial model of state enclosures at stage times
//!   - <a>Exk</a> [output] ellipsoidal remainders of polynomial models at stage times (default: NULL)
//!   - <a>os</a> [input] output stream (default: std::cout)
//! .
//! The return value is the status.
template <typename T, typename PMT, typename PVT>
inline typename ODEBND_GSL<T,PMT,PVT>::STATUS
ODEBND_GSL<T,PMT,PVT>::bounds
( const unsigned int ns, const double*tk, const PVT*PMp, PVT**PMxk, E*Exk,
  std::ostream&os )
{
  // Initialize trajectory integration with GSL
  if( !_GSL_init( PMp ) ) return FATAL;

  try{
    // Bounds on initial conditions
    if( !_vIC.size() )
      { _GSL_term(); return FATAL; }
    _pIC = _vIC.at(0);
    _opIC = _pDAG->subgraph( _nx, _pIC );
    _init( PMxk[0] );
    if( Exk && options.WRAPMIT == Options::ELLIPS ) Exk[0] = _Ed;
    else if( Exk ) Exk[0].reset();
    if( options.DISPLAY >= 1 )
      switch( options.WRAPMIT ){
        case Options::NONE:
        case Options::DINEQ:
          _print_interm( tk[0], PMxk[0], "x", os );         break;
        case Options::ELLIPS:
          _print_interm( tk[0], PMxk[0], _Ed, "x", os ); break;
      }
    // Record initial results
    if( options.RESRECORD )
      _results.push_back( Results( tk[0], _nx, PMxk[0] ) );

    // Integrate ODEs through each stage using GSL
    _t = tk[0];
    _h = options.H0;
    pODEBND_GSL = this;

    for( _istg=0; _istg<ns; _istg++ ){
      if( _istg && _vIC.size()>1 ) throw Exceptions( Exceptions::REINIT );

      // update list of operations in RHS and JAC
      const unsigned int iRHS = ( _vRHS.size()<2? 0: _istg );
      if( (!_istg || iRHS) && !_set_RHS_dineqPM( iRHS ) )
        { _GSL_term(); return FATAL; }

      while( _t < tk[_istg+1] ){
        if( gsl_odeiv2_evolve_apply( _driver_dineqPM->e, _driver_dineqPM->c, _driver_dineqPM->s,
            &_sys_dineqPM, &_t, tk[_istg+1], &_h, _vec_state ) != GSL_SUCCESS
         || _h < options.HMIN
         || (options.NMAX && stats_traj.numSteps > options.NMAX) )
          throw Exceptions( Exceptions::UNDEF );
        stats_traj.numSteps++;
        if( options.HMAX > 0 && _h > options.HMAX ) _h = options.HMAX;
      }

      switch( options.WRAPMIT){
      case Options::NONE:
      case Options::DINEQ:
        _vec2PMI( _vec_state, PMxk[_istg+1] ); break;
        if( Exk ) Exk[_istg+1].reset();
      case Options::ELLIPS:
      default:
        _vec2PME( _vec_state, PMxk[_istg+1] ); break;
        if( Exk ) Exk[_istg+1] = _Ed;
      }
      if( options.DISPLAY >= 1 )
        switch( options.WRAPMIT ){
          case Options::NONE:
          case Options::DINEQ:
            _print_interm( tk[_istg+1], PMxk[_istg+1], "x", os );               break;
          case Options::ELLIPS:
            _print_interm( tk[_istg+1], PMxk[_istg+1], _Ed, "x", os ); break;
        }
      // Record intermediate results
      if( options.RESRECORD )
        _results.push_back( Results( tk[_istg+1], _nx, PMxk[_istg+1] ) );
    }
  }
  catch(...){
    _GSL_term();
    if( options.DISPLAY >= 1 ) _print_stats( stats_traj, os );
    return FAILURE;
  }

  _GSL_term(); 
  if( options.DISPLAY >= 1 ) _print_stats( stats_traj, os );
  return NORMAL;
}

template <typename T, typename PMT, typename PVT> inline typename ODEBND_GSL<T,PMT,PVT>::STATUS
ODEBND_GSL<T,PMT,PVT>::_remainders
( const unsigned int ns, const double*tk, const T*Ip, const PVT*const*PMxk,
  T**Rxk, const unsigned int nsamp, std::ostream&os )
{
  STATUS flag = NORMAL;
  
  // Initialization of sampled bounds at parameter lower bound
  double *p = new double[_np];
  for( unsigned int ip=0; ip<_np; ip++ )
    p[ip] = Op<T>::l(Ip[ip]);
  double **xk = new double*[ns+1];
  for( unsigned int is=0; is<=ns; is++ )
    xk[is] = new double[_nx];
  flag = _ODESLV_GSL->states( ns, tk, p, xk, 0, os );
  if( flag != NORMAL || nsamp <= 1 ){
    delete[] p;
    for( unsigned int is=0; is<=ns; is++ ) delete[] xk[is]; delete[] xk;
    return flag;
  }   
  for( unsigned int is=0; is<=ns; is++ )
    for( unsigned int ix=0; ix<_nx; ix++ )
      Rxk[is][ix] = xk[is][ix] - PMxk[is][ix].polynomial( p );
  
  // Start sampling process
  unsigned int* vsamp = new unsigned int[_np];
  flag = _remainders( ns, tk, Ip, PMxk, Rxk, nsamp, vsamp, 0, p, xk, os );
  
  // Clean-up
  delete[] p;
  for( unsigned int is=0; is<=ns; is++ ) delete[] xk[is]; delete[] xk;
  delete[] vsamp;
  
  return flag;
}

template <typename T, typename PMT, typename PVT> inline typename ODEBND_GSL<T,PMT,PVT>::STATUS
ODEBND_GSL<T,PMT,PVT>::_remainders
( const unsigned int ns, const double*tk, const T*Ip, const PVT*const*PMxk,
  T**Rxk, const unsigned int nsamp, unsigned int* vsamp, const unsigned int ip,
  double*p, double**xk, std::ostream&os )
{
  STATUS flag = NORMAL;

  // Update bounds for all sampling points
  for( unsigned int isamp=0; isamp<nsamp; isamp++ ){
    vsamp[ip] = isamp;

    // Continue recursive call
    if( ip+1 < _np ){
      flag = _remainders( ns, tk, Ip, PMxk, Rxk, nsamp, vsamp, ip+1, p, xk, os );
      if( flag != NORMAL ) return flag;
      continue;
    }

    // Update bounds for current point
#ifdef MC__ODEBND_GSL_SAMPLE_DEBUG
    std::cout << "Sample: ";
#endif
    for( unsigned int ip=0; ip<_np; ip++ ){
      p[ip] = Op<T>::l( Ip[ip] ) + vsamp[ip]/(nsamp-1.) * Op<T>::diam( Ip[ip] );
#ifdef MC__ODEBND_GSL_SAMPLE_DEBUG
      std::cout << p[ip] << "  ";
#endif
    }
#ifdef MC__ODEBND_GSL_SAMPLE_DEBUG
    std::cout << std::endl;
#endif
    flag = _ODESLV_GSL->states( ns, tk, p, xk, 0, os );
    if( flag != NORMAL ) return flag;
    for( unsigned int is=0; is<=ns; is++ )
      for( unsigned int ix=0; ix<_nx; ix++ )
        Rxk[is][ix] = Op<T>::hull( xk[is][ix]-PMxk[is][ix].polynomial(p),
                                   Rxk[is][ix] );
  }

  return flag;
}  

//! @fn template <typename T, typename PMT, typename PVT> inline typename ODEBND_GSL<T,PMT,PVT>::STATUS ODEBND_GSL<T,PMT,PVT>::hausdorff(
//! const unsigned int ns, const double*tk, const PVT*PMp, double**Hxk,
//! const unsigned int nsamp, std::ostream&os )
//!
//! This function computes the Hausdorff distance between the polynomial model
//! remainder and the actual (sampled) range of the remainder function
//! in projection onto each variable and for each stage time
//! remainder and the actual range of the remainder function:
//!   - <a>ns</a> [input] number of time stages
//!   - <a>tk</a> [input] stage times, including the initial time
//!   - <a>PMp</a> [input] polynomial model of parameter set
//!   - <a>Hxk</a> [output] Hausdorff distance between the polynomial model
//!     remainder and the actual (sampled) range of the remainder function,
//!     at stage times
//!   - <a>nsamp</a> [input] number of samples for each parameter
//!   - <a>os</a> [input] output stream (default: std::cout)

template <typename T, typename PMT, typename PVT> inline typename ODEBND_GSL<T,PMT,PVT>::STATUS
ODEBND_GSL<T,PMT,PVT>::hausdorff
( const unsigned int ns, const double*tk, const PVT*PMp, double**Hxk,
  const unsigned int nsamp, std::ostream&os )
{
  int DISPLAY_ODEBND = options.DISPLAY;
  int DISPLAY_ODESLV = options.ODESLVOPT.DISPLAY;
  options.DISPLAY = 0;
  options.ODESLVOPT.DISPLAY = 0;
  _ODESLV_GSL->set( *this ); // in base class BASE_ODE
  _ODESLV_GSL->options = options.ODESLVOPT;
  _ODESLV_GSL->options.RESRECORD = false;

  // Compute approximate bounds
  PVT** PMxk = new PVT*[ns+1];
  for( unsigned int is=0; is<ns+1; is++ ) PMxk[is] = new PVT[_nx];
  try{ bounds( ns, tk, PMp, PMxk, 0, os ); }
  catch(...){;}
  unsigned int nsf = _istg;
  for( unsigned int is=nsf; is<ns; is++ )
    for( unsigned int ix=0; ix<_nx; ix++ )
      PMxk[is+1][ix] = T( -1e20, 1e20 );

  // Compute remainder bounds 
  T* Ip = new T[_np];
  for( unsigned int ip=0; ip<_np; ip++ ) Ip[ip] = PMp[ip].B();
  T** Rxk = new T*[ns+1];
  for( unsigned int is=0; is<ns+1; is++ ) Rxk[is] = new T[_nx];
  _remainders( nsf, tk, Ip, PMxk, Rxk, nsamp, os );
  for( unsigned int is=nsf; is<ns; is++ )
    for( unsigned int ix=0; ix<_nx; ix++ )
      Rxk[is+1][ix] = T( 0. );

  // Compute Hausdorff metric at each time step
  struct loc{ static double hausdorff
    ( const T&Ix, const T&Ix0 )
    { return std::max( std::fabs(Op<T>::l(Ix)-Op<T>::l(Ix0)),
                       std::fabs(Op<T>::u(Ix)-Op<T>::u(Ix0)) ); }
  };

  options.DISPLAY = DISPLAY_ODEBND;
  options.ODESLVOPT.DISPLAY = DISPLAY_ODESLV;
  for( unsigned int is=0; is<ns+1; is++ ){
    for( unsigned int ix=0; ix<_nx; ix++ )
      Hxk[is][ix] = loc::hausdorff( PMxk[is][ix].R(), Rxk[is][ix] );
    if( options.DISPLAY >= 1 ){
      _print_interm( tk[is], Hxk[is], "dH", os );
    }
  }

  for( unsigned int is=0; is<ns+1; is++ ) delete[] Rxk[is];
  delete[] Rxk;
  for( unsigned int is=0; is<ns+1; is++ ) delete[] PMxk[is];
  delete[] PMxk;
  delete[] Ip;

  return NORMAL;
}

template <typename T, typename PMT, typename PVT>
template <typename U> inline void
ODEBND_GSL<T,PMT,PVT>::_print_interm
( const double t, const U*x, const std::string&var, std::ostream&os ) const
{
  os << " @t = " << std::scientific << std::setprecision(4)
                 << std::left << t << " :" << std::endl;
  for( unsigned int ix=0; ix<_nx; ix++ )
    os << " " << var.c_str() << "[" << ix << "] = " << x[ix] << std::endl;
  return;
}

template <typename T, typename PMT, typename PVT>
template <typename U, typename V> inline void
ODEBND_GSL<T,PMT,PVT>::_print_interm
( const double t, const U*x, const V&r, const std::string&var, std::ostream&os ) const
{
  os << " @t = " << std::scientific << std::setprecision(4)
                 << std::left << t << " :" << std::endl;
  for( unsigned int ix=0; ix<_nx; ix++ )
    os << " " << var.c_str() << "[" << ix << "] = " << x[ix] << std::endl;
  os << " " << "R" << var.c_str() << " =" << r << std::endl;
  return;
}

template <typename T, typename PMT, typename PVT> inline void
ODEBND_GSL<T,PMT,PVT>::record
( std::ofstream&bndrec, const unsigned int iprec ) const
{
  if( !bndrec ) return;

  // Specify format
  bndrec << std::right << std::scientific << std::setprecision(iprec);

  // Record computed interval bounds at stage times
  typename std::vector< Results >::const_iterator it = _results.begin();
  for( ; it != _results.end(); ++it ){
    bndrec << std::setw(iprec+9) << (*it).t;
    for( unsigned int ix=0; ix<_nx; ix++ )
      bndrec << std::setw(iprec+9) << mc::Op<T>::l( (*it).X[ix] )
             << std::setw(iprec+9) << mc::Op<T>::u( (*it).X[ix] );
    bndrec << std::endl;
  }
}

} // end namescape mc

#endif

