// Copyright (C) 2012-2014 Benoit Chachuat, Imperial College London.
// All Rights Reserved.
// This code is published under the Eclipse Public License.

#ifndef MC__ODESLV_GSL_HPP
#define MC__ODESLV_GSL_HPP

#undef  MC__ODESLV_GSL_SAMPLE_DEBUG
#undef  MC__ODESLV_GSL_DEBUG_INTERP
#undef  MC__ODESLV_GSL_DEBUG_ADJOINT
#define MC__ODESLV_GSL_CHECK

#include <stdexcept>
#include <cassert>
#include <cmath>
#include <fstream>
#include <vector>
#include <sys/time.h>

#include "base_gsl.hpp"

namespace mc
{
//! @brief C++ class computing solutions of parametric ODEs using non-validated integration.
////////////////////////////////////////////////////////////////////////
//! mc::ODESLV_GSL is a C++ class that computes solutions of parametric
//! ordinary differential equations (ODEs) using GSL.
////////////////////////////////////////////////////////////////////////
template <typename T>
class ODESLV_GSL: public BASE_GSL
{
private:
  //! @brief GSL data type for ODE integration
  gsl_odeiv2_system _sys_traj;

  //! @brief GSL driver for ODE integration
  gsl_odeiv2_driver *_driver_traj;

  //! @brief GSL data type for adjoint ODE integration
  gsl_odeiv2_system _sys_adj;

  //! @brief GSL drivers for adjoint ODE integration
  std::vector<gsl_odeiv2_driver*> _driver_adj;

  //! @brief GSL lookup accelerators for interpolation
  std::vector<gsl_interp_accel*> _accel_interp;

  //! @brief GSL drivers for interpolation
  std::vector<gsl_interp*> _driver_interp;

  //! @brief number of variables for DAG evaluation
  unsigned int _nVAR;

  //! @brief pointer to variables for DAG evaluation
  FFVar* _pVAR;

  //! @brief pointer to variable values for DAG evaluation
  double *_dVAR;

  //! @brief list of operations in RHS evaluation
  std::list<const FFOp*> _opRHS;

  //! @brief list of operations in RHS Jacobian
  std::list<const FFOp*> _opJAC;

  //! @brief list of operations in IC evalution
  std::list<const FFOp*> _opIC;

  //! @brief list of operations in adjoint RHS evaluation
  std::list<const FFOp*> _opADJ;

  //! @brief list of operations in adjoint TC evaluation
  std::list<const FFOp*> _opTC;

  //! @brief const pointer to RHS function in current stage of ODE system
  const FFVar* _pRHS;

  //! @brief preallocated array for evaluation of RHS function
  double* _dRHS;

  //! @brief const pointer to RHS Jacobian in current stage of ODE system
  const FFVar* _pJAC;

  //! @brief preallocated array for evaluation of RHS Jacobian
  double* _dJAC;

  //! @brief const pointer to IC function in current stage of ODE system
  const FFVar* _pIC;

  //! @brief const pointer to adjoint RHS function in current stage of ODE system
  const FFVar* _pADJ;

  //! @brief preallocated array for evaluation of adjoint RHS function
  double* _dADJ;

  //! @brief const pointer to adjoint TC function in current stage of ODE system
  const FFVar* _pTC;

  //! @brief array for storing TC Jacobian in current stage
  double* _dTC;

public:
  /** @defgroup ODESLV_GSL Real-valued (non-validated) integration of parametric ODEs
   *  @{
   */
  //! @brief Default class constructor
  ODESLV_GSL()
    : BASE_GSL(), _driver_traj(0), _nVAR(0), _pVAR(0), _dVAR(0),
      _pRHS(0), _dRHS(0), _pJAC(0), _dJAC(0), _pIC(0), _pADJ(0),
      _dADJ(0), _pTC(0), _dTC(0), _mesh(0), _meshndx(0), _ifct(0)
    {}

  //! @brief Default destructor
  virtual ~ODESLV_GSL()
    {    
      if( _driver_traj ) gsl_odeiv2_driver_free( _driver_traj );
      for( unsigned i=0; i<_driver_adj.size(); i++ )
        if( _driver_adj[i] )  gsl_odeiv2_driver_free( _driver_adj[i] );
      for( unsigned i=0; i<_driver_interp.size(); i++ )
        if( _driver_interp[i] ) gsl_interp_free( _driver_interp[i] );
      for( unsigned i=0; i<_accel_interp.size(); i++ )
        if( _accel_interp[i] )  gsl_interp_accel_free( _accel_interp[i] );
      delete[] _dVAR;
      delete[] _dRHS;
      delete[] _dJAC;
      delete[] _dADJ;
      delete[] _dTC;
      delete[] _pVAR;
      /* DO NOT FREE _pRHS, _pIC */
      delete[] _pJAC;
      delete[] _pADJ;
      delete[] _pTC;
      delete[] _mesh;
      delete[] _meshndx;
    }

  //! @brief Integrator options
  struct Options: public BASE_GSL::Options
  {
    //! @brief Constructor
    Options():
      BASE_GSL::Options(), INTERPMETH(CSPLINE), MESHPREALLOC(0), DISPLAY(0), RESRECORD(false)
      {}
    //! @brief Assignment operator
    template <typename U> Options& operator=
      ( U&options ){
        BASE_GSL::Options::operator=(options);
        INTERPMETH   = options.INTERPMETH;
        MESHPREALLOC = options.MESHPREALLOC;
        DISPLAY      = options.DISPLAY;
        RESRECORD    = options.RESRECORD;
        return *this;
      }
    //! @brief Enumeration of interpolation algorithms
    enum INTERPOLATION_METHOD{
      LINEAR=0,		//!< Linear interpolation. [Default]
      CSPLINE,		//!< Cubic spline with natural boundary conditions.
      AKIMA		//!< Non-rounded Akima spline with natural boundary conditions.
    } options;
    //! @brief Numerical integration method
    INTERPOLATION_METHOD INTERPMETH;
    //! @brief Preallocated mesh size (default: 0)
    bool MESHPREALLOC;
    //! @brief Display level
    int DISPLAY;
    //! @brief Whether or not to record results (default: false)
    bool RESRECORD;
  } options;

  //! @brief Structure for setting up storing the solver exceptions
  class Exceptions
  {
  public:
    //! @brief Enumeration type for NLPSBB exception handling
    enum TYPE{
      UNDEF=-33	//!< Error due to calling a function/feature not yet implemented
    };
    //! @brief Constructor for error <a>ierr</a>
    Exceptions( TYPE ierr ) : _ierr( ierr ){}
    //! @brief Inline function returning the error flag
    int ierr(){ return _ierr; }
    //! @brief Inline function returning the error description
    std::string what(){
      switch( _ierr ){
      case UNDEF: default:
        return "ODESLV_GSL::Exceptions  Error due to calling a not yet implemented feature";
      }
    }
  private:
    TYPE _ierr;
  };

  //! @brief Store integration bounds at a given time instant
  struct Results
  {
    //! @brief Constructors
    Results
      ( const double tk, const unsigned nxk, const T*Ixk ):
      t( tk ), nx( nxk )
      { X = new T[nx];
        for( unsigned ix=0; ix<nx; ix++ ) X[ix] = Ixk[ix]; }
    Results
      ( const double tk, const unsigned nxk, const T*Ixk,
        const unsigned nyk, const T*Iyk ):
      t( tk ), nx( nxk+nyk )
      { X = new T[nx]; unsigned i(0);
        for( unsigned ix=0; ix<nxk; ix++, i++ ) X[i] = Ixk[ix];
        for( unsigned iy=0; iy<nyk; iy++, i++ ) X[i] = Iyk[iy]; }
    Results
      ( const Results&res ):
      t( res.t ), nx( res.nx )
      { X = new T[nx];
        for( unsigned int ix=0; ix<nx; ix++ ) X[ix] = res.X[ix]; }
    //! @brief Destructor
    ~Results()
      { delete[] X; }
    //! @brief Time instant
    double t;
    //! @brief Solution dimension
    unsigned int nx;
    //! @brief Solution bounds
    T* X;
  };

  //! @brief Statistics for state integration
  Stats stats_traj;

  //! @brief Statistics for adjoint integration
  Stats stats_adj;

  //! @brief Integrate trajectory of parametric ODEs
  STATUS states
    ( const unsigned int ns, const double*tk, const double*p, double**xk,
      double*f, std::ostream&os=std::cout );

  //! @brief Integrate trajectory of parametric ODEs and adjoint ODEs
  STATUS states_ASA
    ( const unsigned int ns, const double*tk, const double*p, double**xk,
      double*f, double**lk, double*df, std::ostream&os=std::cout );

  //! @brief Compute approximate interval enclosure of reachable set of parametric ODEs using parameter sampling
  STATUS bounds
    ( const unsigned int ns, const double*tk, const T*Ip, T**Ixk,
      T*If, const unsigned int nsamp, std::ostream&os=std::cout );

  //! @brief Compute approximate interval enclosure of reachable set of parametric ODEs and adjoint ODEs using parameter sampling
  STATUS bounds_ASA
    ( const unsigned int ns, const double*tk, const T*Ip, T**Ixk,
      T*If, T**Ilk, T*Idf, const unsigned int nsamp, std::ostream&os=std::cout );

  //! @brief Record results in file <a>bndrec</a>, with accuracy of <a>iprec</a> digits
  void record
    ( std::ofstream&bndrec, const unsigned int iprec=5 ) const;
  /** @} */

  //! @brief static pointer to class
  static ODESLV_GSL<T> *pODESLV_GSL;

private:
  //! @brief Vector storing interval bound results (see Options::RESRECORD)
  std::vector< Results > _results;

  //! @brief array of vectors storing forward integration results
  std::vector<double>* _mesh;

  //! @brief array holding indexes at stage times
  unsigned* _meshndx;

  //! @brief current function
  unsigned int _ifct;

  //! @brief vector storing stepsize during adjoint integration
  std::vector<double> _h_adj;

  //! @brief Function to finalize statistics for GSL
  void _GSL_term
    ( Stats &stats );

  //! @brief Function to initialize GSL numerical integration driver
  void _GSL_init
    ( gsl_odeiv2_system &sys, gsl_odeiv2_driver *&driver );

  //! @brief Function to initialize GSL numerical integration drivers
  void _GSL_init
    ( gsl_odeiv2_system &sys, std::vector<gsl_odeiv2_driver*>&driver );

  //! @brief Function to initialize GSL interpolation driver
  void _GSL_init
    ( std::vector<gsl_interp*>&driver, const unsigned size );

  //! @brief Function to initialize GSL interpolation lookup accelerator
  void _GSL_init
    ( std::vector<gsl_interp_accel*>&accel );

  //! @brief Initialize GSL for state integration
  void _GSL_init
    ( const double* p, const bool store, const unsigned ns );

  //! @brief Initialize GSL for adjoint integration
  void _GSL_init
    ( const double* p );

  //! @brief Set RHS pointer and corresponding Jacobian
  bool _set_RHS
    ( const unsigned int iRHS );

  //! @brief Set adjoint RHS pointer and corresponding Jacobian
  bool _set_ADJRHS
    ( const unsigned int iRHS );

  //! @brief Initialize mesh storage for state trajectories
  void _mesh_init
    ( const unsigned ns );

  //! @brief Append current point to state trajectories mesh
  void _mesh_add
    ( const double&t, const double*x, const bool index );

  //! @brief Interpolate states stored along the current stage <a>_istg<\a>
  bool _mesh_interp
    ( const unsigned ns );

  //! @brief Evaluate interpolated state variable <a>i<\a> at time <a>t</a> in <a>x<\a>
  bool _mesh_eval
    ( const double t, double*x );

  //! @brief Integrate state trajectories on current stages
  STATUS _states_traj
    ( const double tf, double*xf, const bool store );

  //! @brief Integrate state trajectories on every time stages
  STATUS _states
    ( const unsigned int ns, const double*tk, const double*p, double**xk,
      double*f, const bool store, std::ostream&os );

  //! @brief Integrate adjoint trajectories on current stages
  STATUS _adjoints_traj
    ( const double tf, double&h );

  //! @brief Integrate adjoint trajectories on every time stages
  STATUS _adjoints
    ( const unsigned int ns, const double*tk, const double*p,
      const double*const*xk, double**lk, double*df, std::ostream&os );

  //! @brief Static wrapper to function to calculate the ODEs RHS values
  static int MC_GSLRHS__
    ( double t, const double* x, double* xdot, void* user_data );

  //! @brief Function to calculate the ODEs RHS values
  int _ODESLV_GSL_RHS
    ( double t, const double* x, double* xdot, void* user_data );

  //! @brief Static wrapper to function to calculate the ODEs RHS derivatives
  static int MC_GSLJAC__
    ( double t, const double* x, double* jac, double* xdot, void* user_data );

  //! @brief Function to calculate the ODEs RHS derivatives
  int _ODESLV_GSL_JAC
    ( double t, const double* x, double* jac, double* xdot, void* user_data );

  //! @brief Static wrapper to function to calculate the adjoint ODEs RHS values
  static int MC_GSLADJRHS__
    ( double t, const double* l, double* ldot, void* user_data );

  //! @brief Function to calculate the adjoint ODEs RHS values
  int _ODESLV_GSL_ADJRHS
    ( double t, const double* l, double* ldot, void* user_data );

  //! @brief Static wrapper to function to calculate the adjoint ODEs RHS derivatives
  static int MC_GSLADJJAC__
    ( double t, const double* l, double* jac, double* xdot, void* user_data );

  //! @brief Function to calculate the adjoint ODEs RHS derivatives
  int _ODESLV_GSL_ADJJAC
    ( double t, const double* l, double* jac, double* xdot, void* user_data );

  //! @brief Recursive function computing bounds on solutions of IVP in ODEs using sampling
  STATUS _states
    ( const unsigned int ns, const double*tk, const T*Ip, T**Ixk, T*If,
      const unsigned int nsamp, unsigned int* vsamp, const unsigned int ip,
      double*p, double**xk, double*f, std::ostream&os );

  //! @brief Recursive function computing bounds on solutions of IVP in ODEs using sampling
  STATUS _states_ASA
    ( const unsigned int ns, const double*tk, const T*Ip, T**Ixk, T*If,
      T**Ilk, T*Idf, const unsigned int nsamp, unsigned int* vsamp,
      const unsigned int ip, double*p, double**xk, double*f, double**lk,
      double*df, std::ostream&os );

  //! @brief Function to display intermediate results
  template<typename U> static void _print_interm
    ( const double t, const unsigned nx, const U*x, const std::string&var,
      std::ostream&os=std::cout );

  //! @brief Function to display intermediate results
  template<typename U> static void _print_interm
    ( const std::string&header, const unsigned nx, const U*x, const std::string&var,
      std::ostream&os=std::cout );

  //! @brief Private methods to block default compiler methods
  ODESLV_GSL(const ODESLV_GSL&);
  ODESLV_GSL& operator=(const ODESLV_GSL&);
};

template <typename T>
ODESLV_GSL<T>* ODESLV_GSL<T>::pODESLV_GSL = 0;

template <typename T> inline void
ODESLV_GSL<T>::_GSL_init
( const double*p, const bool store, const unsigned ns )
{
  // Define ODE system in GSL format
  _sys_traj.function = MC_GSLRHS__;
  _sys_traj.jacobian = MC_GSLJAC__;
  _sys_traj.dimension = _nx;
  _sys_traj.params = 0;
  
  // Set GSL drivers for ODE integration
  _GSL_init( _sys_traj, _driver_traj );

  // Initialize mesh
  if( store ) _mesh_init( ns );

  // Size and set DAG evaluation arrays
  if( _nVAR < _nx+_np+1 ){
    _nVAR = _nx+_np+1;
    delete[] _pVAR; _pVAR = new FFVar[_nVAR];
    delete[] _dVAR; _dVAR = new double[_nVAR];
  }
  for( unsigned ix=0; ix<_nx; ix++ ) _pVAR[ix] = _pX[ix];
  for( unsigned ip=0; ip<_np; ip++ ) _pVAR[_nx+ip] = _pP[ip];
  _pVAR[_nx+_np] = (_pT? *_pT: 0. );
  for( unsigned ip=0; ip<_np; ip++ ) _dVAR[_nx+ip] = p[ip];

  // Initialize statistics
  _init_stats( stats_traj );

  return;
}
template <typename T> inline void
ODESLV_GSL<T>::_GSL_init
( const double*p )
{
  // Define adjoint ODE system in GSL format
  _sys_adj.function = MC_GSLADJRHS__;
  _sys_adj.jacobian = MC_GSLADJJAC__;
  _sys_adj.dimension = _nx+_np;
  _sys_adj.params = 0;
  
  // Set GSL drivers for adjoint ODE integration
  _GSL_init( _sys_adj, _driver_adj );

  // Set GSL lookup accelerator for interpolation
  _GSL_init( _accel_interp );

  // Size and set DAG evaluation arrays
  if( _nVAR < 2*_nx+_np+1 ){
    _nVAR = 2*_nx+_np+1;
    delete[] _pVAR; _pVAR = new FFVar[_nVAR];
    delete[] _dVAR; _dVAR = new double[_nVAR];
  }

  set_adjoint();
  for( unsigned ix=0; ix<_nx; ix++ ) _pVAR[ix] = _pY[ix];
  for( unsigned ix=0; ix<_nx; ix++ ) _pVAR[_nx+ix] = _pX[ix];
  for( unsigned ip=0; ip<_np; ip++ ) _pVAR[2*_nx+ip] = _pP[ip];
  _pVAR[2*_nx+_np] = (_pT? *_pT: 0. );

  for( unsigned ip=0; ip<_np; ip++ ) _dVAR[2*_nx+ip] = p[ip];

  // Initialize statistics
  _init_stats( stats_adj );

  return;
}

template <typename T> inline void
ODESLV_GSL<T>::_GSL_init
( gsl_odeiv2_system &sys, gsl_odeiv2_driver *&driver )
{
  // Set GSL numerical integration driver
  if( driver ) gsl_odeiv2_driver_free( driver );
  switch( options.INTMETH ){
  case Options::RK8PD:
    driver = gsl_odeiv2_driver_alloc_y_new( &sys, gsl_odeiv2_step_rk8pd,
      options.H0, options.ATOL, options.RTOL );
    break;
  case Options::MSADAMS:
    driver = gsl_odeiv2_driver_alloc_y_new( &sys, gsl_odeiv2_step_msadams,
      options.H0, options.ATOL, options.RTOL );
    break;
  case Options::MSBDF:
    driver = gsl_odeiv2_driver_alloc_y_new( &sys, gsl_odeiv2_step_msbdf,
      options.H0, options.ATOL, options.RTOL );
    break;
  case Options::RKF45: default:
    driver = gsl_odeiv2_driver_alloc_y_new( &sys, gsl_odeiv2_step_rkf45,
      options.H0, options.ATOL, options.RTOL );
    break;
  }
  gsl_odeiv2_driver_set_hmin( driver, options.HMIN );  
  gsl_odeiv2_driver_set_nmax( driver, options.NMAX );  

  // Set state storage
  delete [] _vec_state;
  _vec_state = new double[ sys.dimension ];

  return;
}

template <typename T> inline void
ODESLV_GSL<T>::_GSL_init
( gsl_odeiv2_system &sys, std::vector<gsl_odeiv2_driver*>&driver )
{
  // Reset GSL numerical integration drivers
  for( unsigned i=0; i<driver.size(); i++ )
    gsl_odeiv2_driver_free( driver[i] );
  driver.clear();

  for( unsigned i=0; i<_nf; i++ ){
    switch( options.INTMETH ){
    case Options::RK8PD:
      driver.push_back( gsl_odeiv2_driver_alloc_y_new( &sys,
        gsl_odeiv2_step_rk8pd, options.H0, options.ATOL, options.RTOL ) );
      break;
    case Options::MSADAMS:
      driver.push_back( gsl_odeiv2_driver_alloc_y_new( &sys,
        gsl_odeiv2_step_msadams, options.H0, options.ATOL, options.RTOL ) );
      break;
    case Options::MSBDF:
      driver.push_back( gsl_odeiv2_driver_alloc_y_new( &sys,
        gsl_odeiv2_step_msbdf, options.H0, options.ATOL, options.RTOL ) );
      break;
    case Options::RKF45: default:
      driver.push_back( gsl_odeiv2_driver_alloc_y_new( &sys,
        gsl_odeiv2_step_rkf45, options.H0, options.ATOL, options.RTOL ) );
      break;
    }
    gsl_odeiv2_driver_set_hmin( driver[i], options.HMIN );  
    gsl_odeiv2_driver_set_nmax( driver[i], options.NMAX );  
  }

  // Set intermediate storage
  delete [] _vec_state; _vec_state = new double[ sys.dimension ];
  delete [] _dTC; _dTC = new double[ sys.dimension*_nf ];

  return;
}

template <typename T> inline void
ODESLV_GSL<T>::_GSL_init
( std::vector<gsl_interp_accel*>&accel )
{
  // Set GSL interpolation lookup accelerator - expand vector if necessary
  for( unsigned i=accel.size(); i<_nx; i++ )
    accel.push_back( gsl_interp_accel_alloc() );
}

template <typename T> inline void
ODESLV_GSL<T>::_GSL_init
( std::vector<gsl_interp*>&driver, const unsigned size )
{
  // Reset GSL interpolation drivers
  for( unsigned i=0; i<driver.size(); i++ )
    gsl_interp_free( driver[i] );
  driver.clear();

  for( unsigned i=0; i<_nx; i++ )
    switch( options.INTERPMETH ){
    case Options::AKIMA:
      if( size >= 5 ){
        driver.push_back( gsl_interp_alloc( gsl_interp_akima, size ) );
        break;
      }
    case Options::CSPLINE:
      if( size >= 3 ){
        driver.push_back( gsl_interp_alloc( gsl_interp_cspline, size ) );
        break;
      }
    case Options::LINEAR: default:
      driver.push_back( gsl_interp_alloc( gsl_interp_linear, size ) );
      break;
    }
}

template <typename T> inline void
ODESLV_GSL<T>::_mesh_init
( const unsigned ns )
{
  // Allocate mesh for holding intermediate states
  if( !_mesh ) _mesh = new std::vector<double>[_sys_traj.dimension+1];
  for( unsigned ix=0; options.MESHPREALLOC && ix<=_sys_traj.dimension; ix++ )
    { _mesh[ix].clear(); _mesh[ix].reserve( options.MESHPREALLOC ); }

  // Allocate array for holding index at stage times
  delete[] _meshndx; _meshndx = new unsigned[ns];
}

template <typename T> inline bool
ODESLV_GSL<T>::_mesh_interp
( const unsigned ns )
{
  const unsigned offset = _meshndx[_istg-1];
  const unsigned size = ( _istg==ns? _mesh[0].size(): _meshndx[_istg] ) - offset; 

  // Initialize interpolation driver
  _GSL_init( _driver_interp, size );

  // Interpolate every state in current stage
  for( unsigned i=0; i<_nx; i++ ){
#ifdef MC__ODESLV_GSL_DEBUG_INTERP
    std::cout << "Stage #" << _istg << "  State #" << i << std::endl; 
    std::cout << _mesh[0].data() << "  " << _mesh[i+1].data() << std::endl;
    for( unsigned k=0; k<size; k++ )
      std::cout << (_mesh[0].data()+offset)[k] << "  " << (_mesh[i+1].data()+offset)[k] << std::endl;
    { int dum; std::cin >> dum; }
#endif
    int flag = gsl_interp_init( _driver_interp[i], _mesh[0].data()+offset,
                                _mesh[i+1].data()+offset, size );
#ifdef MC__ODESLV_GSL_DEBUG_INTERP
    std::cout << "flag: " << flag << std::endl;
#endif
    if( flag != GSL_SUCCESS ) return false;
  }
  // !!!gsl_interp_init returns an int flag, but it is unclear what it is!!!
  // Function: int gsl_interp_init (gsl_interp * interp, const double xa[], const double ya[], size_t size)

  // Reset lookup accelerator
#ifdef MC__ODESLV_GSL_CHECK
  if( _accel_interp.size() < _nx ) return false;
#endif
  for( unsigned i=0; i<_nx; i++ )
    if( gsl_interp_accel_reset( _accel_interp[i] ) != GSL_SUCCESS ) return false;

  return true;
}

template <typename T> inline void
ODESLV_GSL<T>::_mesh_add
( const double&t, const double*x, const bool index )
{
  // Index current location
  if( index ) _meshndx[_istg] = _mesh[0].size();

  // Store time and state into mesh
  _mesh[0].push_back( t );
  for( unsigned ix=0; ix<_sys_traj.dimension; ix++ )
    _mesh[ix+1].push_back( x[ix] );
}

template <typename T> inline bool
ODESLV_GSL<T>::_mesh_eval
( const double t, double*x )
{
  const unsigned offset = _meshndx[_istg-1];
#ifdef MC__ODESLV_GSL_CHECK
  if( _driver_interp.size() != _nx ) return false;
#endif
  for( unsigned i=0; i<_nx; i++ )
    if( gsl_interp_eval_e( _driver_interp[i], _mesh[0].data()+offset,
        _mesh[i+1].data()+offset, t, _accel_interp[i], x+i ) == GSL_EDOM )
      return false;
  return true;
}

template <typename T> inline void
ODESLV_GSL<T>::_GSL_term
( Stats& stats )
{
  // Get final CPU time
  _final_stats( stats );
}

template <typename T> inline int
ODESLV_GSL<T>::_ODESLV_GSL_RHS
( double t, const double* x, double* xdot, void* user_data )
{
  stats_traj.numRHS++; // increment RHS counter
  if( !_pRHS ) return GSL_EBADFUNC; // **error** RHS not defined
  for( unsigned i=0; i<_nx; i++ ) _dVAR[i] = x[i]; // set current state values
  _dVAR[_nx+_np] = t; // set current time
  _pDAG->eval( _opRHS, _dRHS, _nx, _pRHS, xdot, _nVAR, _pVAR, _dVAR );

  return GSL_SUCCESS;  
}

template <typename T> inline int
ODESLV_GSL<T>::MC_GSLRHS__
( double t, const double* x, double* xdot, void* user_data )
{
  ODESLV_GSL<T> *pODESLV_GSL = ODESLV_GSL<T>::pODESLV_GSL;
  int flag = pODESLV_GSL->_ODESLV_GSL_RHS( t, x, xdot, user_data );
  ODESLV_GSL<T>::pODESLV_GSL = pODESLV_GSL;
  return flag;
}

template <typename T> inline int
ODESLV_GSL<T>::_ODESLV_GSL_JAC
( double t, const double* x, double* jac, double* xdot, void* user_data )
{
  stats_traj.numJAC++;
  if( !_pRHS || !_pJAC ) return GSL_EBADFUNC; // **error** RHS or JAC not defined
  for( unsigned i=0; i<_nx; i++ ) _dVAR[i] = x[i]; // set current state values
  _dVAR[_nx+_np] = t; // set current time
  _pDAG->eval( _opRHS, _dRHS, _nx, _pRHS, xdot, _nVAR, _pVAR, _dVAR );
  _pDAG->eval( _opJAC, _dJAC, _nx*_nx, _pJAC, jac, _nVAR, _pVAR, _dVAR );

  return GSL_SUCCESS;  
}

template <typename T> inline int
ODESLV_GSL<T>::MC_GSLJAC__
( double t, const double* x, double* jac, double* xdot, void* user_data )
{
  ODESLV_GSL<T> *pODESLV_GSL = ODESLV_GSL<T>::pODESLV_GSL;
  int flag = pODESLV_GSL->_ODESLV_GSL_JAC( t, x, jac, xdot, user_data );
  ODESLV_GSL<T>::pODESLV_GSL = pODESLV_GSL;
  return flag;
}

template <typename T> inline bool
ODESLV_GSL<T>::_set_RHS
( const unsigned int iRHS )
{
  if( _vRHS.size() <= iRHS ) return false;

  _pRHS = _vRHS.at( iRHS );
  _opRHS = _pDAG->subgraph( _nx, _pRHS );
  delete[] _dRHS; _dRHS = new double[ _opRHS.size() ];
  
  switch( options.INTMETH ){
  case Options::MSADAMS:
  case Options::MSBDF:
    delete[] _pJAC; _pJAC  = _pDAG->FAD( _nx, _pRHS, _nx, _pX );
    _opJAC = _pDAG->subgraph( _nx*_nx, _pJAC );
    delete[] _dJAC; _dJAC  = new double[ _opJAC.size() ];
    break;
  case Options::RK8PD:
  case Options::RKF45:
  default:
    delete[] _pJAC; _pJAC = 0;
    _opJAC.clear();
    delete[] _dJAC; _dJAC = 0;
    break;
  }

  return true;
}

template <typename T> inline int
ODESLV_GSL<T>::_ODESLV_GSL_ADJRHS
( double t, const double* y, double* ydot, void* user_data )
{
  stats_adj.numRHS++; // increment RHS counter
  if( !_pADJ ) return GSL_EBADFUNC; // **error** RHS not defined
  if( !_mesh_eval( -t, _dVAR+_nx ) ) return GSL_EBADFUNC; // set interpolated state values
  for( unsigned i=0; i<_nx; i++ ) _dVAR[i] = y[i]; // set current adjoint/quadrature values
  _dVAR[2*_nx+_np] = -t; // set current time
  _pDAG->eval( _opADJ, _dADJ, _nx+_np, _pADJ, ydot, _nVAR, _pVAR, _dVAR );

  return GSL_SUCCESS;  
}

template <typename T> inline int
ODESLV_GSL<T>::MC_GSLADJRHS__
( double t, const double* y, double* ydot, void* user_data )
{
  ODESLV_GSL<T> *pODESLV_GSL = ODESLV_GSL<T>::pODESLV_GSL;
  int flag = pODESLV_GSL->_ODESLV_GSL_ADJRHS( t, y, ydot, user_data );
  ODESLV_GSL<T>::pODESLV_GSL = pODESLV_GSL;
  return flag;
}

template <typename T> inline int
ODESLV_GSL<T>::_ODESLV_GSL_ADJJAC
( double t, const double* y, double* jac, double* ydot, void* user_data )
{
  stats_adj.numJAC++;
  if( !_pADJ || !_pJAC ) return GSL_EBADFUNC; // **error** RHS or JAC not defined
  if( !_mesh_eval( -t, _dVAR+_nx ) ) return GSL_EBADFUNC; // set interpolated state values
  for( unsigned i=0; i<_nx; i++ ) _dVAR[i] = y[i]; // set current adjoint/quadrature values
  _dVAR[2*_nx+_np] = -t; // set current time
  _pDAG->eval( _opADJ, _dADJ, _nx+_np, _pADJ, ydot, _nVAR, _pVAR, _dVAR );
  _pDAG->eval( _opJAC, _dJAC, (_nx+_np)*(_nx+_np), _pJAC, jac, _nVAR, _pVAR, _dVAR );

  return GSL_SUCCESS;  
}

template <typename T> inline int
ODESLV_GSL<T>::MC_GSLADJJAC__
( double t, const double* y, double* jac, double* ydot, void* user_data )
{
  ODESLV_GSL<T> *pODESLV_GSL = ODESLV_GSL<T>::pODESLV_GSL;
  int flag = pODESLV_GSL->_ODESLV_GSL_ADJJAC( t, y, jac, ydot, user_data );
  ODESLV_GSL<T>::pODESLV_GSL = pODESLV_GSL;
  return flag;
}

template <typename T> inline bool
ODESLV_GSL<T>::_set_ADJRHS
( const unsigned int iRHS )
{
  if( _vRHS.size() <= iRHS ) return false;

  _pRHS = _vRHS.at( iRHS );
  FFVar pHAM( 0. );
  for( unsigned ix=0; ix<_nx; ix++ ) pHAM += _pVAR[ix] * _pRHS[ix];
  delete[] _pADJ; _pADJ = _pDAG->BAD( 1, &pHAM, _nx+_np, _pVAR+_nx );
  _opADJ = _pDAG->subgraph( _nx+_np, _pADJ );
  
  switch( options.INTMETH ){
  case Options::MSADAMS:
  case Options::MSBDF:
    delete[] _pJAC; _pJAC  = _pDAG->FAD( _nx+_np, _pADJ, _nx+_np, _pVAR+_nx );
    _opJAC = _pDAG->subgraph( (_nx+_np)*(_nx+_np), _pJAC );
    delete[] _dJAC; _dJAC  = new double[ _opJAC.size() ];
    break;
  case Options::RK8PD:
  case Options::RKF45:
  default:
    delete[] _pJAC; _pJAC = 0;
    _opJAC.clear();
    delete[] _dJAC; _dJAC = 0;
    break;
  }

  delete[] _dADJ; _dADJ = new double[ _opADJ.size() ];

  return true;
}

template <typename T> inline typename ODESLV_GSL<T>::STATUS
ODESLV_GSL<T>::_states_traj
( const double tf, double*xf, const bool store )
{
  // Store initial point
  if( store ) _mesh_add( _t, xf, true );

  // integrate till end of time stage
  while( _t < tf ){
    if( gsl_odeiv2_evolve_apply( _driver_traj->e, _driver_traj->c, _driver_traj->s,
        &_sys_traj, &_t, tf, &_h, xf ) != GSL_SUCCESS
     || _h < options.HMIN
     || (options.NMAX && stats_traj.numSteps > options.NMAX) ) return FAILURE;
    stats_traj.numSteps++;
    if( options.HMAX > 0 && _h > options.HMAX ) _h = options.HMAX;

    // Store intermediate point
    if( store ) _mesh_add( _t, xf, false );
  }
  return NORMAL;
}

template <typename T> inline typename ODESLV_GSL<T>::STATUS
ODESLV_GSL<T>::_states
( const unsigned int ns, const double*tk, const double*p, double**xk,
  double*f, const bool store, std::ostream&os )
{
  // Initialize trajectory integration with GSL
  _GSL_init( p, store, ns );

  // States initial conditions
  if( !_vIC.size() )
    { _GSL_term( stats_traj ); return FATAL; }
  _pIC = _vIC.at(0);
  _opIC = _pDAG->subgraph( _nx, _pIC );
  _dVAR[_nx+_np] = tk[0]; // set initial time
  _pDAG->eval( _opIC, _nx, _pIC, xk[0], _np+1, _pVAR+_nx, _dVAR+_nx );
  if( options.DISPLAY >= 1 ) _print_interm( tk[0], _nx, xk[0], "x", os );

  // Integrate ODEs through each stage using GSL
  _t = tk[0];
  _h = options.H0;
  pODESLV_GSL = this;

  for( _istg=0; _istg<ns; _istg++ ){

    // Reinitialize states at intermediate times
    if( _istg && _vIC.size() > 1 ){
      _pIC = _vIC.at(_istg);
      _opIC = _pDAG->subgraph( _nx, _pIC );
      for( unsigned i=0; i<_nx; i++ ) _dVAR[i] = xk[_istg][i]; // set current state values
      _dVAR[_nx+_np] = tk[_istg+1]; // set current time
      _pDAG->eval( _opIC, _nx, _pIC, xk[_istg+1], _nVAR, _pVAR, _dVAR );
      // Reset ODE solver - needed to account for discontinuity
      gsl_odeiv2_driver_reset( _driver_traj );
    }
    else
      for( unsigned int ix=0; ix<_nx; ix++ ) xk[_istg+1][ix] = xk[_istg][ix];

    // Update list of operations in RHS and JAC
    const unsigned int iRHS = ( _vRHS.size()<2? 0: _istg );
    if( (!_istg || iRHS) && !_set_RHS( iRHS ) )
      { _GSL_term( stats_traj ); return FATAL; }

    // Integrate until end of time stage
    if( _states_traj( tk[_istg+1], xk[_istg+1], store ) == FAILURE )
      { _GSL_term( stats_traj ); return FAILURE; }
    if( options.DISPLAY >= 1 ) _print_interm( tk[_istg+1], _nx, xk[_istg+1], "x", os );
  }

  // Evaluate functions
  if( f && _vFCT.size() == 1 ){
    const FFVar* pFCT = _vFCT.at( 0 );
    for( unsigned i=0; i<_nx; i++ ) _dVAR[i] = xk[ns][i]; // set final state values
    _dVAR[_nx+_np] = tk[ns]; // set final time
    _pDAG->eval( _nf, pFCT, f, _nx+_np+1, _pVAR, _dVAR, false );
    if( options.DISPLAY >= 1 ) _print_interm( "functions", _nf, f, "f", os );
  }
  else if( f && _vFCT.size() >= ns ){
    for( unsigned int istg=0; istg<ns; istg++ ){
      const FFVar* pFCT = _vFCT.at( istg );
      for( unsigned i=0; i<_nx; i++ ) _dVAR[i] = xk[istg+1][i]; // set intermediate state values
      _dVAR[_nx+_np] = tk[istg+1]; // set stage time
      _pDAG->eval( _nf, pFCT, f, _nx+_np+1, _pVAR, _dVAR, istg?true:false );
    }
    if( options.DISPLAY >= 1 ) _print_interm( "function", _nf, f, "f", os );
  }
  else if( f ){ _GSL_term( stats_traj ); return FATAL; }

  _GSL_term( stats_traj );
  if( options.DISPLAY >= 1 ) _print_stats( stats_traj, os );
  return NORMAL;
}

template <typename T> inline typename ODESLV_GSL<T>::STATUS
ODESLV_GSL<T>::_adjoints_traj
( const double tf, double&h )
{
  // integrate till end of time stage
#ifdef MC__ODESLV_GSL_DEBUG_ADJOINT
    std::cout << "adjoint #" << _ifct << "  " << "stage #" << _istg << "  times:\n";
#endif
  while( _t < tf ){
    if( gsl_odeiv2_evolve_apply( _driver_adj[_ifct]->e, _driver_adj[_ifct]->c,
        _driver_adj[_ifct]->s, &_sys_adj, &_t, tf, &h, _vec_state ) != GSL_SUCCESS
     || h < options.HMIN
     || (options.NMAX && stats_adj.numSteps > options.NMAX) ) return FAILURE;
    stats_adj.numSteps++;
    if( options.HMAX > 0 && h > options.HMAX ) h = options.HMAX;
#ifdef MC__ODESLV_GSL_DEBUG_ADJOINT
    std::cout << -_t << std::endl;
#endif
  }
  return NORMAL;
}

template <typename T> inline typename ODESLV_GSL<T>::STATUS
ODESLV_GSL<T>::_adjoints
( const unsigned int ns, const double*tk, const double*p, const double*const*xk,
  double**lk, double*q, std::ostream&os )
{
  // Initialize trajectory integration with GSL
  _GSL_init( p );

  // Adjoint terminal conditions
  if( _vFCT.size() != 1 && _vFCT.size() < ns )
    { _GSL_term( stats_adj ); return FATAL; }
  const FFVar* pFCT = _vFCT.size()>1? _vFCT.at(ns-1): _vFCT.at(0);
  delete[] _pTC; _pTC = _pDAG->FAD( _nf, pFCT, _nx+_np, _pVAR+_nx );
  _opTC = _pDAG->subgraph( (_nx+_np)*_nf, _pTC );

  for( unsigned i=0; i<_nx; i++ ) _dVAR[_nx+i] = xk[ns][i]; // set current state values
  _dVAR[2*_nx+_np] = tk[ns]; // set current time
  _pDAG->eval( _opTC, (_nx+_np)*_nf, _pTC, _dTC, _nx+_np+1, _pVAR+_nx, _dVAR+_nx );

  for( unsigned k=0, ik=0; k<_nf; k++ ){
    for( unsigned i=0; i<_nx; i++, ik++ ) lk[ns][k*_nx+i] = _dTC[ik];
    for( unsigned i=0; i<_np; i++, ik++ ) q[k*_np+i] = _dTC[ik];
  }
  if( options.DISPLAY >= 1 ){
    _print_interm( tk[ns], _nf*_nx, lk[ns], "l", os );
    //_print_interm( tk[ns], _nf*_np, q, "q", os );
  }

  // Integrate adjoint ODEs backward through each stage using GSL
  _h_adj.assign( _nf, options.H0 );
  pODESLV_GSL = this;

  for( _istg=ns; _istg>0; _istg-- ){

    // Adjoint discontinuity at stage times
    if( _istg<ns ){

      if( _vFCT.size() > 1 ){
        pFCT = _vFCT.at(_istg-1);
        delete[] _pTC; _pTC = _pDAG->FAD( _nf, pFCT, _nx+_np, _pVAR+_nx );
        _opTC = _pDAG->subgraph( (_nx+_np)*_nf, _pTC );

        _dVAR[2*_nx+_np] = tk[_istg]; // set current time
        for( unsigned i=0; i<_nx; i++ ) _dVAR[_nx+i] = xk[_istg][i]; // set current state values
        _pDAG->eval( _opTC, (_nx+_np)*_nf, _pTC, _dTC, _nx+_np+1, _pVAR+_nx, _dVAR+_nx );
        // Reset adjoint ODE solver
        // for( _ifct=0; _ifct<_nf; _ifct++ ) gsl_odeiv2_driver_reset( _driver_adj[_ifct] );
      }
      else
        for( unsigned i=0; i<(_nx+_np)*_nf; i++ ) _dTC[i] = 0.;

      // State discontinuity contribution
      if( _vIC.size() > 1 ){
        _pIC = _vIC.at(_istg);
        FFVar pHAM( 0. );
        for( unsigned ix=0; ix<_nx; ix++ ) pHAM += _pVAR[ix] * _pIC[ix];
        delete[] _pTC; _pTC = _pDAG->BAD( 1, &pHAM, _nx+_np, _pVAR+_nx );
        _opTC = _pDAG->subgraph( _nx+_np, _pTC );

        _dVAR[2*_nx+_np] = tk[_istg]; // set current time
        for( _ifct=0; _ifct<_nf; _ifct++ ){
          for( unsigned i=0; i<_nx; i++ ){
            _dVAR[i] = lk[_istg][_ifct*_nx+i] + _dTC[_ifct*(_nx+_np)+i]; // l(ti+)+dphi/dx(ti)
            _dTC[_ifct*(_nx+_np)+i] = 0.;
            _dVAR[_nx+i] = xk[_istg][i]; // set current state values
          }
          for( unsigned i=0; i<_np; i++ )
            _dTC[_ifct*(_nx+_np)+_nx+i] += q[_ifct*_np+i];
          _pDAG->eval( _opTC, _nx+_np, _pTC, _dTC+_ifct*(_nx+_np), _nVAR, _pVAR, _dVAR, true ); // add result to _dTC
          // Reset adjoint ODE solver
          // gsl_odeiv2_driver_reset( _driver_adj[_ifct] );
        }
      }
      else
        for( _ifct=0; _ifct<_nf; _ifct++ ){
          for( unsigned i=0; i<_nx; i++ ) _dTC[_ifct*(_nx+_np)+i] += lk[_istg][_ifct*_nx+i];
          for( unsigned i=0; i<_np; i++ ) _dTC[_ifct*(_nx+_np)+_nx+i] += q[_ifct*_np+i];
        }
    }

    // Update list of operations in adjoint RHS and JAC
    const unsigned int iRHS = ( _vRHS.size()<2? 0: _istg-1 );
    if( (_istg==ns || iRHS) && !_set_ADJRHS( iRHS ) )
      { _GSL_term( stats_adj ); return FATAL; }

    // Interpolate state mesh
    if( !_mesh_interp( ns ) )
      { _GSL_term( stats_adj ); return FAILURE; }

    // Integrate backward along time stage for each function
    for( _ifct=0; _ifct<_nf; _ifct++ ){
      _t = -tk[_istg]; // need to reinitialize stepsize too???
      for( unsigned i=0; i<_nx+_np; i++ ) _vec_state[i] = _dTC[_ifct*(_nx+_np)+i];
      if( _adjoints_traj( -tk[_istg-1], _h_adj[_ifct] ) == FAILURE )
        { _GSL_term( stats_adj ); return FAILURE; }
      for( unsigned i=0; i<_nx; i++ ) lk[_istg-1][_ifct*_nx+i] = _vec_state[i];
      for( unsigned i=0; i<_np; i++ ) q[_ifct*_np+i] = _vec_state[_nx+i];
    }

    // Initial state contribution
    if( _istg==1 ){
      _pIC = _vIC.at(0);
      FFVar pHAM( 0. );
      for( unsigned ix=0; ix<_nx; ix++ ) pHAM += _pVAR[ix] * _pIC[ix];
      delete[] _pTC; _pTC = _pDAG->BAD( 1, &pHAM, _np, _pVAR+2*_nx );
      _opTC = _pDAG->subgraph( _np, _pTC );

      _dVAR[2*_nx+_np] = tk[0]; // set initial time
      for( _ifct=0; _ifct<_nf; _ifct++ ){
        for( unsigned i=0; i<_nx; i++ ) _dVAR[i] = lk[0][_ifct*_nx+i]; // set initial adjoint
        _pDAG->eval( _opTC, _np, _pTC, q+_ifct*_np, _nVAR, _pVAR, _dVAR, true ); // add result to q
      }
    }

    if( options.DISPLAY >= 1 ){
      _print_interm( tk[_istg-1], _nf*_nx, lk[_istg-1], "l", os );
      //_print_interm( tk[_istg-1], _nf*_np, q, "q", os );
    }
  }

  if( options.DISPLAY >= 1 )
    _print_interm( "gradient", _nf*_np, q, "df", os );

  _GSL_term( stats_adj );
  if( options.DISPLAY >= 1 ) _print_stats( stats_adj, os );
  return NORMAL;
}


template <typename T> inline typename ODESLV_GSL<T>::STATUS
ODESLV_GSL<T>::_states
( const unsigned int ns, const double*tk, const T*Ip, T**Ixk, T*If,
  const unsigned int nsamp, unsigned int* vsamp, const unsigned int ip,
  double*p, double**xk, double*f, std::ostream&os )
{
  STATUS flag = NORMAL;

  // Update bounds for all sampling points
  for( unsigned int isamp=0; isamp<nsamp; isamp++ ){
    vsamp[ip] = isamp;

    // Continue recursive call
    if( ip+1 < _np ){
      flag = _states( ns, tk, Ip, Ixk, If, nsamp, vsamp, ip+1, p, xk, f, os );
      if( flag != NORMAL ) return flag;
      continue;
    }

    // Update bounds for current point
#ifdef MC__ODESLV_GSL_SAMPLE_DEBUG
    std::cout << "Sample: ";
#endif
    for( unsigned int ip=0; ip<_np; ip++ ){
      p[ip] = Op<T>::l( Ip[ip] ) + vsamp[ip]/(nsamp-1.) * Op<T>::diam( Ip[ip] );
#ifdef MC__ODESLV_GSL_SAMPLE_DEBUG
      std::cout << p[ip] << "  ";
#endif
    }
#ifdef MC__ODESLV_GSL_SAMPLE_DEBUG
    std::cout << std::endl;
#endif
    flag = states( ns, tk, p, xk, f, os );
    if( flag != NORMAL ) return flag;
    for( unsigned int is=0; is<=ns; is++ )
      for( unsigned int ix=0; ix<_nx; ix++ )
        Ixk[is][ix] = Op<T>::hull( xk[is][ix], Ixk[is][ix] );
    for( unsigned int ifn=0; If && ifn<_nf; ifn++ )
      If[ifn] = Op<T>::hull( f[ifn], If[ifn] );
  }

  return flag;
}  

template <typename T> inline typename ODESLV_GSL<T>::STATUS
ODESLV_GSL<T>::_states_ASA
( const unsigned int ns, const double*tk, const T*Ip, T**Ixk, T*If,
  T**Ilk, T*Idf, const unsigned int nsamp, unsigned int* vsamp,
  const unsigned int ip, double*p, double**xk, double*f, double**lk,
  double*df, std::ostream&os )
{
  STATUS flag = NORMAL;

  // Update bounds for all sampling points
  for( unsigned int isamp=0; isamp<nsamp; isamp++ ){
    vsamp[ip] = isamp;

    // Continue recursive call
    if( ip+1 < _np ){
      flag = _states_ASA( ns, tk, Ip, Ixk, If, Ilk, Idf, nsamp, vsamp, ip+1, p, xk, f, lk, df, os );
      if( flag != NORMAL ) return flag;
      continue;
    }

    // Update bounds for current point
#ifdef MC__ODESLV_GSL_SAMPLE_DEBUG
    std::cout << "Sample: ";
#endif
    for( unsigned int ip=0; ip<_np; ip++ ){
      p[ip] = Op<T>::l( Ip[ip] ) + vsamp[ip]/(nsamp-1.) * Op<T>::diam( Ip[ip] );
#ifdef MC__ODESLV_GSL_SAMPLE_DEBUG
      std::cout << p[ip] << "  ";
#endif
    }
#ifdef MC__ODESLV_GSL_SAMPLE_DEBUG
    std::cout << std::endl;
#endif
    flag = states_ASA( ns, tk, p, xk, f, lk, df, os );
    if( flag != NORMAL ) return flag;
    for( unsigned int is=0; is<=ns; is++ ){
      for( unsigned int ix=0; ix<_nx; ix++ )
        Ixk[is][ix] = Op<T>::hull( xk[is][ix], Ixk[is][ix] );
      for( unsigned int ix=0; ix<_nf*_nx; ix++ )
        Ilk[is][ix] = Op<T>::hull( lk[is][ix], Ilk[is][ix] );
    }
    for( unsigned int ifn=0; If && ifn<_nf; ifn++ ){
      If[ifn] = Op<T>::hull( f[ifn], If[ifn] );
      for( unsigned int ip=0; ip<_np; ip++ )
        Idf[ifn*_np+ip] = Op<T>::hull( df[ifn*_np+ip], Idf[ifn*_np+ip] );
    }
  }

  return flag;
}

//! @fn template <typename T> inline typename ODESLV_GSL<T>::STATUS ODESLV_GSL<T>::states(
//! const unsigned int ns, const double*tk, const double*p, double**xk,
//! double*f, std::ostream&os )
//!
//! This function computes the solution of the parametric ODEs defined in IVP:
//!   - <a>ns</a> [input]  number of time stages
//!   - <a>tk</a> [input]  stage times, including the initial time
//!   - <a>p</a>  [input]  parameter values
//!   - <a>xk</a> [output] state values at stage times
//!   - <a>f</a>  [output] function values (only if f != 0)
//!   - <a>os</a> [input]  output stream
//!   .
//! The return value is the status.
template <typename T> inline typename ODESLV_GSL<T>::STATUS
ODESLV_GSL<T>::states
( const unsigned int ns, const double*tk, const double*p, double**xk,
  double*f, std::ostream&os )
{
  return _states( ns, tk, p, xk, f, false, os );
}

//! @fn template <typename T> inline typename ODESLV_GSL<T>::STATUS ODESLV_GSL<T>::states_ASA(
//! const unsigned int ns, const double*tk, const double*p, double**xk,
//! double*f, double**lk, double*df, std::ostream&os )
//!
//! This function computes the solution of the parametric ODEs defined in IVP:
//!   - <a>ns</a> [input] number of time stages
//!   - <a>tk</a> [input] stage times, including the initial time
//!   - <a>p</a> [input] parameter values
//!   - <a>xk</a> [output] state values at stage times
//!   - <a>f</a>  [output] function values (only if f != 0)
//!   - <a>lk</a> [output] adjoint values at stage times
//!   - <a>f</a>  [output] function derivatives (stored row-wise)
//!   - <a>os</a> [input] output stream
//!   .
//! The return value is the status.
template <typename T> inline typename ODESLV_GSL<T>::STATUS
ODESLV_GSL<T>::states_ASA
( const unsigned int ns, const double*tk, const double*p, double**xk,
  double*f, double**lk, double*df, std::ostream&os )
{
  STATUS flag = NORMAL;

  flag = _states( ns, tk, p, xk, f, true, os );
  if( flag != NORMAL ) return flag;

  flag = _adjoints( ns, tk, p, xk, lk, df, os );
  return flag;
}

//! @fn template <typename T> inline typename ODESLV_GSL<T>::STATUS ODESLV_GSL<T>::bounds(
//! const unsigned int ns, const double*tk, const T*Ip, T**Ixk, T*f,
//! const unsigned int nsamp, std::ostream&os )
//!
//! This function computes an approximate interval enclosure of the
//! reachable set of the parametric ODEs defined in IVP using equally
//! spaced samples:
//!   - <a>ns</a> [input] number of time stages
//!   - <a>tk</a> [input] stage times, including the initial time
//!   - <a>Ip</a> [input] interval parameter set
//!   - <a>Ixk</a> [output] approximate interval state enclosures at stage times
//!   - <a>If</a>  [output] approximate function enclosure (only if If != 0)
//!   - <a>nsamp</a> [input] number of samples for each parameter
//!   - <a>os</a> [input] output stream
//! .
//! The return value is the status.
template <typename T> inline typename ODESLV_GSL<T>::STATUS
ODESLV_GSL<T>::bounds
( const unsigned int ns, const double*tk, const T*Ip, T**Ixk,
  T*If, const unsigned int nsamp, std::ostream&os )
{
  int DISPLAY_SAVE = options.DISPLAY;
  options.DISPLAY = 0;
  STATUS flag = NORMAL;
  
  // Initialization of sampled bounds at parameter lower bound
  double *p = new double[_np];
  for( unsigned int ip=0; ip<_np; ip++ )
    p[ip] = Op<T>::l(Ip[ip]);
  double **xk = new double*[ns+1];
  for( unsigned int is=0; is<=ns; is++ )
    xk[is] = new double[_nx];
  double *f = If? new double[_nf]: 0;
  flag = states( ns, tk, p, xk, f, os );
  if( flag != NORMAL || nsamp <= 1 ){
    delete[] p; delete[] f;
    for( unsigned int is=0; is<=ns; is++ ) delete[] xk[is]; delete[] xk;
    return flag;
  }   
  for( unsigned int is=0; is<=ns; is++ )
    for( unsigned int ix=0; ix<_nx; ix++ )
      Ixk[is][ix] = xk[is][ix];
  for( unsigned int ifn=0; If && ifn<_nf; ifn++ )
    If[ifn] = f[ifn];
  
  // Start sampling process
  unsigned int* vsamp = new unsigned int[_np];
  flag = _states( ns, tk, Ip, Ixk, If, nsamp, vsamp, 0, p, xk, f, os );

  // Display results
  options.DISPLAY = DISPLAY_SAVE;
  if( options.DISPLAY >= 1 ){
    for( unsigned int is=0; is<=ns; is++ )
      _print_interm( tk[is], _nx, Ixk[is], "x", os );
    if( If ) _print_interm( "function", _nf, If, "f", os );
  }

  // Record intermediate results
  _results.clear();
  if( options.RESRECORD )
    for( unsigned int is=0; is<=ns; is++ )
      _results.push_back( Results( tk[is], _nx, Ixk[is] ) );
  
  // Clean-up
  delete[] p; delete[] f;
  for( unsigned int is=0; is<=ns; is++ ) delete[] xk[is]; delete[] xk;
  delete[] vsamp;
  
  return flag;
}

//! @fn template <typename T> inline typename ODESLV_GSL<T>::STATUS ODESLV_GSL<T>::bounds_ASA(
//! const unsigned int ns, const double*tk, const T*Ip, T**Ixk, T*If,
//! T**Ilk, T*df, const unsigned int nsamp, std::ostream&os )
//!
//! This function computes an approximate interval enclosure of the
//! reachable set of the parametric ODEs defined in IVP using equally
//! spaced samples:
//!   - <a>ns</a> [input] number of time stages
//!   - <a>tk</a> [input] stage times, including the initial time
//!   - <a>Ip</a> [input] interval parameter set
//!   - <a>Ixk</a> [output] approximate interval state enclosures at stage times
//!   - <a>If</a>  [output] approximate function enclosure (only if If != 0)
//!   - <a>Ilk</a> [output] approximate interval adjoint enclosures at stage times
//!   - <a>Idf</a>  [output] approximate function derivative enclosures (stored row-wise)
//!   - <a>nsamp</a> [input] number of samples for each parameter
//!   - <a>os</a> [input] output stream
//! .
//! The return value is the status.
template <typename T> inline typename ODESLV_GSL<T>::STATUS
ODESLV_GSL<T>::bounds_ASA
( const unsigned int ns, const double*tk, const T*Ip, T**Ixk,
  T*If, T**Ilk, T*Idf, const unsigned int nsamp, std::ostream&os )
{
  int DISPLAY_SAVE = options.DISPLAY;
  options.DISPLAY = 0;
  STATUS flag = NORMAL;
  
  // Initialization of sampled bounds at parameter lower bound
  double *p = new double[_np];
  for( unsigned int ip=0; ip<_np; ip++ )
    p[ip] = Op<T>::l(Ip[ip]);
  double **xk = new double*[ns+1];
  double **lk = new double*[ns+1];
  for( unsigned int is=0; is<=ns; is++ ){
    xk[is] = new double[_nx];
    lk[is] = new double[_nf*_nx];
  }
  double *f = If? new double[_nf]: 0;
  double *df = new double[_nf*_nx];
  flag = states_ASA( ns, tk, p, xk, f, lk, df, os );
  if( flag != NORMAL || nsamp <= 1 ){
    delete[] p;
    for( unsigned int is=0; is<=ns; is++ ) delete[] xk[is]; delete[] xk;
    for( unsigned int is=0; is<=ns; is++ ) delete[] lk[is]; delete[] lk;
    delete[] f; delete[] df;
    return flag;
  }   
  for( unsigned int is=0; is<=ns; is++ ){
    for( unsigned int ix=0; ix<_nx; ix++ )
      Ixk[is][ix] = xk[is][ix];
    for( unsigned int ix=0; ix<_nf*_nx; ix++ )
      Ilk[is][ix] = lk[is][ix];
  }
  for( unsigned int ifn=0; If && ifn<_nf; ifn++ ){
    If[ifn] = f[ifn];
    for( unsigned int ip=0; ip<_np; ip++ )
      Idf[ifn*_np+ip] = df[ifn*_np+ip];
  }
  
  // Start sampling process
  unsigned int* vsamp = new unsigned int[_np];
  flag = _states_ASA( ns, tk, Ip, Ixk, If, Ilk, Idf, nsamp, vsamp, 0, p, xk, f, lk, df, os );

  // Display results
  options.DISPLAY = DISPLAY_SAVE;
  if( options.DISPLAY >= 1 ){
    for( unsigned int is=0; is<=ns; is++ )
      _print_interm( tk[is], _nx, Ixk[is], "x", os );
    if( If ) _print_interm( "function", _nf, If, "f", os );
    for( unsigned int is=0; is<=ns; is++ )
      _print_interm( tk[ns-is], _nx*_nf, Ilk[ns-is], "l", os );
    _print_interm( "gradient", _nf*_np, Idf, "df", os );
  }

  // Record intermediate results
  _results.clear();
  if( options.RESRECORD )
    for( unsigned int is=0; is<=ns; is++ )
      _results.push_back( Results( tk[is], _nx, Ixk[is], _nf*_nx, Ilk[is] ) );

  // Clean-up
  delete[] p; delete[] f; delete[] df;
  for( unsigned int is=0; is<=ns; is++ ) delete[] xk[is]; delete[] xk;
  for( unsigned int is=0; is<=ns; is++ ) delete[] lk[is]; delete[] lk;
  delete[] vsamp;
  
  return flag;
}

template <typename T> inline void
ODESLV_GSL<T>::record
( std::ofstream&bndrec, const unsigned int iprec ) const
{
  if( !bndrec ) return;

  // Specify format
  bndrec << std::right << std::scientific << std::setprecision(iprec);

  // Record computed interval bounds at stage times
  typename std::vector< Results >::const_iterator it = _results.begin();
  for( ; it != _results.end(); ++it ){
    bndrec << std::setw(iprec+9) << (*it).t;
    for( unsigned int ix=0; ix<(*it).nx; ix++ )
      bndrec << std::setw(iprec+9) << mc::Op<T>::l( (*it).X[ix] )
             << std::setw(iprec+9) << mc::Op<T>::u( (*it).X[ix] );
    bndrec << std::endl;
  }
}

template <typename T> template<typename U> inline void
ODESLV_GSL<T>::_print_interm
( const double t, const unsigned nx, const U*x, const std::string&var,
  std::ostream&os )
{
  os << " @t = " << std::scientific << std::setprecision(6)
                 << std::left << t << " :" << std::endl;
  for( unsigned int ix=0; ix<nx; ix++ )
    os << " " << var.c_str() << "[" << ix << "] = " << x[ix] << std::endl;
  return;
}

template <typename T> template<typename U> inline void
ODESLV_GSL<T>::_print_interm
( const std::string&header, const unsigned nx, const U*x, const std::string&var,
  std::ostream&os )
{
  os << header.c_str() << " :" << std::endl
     << std::scientific << std::setprecision(6) << std::left;
  for( unsigned int ix=0; ix<nx; ix++ )
    os << " " << var.c_str() << "[" << ix << "] = " << x[ix] << std::endl;
  return;
}

} // end namescape mc

#endif

