const unsigned int NTM   = 3;	// <- Order of Taylor model
const unsigned int NSAMP = 10;	// <- Number of sampling points
#undef  SAVE_RESULTS		// <- Whether to save bounds to file
#undef  TEST_CONVERGENCE	// <- Whether to test Hausdorff convergence of bounds
#define USE_CMODEL		// <- whether to use Chebyshev models or Taylor models

#include "odebnd_gsl.hpp"
#include "odebnd_val.hpp"

#include "interval.hpp"
typedef mc::Interval I;
typedef mc::Ellipsoid E;

#ifdef USE_CMODEL
  #include "cmodel.hpp"
  typedef mc::CModel<I> TM;
  typedef mc::CVar<I> TV;
#else
  #include "tmodel.hpp"
  typedef mc::TModel<I> TM;
  typedef mc::TVar<I> TV;
#endif

int main()
{
  mc::FFGraph IVP;  // DAG describing the problem

  double t0 = 0., tf = 5.;   // Time span
  const unsigned int NS = 1;  // Time stages
  double tk[NS+1]; tk[0] = t0;
  for( unsigned k=0; k<NS; k++ ) tk[k+1] = tk[k] + (tf-t0)/(double)NS;

  const unsigned NP = 1;  // Parameter dimension
  const unsigned NX = 2;  // State dimension

  mc::FFVar P[NP];  // Parameter array
  for( unsigned i=0; i<NP; i++ ) P[i].set( &IVP );

  mc::FFVar X[NX];  // State array
  for( unsigned i=0; i<NX; i++ ) X[i].set( &IVP );

  mc::FFVar RHS[NX];  // Right-hand side function
  RHS[0] = P[0] * X[0] * ( 1. - X[1] );
  RHS[1] = P[0] * X[1] * ( X[0] - 1. );

  mc::FFVar IC[NX*NS];  // Initial value function
  IC[0] = 1.2;
  IC[1] = 1.1;

  I Ip[NP] = { I(2.95,3.05) };
  I *Ixk[NS+1];
  for( unsigned k=0; k<=NS; k++ )
    Ixk[k] = new I[NX];

  TM TMEnv( NP, NTM );
  TV TMp[NP] = { TV( &TMEnv, 0, Ip[0] ) };
  TV *TMxk[NS+1];
  double *Hxk[NS+1];
  for( unsigned k=0; k<=NS; k++ ){
    TMxk[k] = new TV[NX];
    Hxk[k] = new double[NX];
  }
  E RExk[NS+1];

  /////////////////////////////////////////////////////////////////////////
  // Bound ODE trajectories - sampling
  mc::ODESLV_GSL<I> LV0;

  LV0.set_dag( &IVP );
  LV0.set_state( NX, X );
  LV0.set_parameter( NP, P );
  LV0.set_dynamic( RHS );
  LV0.set_initial( IC );

  LV0.options.DISPLAY = 1;
  LV0.options.ATOL    = LV0.options.RTOL = 1e-10;
  LV0.options.INTMETH = mc::ODESLV_GSL<I>::Options::MSBDF;
#if defined( SAVE_RESULTS )
  LV0.options.RESRECORD = true;
#endif

  std::cout << "\nNON_VALIDATED INTEGRATION - APPROXIMATE ENCLOSURE OF REACHABLE SET:\n\n";
  LV0.bounds( NS, tk, Ip, Ixk, 0, NSAMP );
#if defined( SAVE_RESULTS )
  std::ofstream apprec( "approximate.out", std::ios_base::out );
  LV0.record( apprec );
#endif

  /////////////////////////////////////////////////////////////////////////
  // Bound ODE trajectories - validated integrator
  mc::ODEBND_VAL<I,TM,TV> LV1;

  LV1.set_dag( &IVP );
  LV1.set_state( NX, X );
  LV1.set_parameter( NP, P );
  LV1.set_dynamic( RHS );
  LV1.set_initial( IC );

  LV1.options.DISPLAY   = 1;
#if defined( SAVE_RESULTS )
  LV1.options.RESRECORD = true;
#endif
  LV1.options.SCALING   = false;
  LV1.options.TSTOP     = false;
  LV1.options.TSORDER   = 5;
  LV1.options.PMVALID   = false;
  LV1.options.HSTAB     = false;
  LV1.options.TOL       = 1e-10;
  //LV1.options.QTOL      = 1e-12;
  LV1.options.ORDMIT    = 1; //TMp->nord();
  LV1.options.WRAPMIT   = mc::ODEBND_VAL<I,TM,TV>::Options::ELLIPS;//NONE;

  std::cout << "\nDISCRETIZED SET-VALUED INTEGRATION - POLYNOMIAL MODEL ENCLOSURE OF REACHABLE SET:\n\n";
  LV1.bounds( NS, tk, TMp, TMxk );
#if defined( SAVE_RESULTS )
  std::ofstream bnd1rec( "bounds_gsl.out", std::ios_base::out );
      LV1.record( bnd1rec );
#endif

#if defined( TEST_CONVERGENCE )
  LV1.options.DISPLAY = 0;
  std::cout << std::scientific << std::setprecision(5);
  for( unsigned int k=0; k<20; k++ ){
    TV TMp0[NP] = { TV( &TMEnv, 0, Ip[0]/pow(1.5,k) ) };
    LV1.hausdorff( NS, tk, TMp0, Hxk, NSAMP );
    std::cout << mc::Op<I>::diam( TMp0[0].B() ) << "  " << Hxk[NS][0] << std::endl;
  }
#endif

  /////////////////////////////////////////////////////////////////////////
  // Bound ODE trajectories - differential inequalities
#ifdef USE_CMODEL
  mc::ODEBND_GSL<I,TM,TV> LV2;
#else
  mc::ODEBND_GSL<I> LV2;
#endif
  LV2.set_dag( &IVP );
  LV2.set_state( NX, X );
  LV2.set_parameter( NP, P );
  LV2.set_dynamic( RHS );
  LV2.set_initial( IC );

  LV2.options.DISPLAY   = 1;
#if defined( SAVE_RESULTS )
  LV2.options.RESRECORD = true;
#endif
  LV2.options.ORDMIT    = 1; //TMp->nord();
  LV2.options.ATOL      = LV2.options.RTOL = 1e-10;
  LV2.options.WRAPMIT   = mc::ODEBND_GSL<I,TM,TV>::Options::ELLIPS;//NONE;
  LV2.options.ODESLVOPT = LV0.options;

  std::cout << "\nCONTINUOUS SET-VALUED INTEGRATION - POLYNOMIAL MODEL ENCLOSURE OF REACHABLE SET:\n\n";
  //LV2.bounds( NS, tk, Ip, Ixk );
  LV2.bounds( NS, tk, TMp, TMxk, RExk );
#if defined( SAVE_RESULTS )
  std::ofstream bnd2rec( "bounds_val.out", std::ios_base::out );
  LV2.record( bnd2rec );
#endif
  //LV2.hausdorff( NS, tk, Ip, Hxk, NSAMP );
  //LV2.hausdorff( NS, tk, TMp, Hxk, NSAMP );

#if defined( TEST_CONVERGENCE )
  LV2.options.DISPLAY = 0;
  std::cout << std::scientific << std::setprecision(5);
  for( unsigned int k=0; k<20; k++ ){
    TV TMp0[NP] = { TV( &TMEnv, 0, Ip[0]/pow(1.5,k) ) };
    //LV2.hausdorff( NS, tk, Ip, Hxk, NSAMP );
    LV2.hausdorff( NS, tk, TMp0, Hxk, NSAMP );
    std::cout << mc::Op<I>::diam( TMp0[0].B() ) << "  " << Hxk[NS][0] << std::endl;
  }
#endif

  for( unsigned k=0; k<=NS; k++ ){
    delete[] Ixk[k];
    delete[] TMxk[k];
    delete[] Hxk[k];
  }

  return 0;
}


