const unsigned int NPM   = 2;	// <- Order of Taylor model
const unsigned int NSAMP = 3;	// <- Number of sampling points
#undef  SAVE_RESULTS		// <- Saving the results to file
#define USE_CMODEL		// <- Use Chebyshev models?

#include "odebnd_gsl.hpp"
#include "odebnd_val.hpp"

#include "interval.hpp"
typedef mc::Interval I;

#ifdef USE_CMODEL
  #include "cmodel.hpp"
  typedef mc::CModel<I> PM;
  typedef mc::CVar<I> PV;
#else
  #include "tmodel.hpp"
  typedef mc::TModel<I> PM;
  typedef mc::TVar<I> PV;
#endif

int main()
{
  mc::FFGraph IVP;  // DAG describing the problem

  double t0 = 0., tf = 20.;   // Time span
  const unsigned int NS = 1;  // Time stages
  double tk[NS+1]; tk[0] = t0;
  for( unsigned k=0; k<NS; k++ ) tk[k+1] = tk[k] + (tf-t0)/(double)NS;

  const unsigned NP = 6;  // Parameter dimension
  const unsigned NX = 3;  // State dimension

  mc::FFVar P[NP];  // Parameter array
  for( unsigned i=0; i<NP; i++ ) P[i].set( &IVP );

  mc::FFVar X[NX];  // State array
  for( unsigned i=0; i<NX; i++ ) X[i].set( &IVP );

  mc::FFVar RHS[NX];  // Right-hand side function
  mc::FFVar &X1 = X[0], &X2 = X[1], &X3 = X[2],
            &X10 = P[0], &X20 = P[1], &X30 = P[2],
            &MUM = P[3], &KS = P[4], &A = P[5];              
  double X3M = 50., D = 0.202, X2F = 20, Y = 0.4, B = 0.2;
  mc::FFVar MU = MUM * ( 1. - X3 / X3M ) * X2 / ( KS + X2 );
  RHS[0] = ( MU - D ) * X1;
  RHS[1] = D * ( X2F - X2 ) - MU * X1 / Y;
  RHS[2] = - D * X3 + ( A * MU + B ) * X1;

  mc::FFVar IC[NX*NS];  // Initial value function
  IC[0] = X10;
  IC[1] = X20;
  IC[2] = X30;

  I Ip[NP] = { I(6.48,6.52), I(4.99,5.01), I(14.99,15.01),
               I(0.464,0.466), I(1.074,1.076), I(2.19,2.21) };
  I *Ixk[NS+1];
  for( unsigned k=0; k<=NS; k++ ) Ixk[k] = new I[NX];

  PM PMEnv( NP, NPM );
  PV PMp[NP];
  for( unsigned i=0; i<NP; i++ ) PMp[i].set( &PMEnv, i, Ip[i] );
  PV *PMxk[NS+1];
  for( unsigned k=0; k<=NS; k++ ) PMxk[k] = new PV[NX];

  /////////////////////////////////////////////////////////////////////////
  // Bound ODE trajectories - sampling
  mc::ODESLV_GSL<I> LV0;

  LV0.set_dag( &IVP );
  LV0.set_state( NX, X );
  LV0.set_parameter( NP, P );
  LV0.set_dynamic( RHS );
  LV0.set_initial( IC );

  LV0.options.DISPLAY = 1;
  LV0.options.ATOL = LV0.options.RTOL = 1e-10;
  LV0.options.INTMETH = mc::ODESLV_GSL<I>::Options::MSBDF;
#if defined( SAVE_RESULTS )
  LV0.options.RESRECORD = true;
#endif

  LV0.bounds( NS, tk, Ip, Ixk, 0, NSAMP );
#if defined( SAVE_RESULTS )
  std::ofstream apprec( "approximate.out", std::ios_base::out );
  LV0.record( apprec );
#endif

  /////////////////////////////////////////////////////////////////////////
  // Bound ODE trajectories - validated integrator
  mc::ODEBND_VAL<I,PM,PV> LV1;

  LV1.set_dag( &IVP );
  LV1.set_state( NX, X );
  LV1.set_parameter( NP, P );
  LV1.set_dynamic( RHS );
  LV1.set_initial( IC );

  LV1.options.DISPLAY   = 1;
#if defined( SAVE_RESULTS )
  LV1.options.RESRECORD = true;
#endif
  LV1.options.TSTOP     = false;
  LV1.options.TSORDER   = 5;
  LV1.options.PMVALID   = false;
  LV1.options.HSTAB     = false;
  LV1.options.ORDMIT    = 1; //TMp->nord();
  LV1.options.WRAPMIT   = mc::ODEBND_VAL<I,PM,PV>::Options::ELLIPS;//NONE;

  LV1.bounds( NS, tk, PMp, PMxk );
#if defined( SAVE_RESULTS )
  std::ofstream bnd1rec( "bounds_val.out", std::ios_base::out );
  LV1.record( bnd1rec );
#endif

  /////////////////////////////////////////////////////////////////////////
  // Bound ODE trajectories - differential inequalities
  mc::ODEBND_GSL<I,PM,PV> LV2;

  LV2.set_dag( &IVP );
  LV2.set_state( NX, X );
  LV2.set_parameter( NP, P );
  LV2.set_dynamic( RHS );
  LV2.set_initial( IC );

  LV2.options.DISPLAY   = 1;
#if defined( SAVE_RESULTS )
  LV2.options.RESRECORD = true;
#endif
  LV2.options.ORDMIT    = 1; //TMp->nord();
  LV2.options.WRAPMIT   = mc::ODEBND_GSL<I,PM,PV>::Options::ELLIPS;//NONE;

  //LV2.bounds( NS, tk, Ip, Ixk );
  LV2.bounds( NS, tk, PMp, PMxk );
#if defined( SAVE_RESULTS )
  std::ofstream bnd2rec( "bounds_gsl.out", std::ios_base::out );
  LV2.record( bnd2rec );
#endif

  for( unsigned k=0; k<=NS; k++ ){
    delete[] Ixk[k];
    delete[] PMxk[k];
  }

  return 0;
}


